package clarion.acs;
import clarion.system.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to manage a set of VCGs corresponding to one original output
 * dimension PLUS an additional signal VCG at its end.
 * SIGNAL VCG : determine ONE or MULTIPLE active values at a time.
 * DIMENSION type : SINGAL_ACTIVE_VALUED or MULTI_ACTIVE_VALUED.
 * if dimension type = SINGAL_ACTIVE_VALUED, the candidate from signal VCG is FIXED:
 * allows only ONE active value at a time.
 * if dimension type = MULTI_ACTIVE_VALUED, the candidate from signal VCG is VARIED :
 * allows ONE or MULTIPLE active values at a time.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public short[] decideActiveValues(): Decides active elements from the VCG set
 * (EXCEPT the SIGNAL VCG) according to the dimension type. called by BL action.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is to implement a set of "Value Competition Grouping", new in
 * CLARION mdoel. It makes easy implementation of multiple-values in one dimension.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>AcsAction</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class VCGSet {

  /** dimension type : single-active-valued, multiple-active-valued. */
  protected int type;
  /** full expansion length of this VCG set. */
  protected int fullLength;
  /** the num of VCG in this VCG set. */
  protected int vcgNum;
  /** # of possible values in each VCG. */
  protected short[] elemNums;
  /** start point of each VCG in the format of one dimensional array. */
  protected short[] vcgOffsets;
  /** a set of VCG groups corresponding to one original output dimension. PLUS
   *  an additional signal VCG at its end. */
  protected VCG[] vcgs;
  /** candidates from all of the VCGs. */
  protected short[] candidates;

  /** constructor.
   *  @param type dimension type.
   *  @param vcgNum number of VCG.
   *  @param elemNums number of elements in each VCG.
   *  Both vcgNum and elemNums consider the signal VCG.  */
  public VCGSet( int type, int vcgNum, short[] elemNums )
  {
     this.type = type;
     this.vcgNum = vcgNum;
     this.elemNums = elemNums;
     vcgs = new VCG[ vcgNum ];
     for( int i = 0; i < vcgNum; i++ )
          vcgs[i] = new VCG( elemNums[i] );
     candidates = new short[ vcgNum ];
     fullLength = 0;
     for( int i = 0; i < vcgNum; i++ )
          fullLength += elemNums[i];
     vcgOffsets = new short[vcgNum];
     CommonMethods.calcOffsets( vcgOffsets, elemNums );
  }

  /** Returns full length.
   *  @return full length. */
  public int getFullLength()
  {
     return fullLength;
  }

  /** Returns the dimension type : single-active-valued, multiple-active-valued.
   *  @return the dimension type. */
  public int getType()
  {
     return type;
  }

  /** Returns the VCG num.
   *  @return the VCG num. */
  public int getVCGNum()
  {
     return vcgNum;
  }

  /** # of possible values in each VCG. */
  public short[] getElemNums()
  {
     return elemNums;
  }

  /** Returns the candidates.
   *  @return the candidates. */
  public short[] getCandidates()
  {
     return candidates;
  }

  /** Returns the signal candidate.
   *  @return the signal candidate. */
  public short getSignalCandidate()
  {
     return candidates[vcgNum-1];
  }

  /** Returns the activations.
   *  @return the activations. */
  public double[] getActivations()
  {
     double[] activations = new double[ fullLength ];
     for( int i = 0; i < vcgNum; i++ )
          System.arraycopy( vcgs[i].getActivations(), 0, activations, vcgOffsets[i], elemNums[i] );
     return activations;
 }

  /** Returns the activation of a specific element.
   *  @param vcgIdx an index to a VCG.
   *  @param elemIdx an index to an element in the VCG.
   *  @return the activation of a specific element. */
  public double getActivation( int vcgIdx, int elemIdx )
  {
     return vcgs[vcgIdx].getActivations()[elemIdx];
  }

  /** Returns the MAX activation of a specific VCG.
   *  @param vcgIdx an index to a VCG.
   *  @return the MAX activation of a specific VCG. */
  public double getMaxActivation( int vcgIdx )
  {
     return vcgs[vcgIdx].getMaxActivation();
  }

  /** Sets activations.  called by BL action.
   * @param activations the activations calculated by BL. */
  public void setActivations( double[] activations )
  {
    int offset = 0;
    for( int i = 0; i < vcgNum; i++ ){
         double[] activation = new double[ elemNums[i] ];
         for( int j = 0; j < elemNums[i]; j++ )
              activation[ j ] = activations[ offset + j ];
         vcgs[i].setActivations( activation );
         offset += elemNums[i];
    }
  }

  /** Sets the active values of this dimension. called by TL action.
   * assumed that the input 'activeValues' doesn't include signal value.
   * @param activations the activations to set. */
  public void setActiveValues( short[] activeValues )
  {
     for( int i = 0; i < vcgNum; i++ )
          candidates[i] = Global.EMPTY;
     // fill in the signal VCG.
     candidates[vcgNum-1] = (short)type;
     vcgs[vcgNum-1].setActivations( candidates[vcgNum-1] );
     // fill in the active values
     if( type == Global.SINGLE_ACTIVE_VALUED ){ // only allows one active value at a time.
         candidates[0] = activeValues[0];
         vcgs[0].setActivations( candidates[0] );
         return;
     }
     else // allows multiple active values at a time.
        // ENABLE the dimensional-value for all active VCGs.
        for( int i = 0; i < activeValues.length; i++ ){
             int idx = activeValues[i];
             candidates[idx] = Global.ELEM_ACTIVE;
             vcgs[idx].setActivations( candidates[idx] );
        }
        // DISABLE the dimensional-value for all inactive VCGs.
        for( int i = 0; i < vcgNum - 1; i++ )
             if( candidates[i] == Global.EMPTY ) {
                 candidates[i] = Global.ELEM_INACTIVE;
                 vcgs[i].setActivations( candidates[i] );
             }
  }

  /** Decides active elements from the VCG set (EXCEPT the SIGNAL VCG) according
   *  to the dimension type. called by BL action.
   *  @return the active values. */
  public short[] decideActiveValues()
  {
     // decide candidates from all of the VCGs.
     for( int i = 0; i < vcgNum; i++ )
          candidates[i] = (short)vcgs[i].decideCandidate();
     short[] activeValues =  new short[vcgNum-1];
     // find the active value if the type is SINGLE_ACTIVE_VALUED.
     if( type == Global.SINGLE_ACTIVE_VALUED ){  // only allows one active value.
         activeValues[0] = candidates[0];
         candidates[vcgNum-1] = (short)type;
         return activeValues;
     }
     // find the active values if type is MULTI_ACTIVE_VALUED.
     int count = 0;
     for( int i = 0; i < vcgNum - 1; i++ ){
          if( candidates[i] == Global.ELEM_ACTIVE ){
              activeValues[count] = (short)i;
              count++;
          }
     }
     if( count == 0 )
         return (short[])null;
     // if signal MULTI_ACTIVE_VALUED is ON.
     if( candidates[vcgNum-1] == Global.MULTI_ACTIVE_VALUED ){
         short[] values = new short[count];
         System.arraycopy( activeValues, 0, values, 0, count );
         return values;
     }
     // if signal MULTI_ACTIVE_VALUED is OFF.
     // select one active among the candidates from all of the activated VCGs except the signal VCG.
     double[] activations = new double[count];
     for( int i = 0; i < count; i++ ){
          int idx = activeValues[i];
          activations[i] = vcgs[idx].getActivations()[Global.ELEM_ACTIVE];
     }
     short[] finalActiveValues = new short[1];
     int idx = new StochasticDecider( count, 0.1 ).getRecommendation( activations );
     finalActiveValues[0] = activeValues[idx];
     return finalActiveValues;
  }
}