 /************************** class RerRule begin ********************************/
 package clarion.acs.net;
 import java.util.Arrays;
 import clarion.acs.net.auxiliary.*;
 import clarion.system.*;
 import clarion.common.*;
 import clarion.tools.*;

 /**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement an RER (Rule Extraction Refinement) rule. It extends
 * the more generic rule class <em>TLRule</em>.
 *
 * Notes: in this class,
 * 1. basic match: match all condition dims and the aciton.
 * 2. general match: match all condition dims except one and the action.
 * 3. basic condition match: match all condition dims.
 * 4. general condition match: match all condition dims except one.
 *
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean canGeneralized(): returns if this rule can be geneialized.
 * <li> public boolean canSpecialized(): returns if this rule can be specialized.
 * <li> public boolean decideAction( short[] state, short[] exState, short[] gsState,
 * short[] wmState ): Decides an action by this RER rule.
 * <li> public void update( boolean successFlag, int curTime ): update rontine.
 * <li> public void updateStatistics( boolean successFlag, int matchMode ):
 * updates the rule statistics whenever a match exists.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the RER rule which is one of the three rule knowledge
 * types in the top level of an ACS action-decision network.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>TLRuleSet</em> object.
 * @author Xi Zhang
 *
 */

 public class RerRule extends TLRule {

    /** 1. basic match: match all condition dims and the aciton.
     *  2. general match: match all condition dims except one and the action.
     *  3. basic condition match: match all condition dims.
     *  4. general condition match: match all condition dims except one. */
     protected final int BASIC_MATCH = 0, GENERAL_MATCH = 1;
     /** used for PM, NM statistics. */
     protected final int WITH = 0, WITHOUT = 1;
     /** used for statistics. */
     protected final int PM = 0, NM = 1, RATE = 2;
     /** general-to-specific or specific-to-general. */
     protected int genSpecMode;
     /** condition values in each dim. */
     protected short condValNums[];
     /** location of the max matching rate in the array of statistics. */
     protected int[] potentialLocs[];
     /** if a match-all rule. */
     protected boolean MATCH_ALL_FLAG;
     /** if extraction criterion is met or not. */
     protected boolean extractFlag;
     /** if generalization criterion is met or not. */
     protected boolean geneFlag;
     /** if specialization criterion is met or not. */
     protected boolean specFlag;
     /** basic condition matched flag. */
     protected boolean basicCondMatchFlag;
     /** basic condition & conclusion matched flag. */
     protected boolean basicMatchFlag;
     /** general condition matched flag. */
     protected boolean geneCondMatchFlag;
     /** general condition & conclusion matched flag. */
     protected boolean geneMatchFlag;
     /** action matched flag. */
     protected boolean actionMatchFlag;
     /** used for calculating IG. */
     protected double  c1, c2;
     /** statistics for storing the PM, NM and success rates.
      *  the 3rd dimension records statistics of condition conditions[i][j]
         0: statistics WITH the condition "conditions[i][j]=1"
         1: statistics WITHOUT the condition "conditions[i][j]=1"
         the 4th dimension
         0: Positive Matches, 1: Negative Matches, 2: Rate PM+1/PN+NM+2 */
     protected float[][][][] statistics;

     /** constructor.
      *  @param netIdx the network index.
      *  @param global the associated global settings. */
     public RerRule( int subsysIdx, int netIdx, Global global )
     {
        super( subsysIdx, netIdx, Global.RER, 0, global );
     }

     /** constructor.
      *  @param netIdx the network index.
      *  @param global the associated global settings.
      *  @param state current state.
      *  @param action current action.
      *  @param mFlag match-all flag. */
     public RerRule( int subsysIdx, int netIdx, Global global, short[] state, short[][][] action, boolean mFlag )
     {
         super( subsysIdx, netIdx, Global.RER, 0, global );
         genSpecMode = global.getRerGenSpecMode( netIdx );
         MATCH_ALL_FLAG = mFlag;
         extractFlag = true;  // new extracted rule.
         condValNums = new short[ INPUT_DIM_NUM ];
         potentialLocs = new int[2][2];
         // initialize stats.
         statistics = new float[ INPUT_DIM_NUM ][][][];
         for( int i = 0; i < inputDVs.length; i++ )
             statistics[i] = new float[ inputDVs[i] ][2][3];
         // initialzie states and suggest action.
         conditions = new short[ INPUT_DIM_NUM ][];
         for( int i = 0; i < inputDVs.length; i++ )
              conditions[ i ] = new short[ inputDVs[i] ];
         convertToRuleConditions( state );
         suggestedAction = CommonMethods.copy( action );
     }

/*************************** BOOLEAN METHODS **********************************/
     /** Returns if the basic condition match is met. */
     public boolean isBasicConditionMatched()
     {
         return basicCondMatchFlag;
     }

     /** Returns if the general condition match is met. */
     public boolean isGeneralConditionMatched()
     {
         return geneCondMatchFlag;
     }

     /** Returns if the basic match is met. */
     public boolean isBasicMatched()
     {
         return basicMatchFlag;
     }

     /** Returns if the general match is met. */
     public boolean isGeneralMatched()
     {
         return geneMatchFlag;
     }

     /** Returns if extraction criterion is met. */
     public boolean isExtracted()
     {
         return extractFlag;
     }

     /** Returns if generailizion criterion is met. */
     public boolean isGenerailized()
     {
         return geneFlag;
     }

     /** Returns if speciailizion criterion is met. */
     public boolean isSpecialized()
     {
         return specFlag;
     }

/***************************** GET METHODS ************************************/
     /** Returns the location of the max matching rate in the statistics array.
      *  used for rule generalization and specialization.
      *  @param mode: 0: generalization, 1: specialization. */
     public int[] getMaxIGLoc( int mode )
     {
        return potentialLocs[ mode ];
     }

     /** Returns the PM, NM matches of the rule. */
     public double[] getMatchRate()
     {
        double[] matches = new double[2];
        matches[PM] = attributes.getPM();
        matches[NM] = attributes.getNM();
        return matches;
     }

     /** Returns a match rate of a minor variation of the rule condition.
      *  @param idx1: index to one dimension.
      *  @param idx2: index to a dimensional value.
      *  @param mode = 0: generailization, = 1: specification.
      *  @return a match rate of a minor variation of the rule condition. */
     public double[] getMaxMatches( int idx1, int idx2, int mode )
     {
         double[] matches = new double[2];
         matches[PM] = statistics[ idx1 ][ idx2 ][ 1 - mode ][ PM ];
         matches[NM] = statistics[ idx1 ][ idx2 ][ 1 - mode ][ NM ];
         return matches;
     }

/***************************** SET METHODS ************************************/
     /** Sets the numbers of values in the condiiton dimensions.
      *  @param arr the array used to set the numbers. */
     public void setCondValNums( short[] arr )
     {
        for( int i = 0; i < arr.length; i++ )
              condValNums[ i ] = arr[ i ];
     }

     /** Sets the locations of the max matching rate in the array of statistics.
      *  @param arr the array used to set the locations. */
     public void setPotentialLocs( int[][] arr )
     {
        for( int i = 0; i < arr.length; i++ )
         for( int j = 0; j < arr[ i ].length; j++ )
              potentialLocs[ i ][ j ] = arr[ i ][ j ];
     }

     /** Sets the statistics.
      *  @param stats the array used to set the statistics. */
     public void setStatistics( float[][][][] stats )
     {
        for( int i = 0; i < stats.length; i++ )
         for( int j = 0; j  < stats[ i ].length; j++ )
          for( int k = 0; k < stats[ i ][ j ].length; k++ )
           for( int m = 0; m < stats[ i ][ j ][ k ].length; m++ )
                statistics[ i ][ j ][ k ][ m ] = stats[ i ][ j][ k ][ m ];
     }

     /** Sets the generailization flag.
      *  @param flag the flag used to set the generailization flag. */
     public void setGenerailizedFlag( boolean flag )
     {
         geneFlag = flag;
     }

     /** Sets the speciailization flag.
      *  @param flag the flag used to set the speciailization flag. */
     public void setSpecializedFlag( boolean flag )
     {
         specFlag = flag;
     }

     /** Resets statistics. */
     public void resetStatistics()
     {
        super.resetStatistics();
        for( int i = 0; i < potentialLocs.length; i++ )
          for( int j = 0; j < potentialLocs[i].length; j++ )
               potentialLocs[ i ][ j ] = 0;
        for( int i = 0; i < INPUT_DIM_NUM; i++)
          for( int j = 0; j < inputDVs[i]; j++)
            for( int k = 0; k < 2; k++ )
              for( int m = 0; m < 3; m++ )
                   statistics[i][j][k][m] = 0;
     }

     /** Search the location of the max matching rate in the array statistics.
         @param mode 0: generalization, mode 1: specialization.
         @return the locations. */
     public int[] searchMaxIGLoc( int mode )
     {
        double val, maxVal = 0.0;
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
            for( int j = 0; j < inputDVs[i]; j++ )
                if( conditions[i][j] == mode )
                {
                    val = statistics[i][j][1-mode][RATE];
                    if( maxVal < val )
                    {
                        maxVal = val;
                        potentialLocs[mode][0] = i;
                        potentialLocs[mode][1] = j;
                    }
                }
         return potentialLocs[mode];
     }

   /** Adds a value of one dimension to the rule condiiton. used for generaliztion.
    *  @param idx1 the index to one dimension.
    *  @param idx2 the index to one value in that dimension. */
   public void addCondition( int idx1, int idx2 )
   {
        conditions[ idx1 ][ idx2 ] = 1;
        condValNums[ idx1 ]++;
   }

   /** Deletes a value of one dimension to the rule condiiton. used for specializtion.
    *  @param idx1 the index to one dimension.
    *  @param idx2 the index to one value in that dimension. */
   public void delCondition( int idx1, int idx2 )
   {
        if( condValNums[ idx1 ] <= 1 )
            return;
        conditions[ idx1 ][ idx2 ] = 0;
        condValNums[ idx1 ]--;
   }

   /** Returns if this rule can be generalized. */
   public boolean canGeneralized()
   {
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
             if( i == potentialLocs[Global.GENERALIZATION ][0] && condValNums[ i ] < inputDVs[i] )
                 return true;
        return false;
   }

   /** Returns if this rule can be specialized. */
   public boolean canSpecialized()
   {
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
              if( i == potentialLocs[Global.SPECIFICATION][0] && condValNums[ i ] > 1  )
                  return true;   // modified 07/08/03
        return false;
   }

/************************** RULE MATCH METHODS *********************************/
   /** Checks if the specific action matches the conclusion of the rule. */
   public void matchAction( short[][][] action )
   {
        // check if the action is same as suggestedAction.
        boolean bValue = isActionMatched( action );
        if( bValue ){
            actionMatchFlag = true;
            basicMatchFlag = basicCondMatchFlag;
            geneMatchFlag = geneCondMatchFlag;
        }
        else{
            actionMatchFlag = false;
            basicMatchFlag = false;
            geneMatchFlag = false;
        }
   }

   /** Decides the action and and calculates the rule support.
    *  @param state curent state.
    *  @param exState external state.
    *  @param gsState goal structure state.
    *  @param wmState working memory state.
    *  @return <tt>true</tt> if this rule is matched.
    *  <tt>false>, otherwise. */
   public boolean decideAction( short[] state, short[] exState, short[] gsState, short[] wmState )
   {
        extractFlag = geneFlag = specFlag = false;
        int misMatchNum = isConditionMatched( state );
        if( misMatchNum == 0 ){  // match for all dims.
            basicCondMatchFlag = true;
            geneCondMatchFlag = false;
        }
        if( misMatchNum == 1 ){  // match for all dims except one.
            basicCondMatchFlag = false;
            geneCondMatchFlag = true;
        }
        if( misMatchNum > 1 ){
            basicCondMatchFlag = false;
            geneCondMatchFlag = false;
        }
        return basicCondMatchFlag;
   }

/************************** RULE UPDATE METHODS *******************************/
   /** Updates this rule with flag indicating current step is successful or not.
    *  and current time.
    *  @param successFlag indicate current step is successful or not.
    *  @param curTime current time. */
   public void update( boolean successFlag, int curTime )
   {
        super.update( basicCondMatchFlag, successFlag, curTime );
        if( basicMatchFlag )
            updateStatistics( successFlag, BASIC_MATCH );
        if( geneMatchFlag )
            updateStatistics( successFlag, GENERAL_MATCH );
   }

   /** Updates the match-all rule. the method is only used when this rule is a
    *  match-all rule. */
   public void updateMatchAll( boolean successFlag )
   {
        attributes.setAge( 0 ); // it means the rule is applied this step.
        updateStatistics( successFlag, BASIC_MATCH );
   }

   /** Updates the rule statistics whenever a match exists.
    *  @param successFlag indicate current step is successful or not.
    *  @matchMode: basic match or general match. */
   public void updateStatistics( boolean successFlag, int matchMode )
   {
        int matchType;
        if( successFlag )
            matchType = PM;
        else
            matchType = NM;
        if( matchMode == BASIC_MATCH ) // match all of dims.
            attributes.updateMatches( matchType );
        else
        {  // match all of dims but one.
           Arrays.sort( inputConditions[ DIFF_DIM_IDX ] );
           // update the stats of the input condition values in the mismatched dim
           for( int i = 0; i < inputConditions[DIFF_DIM_IDX].length; i++ )
           {
                int condIdx = inputConditions[ DIFF_DIM_IDX ][ i ];
                statistics[DIFF_DIM_IDX][condIdx][WITH][matchType] += 1;
           }
           // update the stats of the rule condition values except the above ones
           // in the mismatched dim.
           for( int i = 0; i < inputDVs[ DIFF_DIM_IDX ]; i++ )
           {
                if( conditions[DIFF_DIM_IDX][i] == 1 )
                {
                    int val = Arrays.binarySearch(inputConditions[DIFF_DIM_IDX],(short)i);
                    if( val < 0 ) // not included in the input condition.
                        statistics[DIFF_DIM_IDX][i][WITHOUT][matchType] += 1;
                }
           }
           return;
        }
        if( MATCH_ALL_FLAG )  // a match-all rule.
            return;
        // update statistics of the input conditions.
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
         for( int j = 0; j < inputConditions[i].length; j++ )
              statistics[i][inputConditions[i][j]][WITH][matchType] += 1;
        // update statistics of the rule conditions except the above ones.
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
        {
             Arrays.sort( inputConditions[i] );
             for( int j = 0; j < inputDVs[i]; j++ )
             {    // if not included in the input condition.
                  if(conditions[i][j]==1 && Arrays.binarySearch(inputConditions[i],(short)j)<0)
                     statistics[i][j][WITHOUT][matchType] += 1;
            }
        }
        // calculate all of the match rates.
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
          for( int j = 0; j < inputDVs[i]; j++ )
               for( int k = 0; k < 2; k++ )
               {
                    float val1 = statistics[i][j][k][PM];
                    float val2 = statistics[i][j][k][NM];
                    statistics[i][j][k][RATE] = (val1 + 1) / (val1 + val2 + 2);
               }
   }

   /** discount the rule statistics after each trial. */
   public void discountStatistics()
   {
       attributes.discountMatches( DISCOUNT );
       for( int i = 0; i < INPUT_DIM_NUM; i++ )
         for( int j = 0; j < inputDVs[i]; j++ )
            for( int k = 0; k < 2; k++ )
            {
                 statistics[i][j][k][PM] *= DISCOUNT;
                 statistics[i][j][k][NM] *= DISCOUNT;
                 float val1 = statistics[i][j][k][PM];
                 float val2 = statistics[i][j][k][NM];
                 statistics[i][j][k][RATE] = (val1 + 1) / (val1 + val2 + 2);
            }
   }

/************************** OVERRIDE METHODS **********************************/
   /** methods from class <em>Object</em> */
    public String toString()
    {
         int MAX_INPUT_DIM_NUM = INPUT_DIM_NUM;
         String ruleString = "{ ";
         for( int i = 0; i < MAX_INPUT_DIM_NUM; i++ ){
              ruleString += "(";
              for( int j = 0; j < inputDVs[i]; j++ ){
                   if( conditions[ i ][ j ] == 1 )
                      ruleString += ( j + 1 ) + " ";
              }
              ruleString += ")";
              if( i < MAX_INPUT_DIM_NUM - 1 )
                  ruleString += ",";
         }
         ruleString += " }";
         ruleString += " --> action ";

         int actionPartNum = (subsysIdx == Global.ACS && global.isNacsOn()) ? 2 : 1;
         for( int k = 0; k < actionPartNum; k++ ){
              ruleString += "{";
              for( int i = 0; i < suggestedAction[k].length; i++ ){
                    ruleString += "(";
                    for( int j = 0; j < suggestedAction[k][i].length; j++ )
                         ruleString += ( suggestedAction[k][i][j] + 1 ) + " ";
                    ruleString += ")";
                    if( i < suggestedAction[k].length - 1 )
                        ruleString += ",";
              }
              ruleString += "} \n";
         }
         return ruleString;
   }

   /** methods from class <em>Object</em> */
   public int hashCode()
   {
       int num = Math.min( 32, inputDVs[0] );
       int code = 0;
       for( int i = 0; i < num; i++ )
            code = (code <<= 1) + conditions[0][i];
       return code;
   }

   /** methods from class <em>Object</em> */
   public boolean equals( Object obj )
   {
       if( !(obj instanceof RerRule ) || obj == null )
            return false;
       RerRule otherRule = (RerRule)obj;
       // if this covers other rule
       int retVal = this.cover( otherRule );
       if( retVal >= 0 || retVal == -1 )
           return false;
       // if other rule covers this rule.
       retVal = otherRule.cover( this );
       if( retVal >= 0 || retVal == -1 )
           return false;
       return true;
   }

   /** methods from class <em>Object</em> */
   protected Object clone( )
   {
         short[] state = new short[ global.getIdnInputNum( netIdx ) ];
         RerRule rule = new RerRule( subsysIdx, netIdx, global, state, suggestedAction,  MATCH_ALL_FLAG );
         rule.setCondition( conditions );
         rule.setCondValNums( condValNums );
         rule.setPotentialLocs( potentialLocs );
         rule.setStatistics( statistics );
         rule.setAttributes( attributes );
         return rule;
   }

/************************** PRIVATE METHODS ************************************/
   /** Converts current state into rule condition. */
   private void convertToRuleConditions( short[] state )
   {    // decide the source of the rule condition for each dim.
        int NON_MASK_NUM = 1;
        int masks[] = new int[ INPUT_DIM_NUM ];
        if( ( genSpecMode == global.DBL_DIRECT && Math.random() < 0.5 ) ||
              genSpecMode == global.GEN_TO_SPEC )
        {
            for( int i = 0; i < INPUT_DIM_NUM; i++ )
                 masks[ i ] = 1;
            int count = 0;
            while( count < NON_MASK_NUM )
            {
                int dimIdx = (int)(Math.random() * INPUT_DIM_NUM);
                if( masks[ dimIdx ] == 1 )
                {
                    masks[ dimIdx ] = 0;
                    count++;
                }
            }
        }
        else
            for( int i = 0; i < INPUT_DIM_NUM; i++ )
                 masks[ i ] = 0;
        // generate the rule conditions.
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
          for( int j = 0; j < inputDVs[i]; j++)
                 conditions[i][j] = 0;
        for( int i = 0; i < INPUT_DIM_NUM; i++ )
        {
             for( int j = 0; j < inputDVs[i]; j++ )
             {
                  if( masks[ i ] == 0 )
                      conditions[ i ][ j ] = state[ inputOffsets[i] + j ];
                  else
                      conditions[ i ][ j ] = 1;
                  if( conditions[ i ][ j ] == 1 )
                      condValNums[ i ]++;
             }
        }
   }
 }