package clarion.acs.net;
import clarion.system.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement a variation of an RER (Rule Extraction Refinement)
 * rule. It extends the rule class <em>RerRule</em>. In this variation, different
 * generalization and specialization processes are used.
 *
 * Notes: in this class,
 * 1. basic match: match all condition dims and the aciton.
 * 2. general match: match all condition dims except one and the action.
 * 3. basic condition match: match all condition dims.
 * 4. general condition match: match all condition dims except one.
 *
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean canGeneralized(): returns if this rule can be geneialized.
 * <li> public boolean canSpecialized(): returns if this rule can be specialized.
 * <li> public void updateStatistics( boolean successFlag, int matchMode ):
 * updates the rule statistics whenever a match exists.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements a variaiton of RER rule which is one of the three rule
 * knowledge types in the top level of an ACS action-decision network.
 * <li> Usage of this class. <br>
 * It is instantiated inside an <em>TLRuleSet</em> object.
 * @author Xi Zhang
 *
 */

public class RerRuleEx extends RerRule {

     /** record statistics of all the values in conditions in order to determine
      *  which dimension will be specialized. the 3rd dimension:
      *  0: Positive Matches, 1: Negative Matches, 2: Rate PM + 1/ PN + NM + 2  */
     protected float[][][] specStats;
     /** record statistics of all the dimensions of conditions in order to determine
      *  which dimension will be generalized. the 2nd dimension:
      *  0: Positive Matches, 1: Negative Matches, 2: Rate PM + 1/ PN + NM + 2  */
     protected float[][] genStats;
     /** used for which dimension of conditions can be generalized or specialized
      *  only record the valid dimensions which has non-zero values in its elements
      *  1st dimension contains the indices of all the valid dimensions.
      *  2nd dimension indicates the corresponding dimension can be generalized or
      *  specialized.
      *  = 0: can be generalized; = 1: can be specialized. */
     protected int[][] dimGenSpecInfo;
     /** VALID_DIM_NUM means the # of dimensions which has non-zero elements. */
     protected int VALID_DIM_NUM;
     /** indicating the dimension which can generalized at the most priority. */
     protected int genIdx;
     /** indicating the dimension and one of its elements which can be specialized
      *  at the most priority. */
     protected int specIdx[];

 /** constructor.
  *  @param netIdx the network index.
  *  @param global the associated global settings.
  *  @param state current state.
  *  @param action current action.
  *  @param mFlag match-all flag.
  *  @param vDimnum nubmer of valid dimensions. */
 public RerRuleEx( int subsysIdx, int netIdx, Global g, short[] state, short[][][] action, boolean mFlag, int vDimNum )
 {
        super( subsysIdx, netIdx, g, state, action, mFlag );
        VALID_DIM_NUM = vDimNum;
        genStats = new float[INPUT_DIM_NUM][3];
        specStats = new float[INPUT_DIM_NUM][][];
        for( int i = 0; i < inputDVs.length; i++ )
             specStats[i] = new float[inputDVs[i]][3];
        dimGenSpecInfo = new int[VALID_DIM_NUM][2];
        if( !MATCH_ALL_FLAG )
            setValidDims( state );
 }

 /** Sets the valid dimensions from current state.
  *  @param state current state. */
 public void setValidDims( short[] state )
 {
     int count = 0;
     for( int i = 0; i < INPUT_DIM_NUM; i++ )
      for( int j = 0; j < inputDVs[i]; j++ )
      {
           if( state[ inputOffsets[i] + j ] == 1 )
           {
               dimGenSpecInfo[count++][0] = i;
               break;
           }
      }
 }

 /** Returns the generalization/specialization info.
  *  @param idx index to one dimension.
  *  @return the info. */
 public int getDimGenSpecInfoIdx( int idx )
 {
     int retVal = 0;
     for( int i = 0; i < dimGenSpecInfo.length; i++ )
          if( dimGenSpecInfo[ i ][ 0 ] == idx )
          {
              retVal = i;
              break;
          }
     return retVal;
 }

 /** Returns the maximal matches with specific mode.
  *  @param mode: 0: generalizaiton; 1: specialization.
  *  @return the maximal matches. */
 public double[] getMaxMatches( int mode )
 {
     return(mode == 0) ? getMaxGenMatches(): getMaxSpecMatches();
 }

 /** Returns the maximal matches from generalization. */
 public double[] getMaxGenMatches()
 {
     double val1, val2, val, maxVal = 0.0;
     double[] maxVals = new double[2];
     for( int i = 0; i < VALID_DIM_NUM; i++ ){
          if( dimGenSpecInfo[i][1] == 0 ){
              val1 = genStats[dimGenSpecInfo[i][0]][0];
              val2 = genStats[dimGenSpecInfo[i][0]][1];
              val = ( val1 + 1 ) / ( val1 + val2 + 2 );
              if( val > maxVal ){
                  maxVal = val;     // modified by xi zhang 11/28/01
                  maxVals[PM] = val1;  // modified by xi zhang 5/5/03
                  maxVals[NM] = val2;  // modified by xi zhang 5/5/03
                  genIdx = dimGenSpecInfo[i][0];
              }
           }
     }
     return maxVals;
 }

 /** Returns the maximal matches from specialization. */
 public double[] getMaxSpecMatches()
 {
     double val1, val2, val, maxVal = 0.0;
     double[] maxVals = new double[2];
     specIdx = new int[2];
     for( int i = 0; i < VALID_DIM_NUM; i++ ){
          if( dimGenSpecInfo[i][1] == 1 ){
              for( int j = 0; j < inputDVs[i]; j++ ){
                   val1 = specStats[dimGenSpecInfo[i][0]][j][0];
                   val2 = specStats[dimGenSpecInfo[i][0]][j][1];
                   val = ( val1 + 1 ) / ( val1 + val2 + 2 );
                   if( val > maxVal ){
                       maxVal = val;        // modified by xi zhang 11/28/01
                       maxVals[PM] = val1;  // modified by xi zhang 5/5/03
                       maxVals[NM] = val2;  // modified by xi zhang 5/5/03
                       specIdx[0] = dimGenSpecInfo[i][0];
                       specIdx[1] = j;
                   }
              }
          }
     }
     return maxVals;
 }

 /** Sets the generalization/specialization info. */
 public void setDimGenSpecInfo( int[][] info )
 {
     for( int i = 0; i < info.length; i++ )
      for( int j = 0; j < info[ i ].length; j++ )
           dimGenSpecInfo[ i ][ j ] = info[ i ][ j ];
 }

 /** Sets the generalization statistics. */
 public void setGenStats( float[][] info )
 {
     for( int i = 0; i < info.length; i++ )
       for( int j = 0; j < info[ i ].length; j++ )
            genStats[ i ][ j] = info[ i ][ j ];
 }

 /** Sets the specialization statistics. */
 public void setSpecStats( float[][][] info )
 {
     for( int i = 0; i < info.length; i++ )
       for( int j = 0; j < info[ i ].length; j++ )
         for( int k = 0; k < info[ i ][ j ].length; k++ )
              specStats[ i ][ j ][ k ] = info[ i ][ j ][ k ];
 }

 /** Resets statistics. */
 public void resetStatistics()
 {
     attributes.resetMatches();
     for( int i = 0; i < INPUT_DIM_NUM; i++)
     {
          for( int j = 0; j < 3; j++ )
               genStats[i][j] = 0;
          for( int j = 0; j < inputDVs[i]; j++)
            for( int k = 0; k < 3; k++ )
                 specStats[i][j][k] = 0;
     }
 }

 /** Returns if this rule can be generalized. */
 public boolean canGeneralized()
 {
     for( int i = 0; i < VALID_DIM_NUM; i++ )
          if( dimGenSpecInfo[i][1] == 0 )
              return true;
     return false;
 }

 /** Returns if this rule can be specialized. */
 public boolean canSpecialized()
 {
     for( int i = 0; i < VALID_DIM_NUM; i++ )
          if( dimGenSpecInfo[i][1] == 1 )
              return true;
     return false;
 }

 /** Generalize one dimension. */
 public void generalizeADim()
 {
     for( int i = 0; i < inputDVs[i]; i++ )
          conditions[ genIdx ][ i ] = 1;
     int idx = getDimGenSpecInfoIdx( genIdx );
     // it means this dimension can only be specialized and not generalized in the future.
     dimGenSpecInfo[ idx ][ 1 ] = 1;
 }

 /** Specialize one dimension. */
 public void specializeADim()
 {
     for( int i = 0; i < inputDVs[i]; i++ )
          conditions[ specIdx[0] ][ i ] = 0;
     conditions[ specIdx[0] ][ specIdx[1] ] = 1;
     int idx = getDimGenSpecInfoIdx( specIdx[0] );
     //it means this dimension can only be generalized and not specialized in the future.
     dimGenSpecInfo[ idx ][ 1 ] = 0;
 }

   /** Updates the rule statistics whenever a match exists.
    *  @param successFlag indicate current step is successful or not.
    *  @matchMode: basic match or general match. */
 public void updateStatistics( boolean successFlag, int matchMode )
 {
     // match current condition and action.
     int matchType;
     if( successFlag )
         matchType = PM;
     else
         matchType = NM;
     if( matchMode == BASIC_MATCH ) // match all of dims.
         attributes.updateMatches( matchType );
     else
     {   // match all of dims but one.
         // update the stats of the input condition values in the mismatched dim
         for( int i = 0; i < inputConditions[DIFF_DIM_IDX].length; i++ )
         {
              int condIdx = inputConditions[ DIFF_DIM_IDX ][ i ];
              specStats[DIFF_DIM_IDX][condIdx][matchType] += 1;
         }
         genStats[DIFF_DIM_IDX][matchType] += 1;
         return;
     }
     if( MATCH_ALL_FLAG )
         return;
     // update statistics of the input conditions.
     for( int i = 0; i < INPUT_DIM_NUM; i++ )
       for( int j = 0; j < inputConditions[i].length; j++ )
            specStats[i][inputConditions[i][j]][matchType] += 1;
 }

 /** Discounts the rule statistics after each trial. */
 public void discountStatistics()
 {
     attributes.discountMatches( DISCOUNT );
     for( int i = 0; i < INPUT_DIM_NUM; i++ )
     {
          genStats[i][0] *= DISCOUNT;
          genStats[i][1] *= DISCOUNT;
          for( int j = 0; j < inputDVs[i]; j++ )
          {
               specStats[i][j][0] *= DISCOUNT;
               specStats[i][j][1] *= DISCOUNT;
          }
     }
 }

/******************************* PROTECTED METHODS *****************************/
 /** methods from class <em>Object</em> */
 protected Object clone( )
 {
     short[] state = new short[ global.getIdnInputNum( netIdx ) ];
     RerRuleEx rule = new RerRuleEx( subsysIdx, netIdx, global, state, suggestedAction,
                                     MATCH_ALL_FLAG, VALID_DIM_NUM );
     rule.setDimGenSpecInfo( dimGenSpecInfo );
     rule.setGenStats( genStats );
     rule.setSpecStats( specStats );
     return rule;
 }
}