package clarion.acs.wm;
import  clarion.system.*;
import clarion.common.*;
import  clarion.nacs.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the concept of working memory.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean actionSet( int idx ): the 'set' action.
 * <li> public boolean actionReset( int idx ): the 'reset' action.
 * <li> public void updateState( double[] state ): updates current WM state as
 * input to WMNet.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the concept of working memory which is for storing information
 * on a temporary basis for the express purpose of facilitating subsequent action
 * decision making or reasoning.
 * <li> Usage of this class. <br>
 * It is instantiated inside <em>WMNet</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class WM{

  /** global settings. */
  protected Global global;
  /** net index. */
  protected int    netIdx;
  /** size of WM. */
  protected int    size_WM;
  /** activation threshold. */
  protected double threshold_WM;
  /** BLA option: constant or recency-based. */
  protected int    blaOption;
  /** three parameters for BLA: c, d, iBLA. */
  protected double blaParams[];
  /** current time, used for calculating BLA. */
  protected long   currentTime;
  /** the length of each WM item. */
  protected short  wmItemLengths[];
  /** start points of each item in overall input in terms of one dimension array. */
  protected int    offsets[];
  /** indicates the flag items is on or not. */
  protected boolean flagItemOns[];
  /** working memory items. */
  protected Chunk  items[];

  /** constructor.
   *  @param g the global setting. */
  public WM( Global g )
  {
      global = g;
      netIdx = global.getWmNetIdx();
      size_WM = global.getWmSize();
      threshold_WM = global.getWmActivationThreshold();
      blaOption = global.getWmBlaOption();
      blaParams = global.getWmBlaParas();
      wmItemLengths = global.getWmItemLengths();
      offsets = new int[ size_WM ];
      offsets[ 0 ] = 0;
      for( int i = 1; i < size_WM; i++ )
           offsets[ i ] = offsets[ i-1 ] + wmItemLengths[ i-1 ];
      flagItemOns = new boolean[ size_WM ];
      for( int i = 0; i < size_WM; i++ )
           flagItemOns[i] = false;
      items = new Chunk[ size_WM ];
      if( global.isWmFlagsOn() ){
          int flagNum = global.getWmFlagNum();
          for( int i = 0; i < flagNum; i++ ){
               int type = Global.CHK_WM_FLAG;
               String label = Global.CHUNK_TYPE_LABELS[type];
               Feature[] featList = new Feature[1];
               featList[0] = new Feature( "wm flags", 2, Global.FALSE );
               items[i] = new Chunk( type, label, featList, global );
               flagItemOns[i] = true;
          }
      }
      global.setWM( this );
  }

  /** Returns the number of items in this working memory. */
  public int getSize()
  {
      return size_WM;
  }

  /** Returns a specific flag item is on or not.
   *  @param idx the index to the flag item. */
  public boolean getFlagItemOn( int idx )
  {
      return flagItemOns[ idx ];
  }

  /** Returns the working memory items.
   *  @return the working memory items. */
  public Chunk[] getItems()
  {
      Chunk[] wmItems = new Chunk[size_WM];
      try{
         for( int i = 0; i < size_WM; i++ )
              wmItems[i] = (items[i]==null) ? null : (Chunk)(items[i].clone());
      }
      catch(  CloneNotSupportedException e ) {
         System.out.println( "clone error from GKSChunk in WM." );
      }
      return wmItems;
  }

  /** Sets current time.
   *  @param time the time to set. */
  public void setCurrentTime( long time )
  {
      currentTime = time;
  }

/************************ Working Memory actions ******************************/
  /** The process of 'set' action.
   *  @param idx the index to an item.
   *  @param chunk the chunk to set. */
  public boolean actionSet( int idx, Chunk chunk )
  {
      if( flagItemOns[idx] ){
          int type = Global.CHK_WM_FLAG;
          String label = Global.CHUNK_TYPE_LABELS[type];
          Feature[] featList = new Feature[1];
          featList[0] = new Feature( "wm flags", 2, Global.TRUE );
          items[idx] = new Chunk( type, label, featList, global );
          return true;
      }
      if( items[ idx ] != null ){
         if( items[ idx ].equals( chunk ) ){
             if( items[idx].getBlaOption() == Global.VARIED_OPTION )
                 items[ idx ].increaseUsage( currentTime );
             return false;
         }
      }
      try{
             items[ idx ] = (GKSChunk)chunk.clone();
             if( items[idx].getBlaOption() == Global.VARIED_OPTION ){
                 items[ idx ].setBlaParams( blaParams );
                 items[ idx ].resetUsage();
                 items[ idx ].increaseUsage( currentTime );
             }
      }
      catch( CloneNotSupportedException e ){
         System.out.println( "clone error!" );
      }
      return true;
  }

  /** The process of 'reset' action.
   *  @param idx the index to an item. */
  public boolean actionReset( int idx )
  {
      if( items[ idx ] == null )
          return false;
      if( flagItemOns[idx] ){
          int type = Global.CHK_WM_FLAG;
          String label = Global.CHUNK_TYPE_LABELS[type];
          Feature[] featList = new Feature[1];
          featList[ 0 ] = new Feature( "wm flags", 2, Global.FALSE );
          items[ idx ] = new Chunk( type, label, featList, global );
          return true;
      }
      items[ idx ] = null;
      return true;
  }

  /** The process of 'reset all' action.  */
  public boolean actionResetAll()
  {
      boolean retVal = false;
      for( int i = 0; i < size_WM; i++ )
           if( items[ i ] != null ){
               actionReset( i );
               retVal = true;
           }
      return retVal;
  }

  /** The process of 'do-nothing' action.  */
  public boolean actionDoNothing()
  {
     return false;
  }

  /** Updates current WM state as input to WMNet.
   *  @param state current WM state to update. */
  public void updateState( double[] state )
  {
     for( int i = 0; i < size_WM; i++ ){
          if( items[i] == null ){
              for( int j = 0; j < wmItemLengths[i]; j++ )
                   state[ offsets[i] + j ] = 0.0;
          }
          else{
              items[i].calcBLA( currentTime );
              if( items[i].getBLA() < threshold_WM )   // input with strengh 0
                  for( int j = 0; j < wmItemLengths[i]; j++ )
                       state[ offsets[i] + j ] = 0.0;
              else{  // input with strengh 1
                  short[] content = new short[ wmItemLengths[i] ];
                  items[i].getContent( content );
                  for( int j = 0; j < wmItemLengths[i]; j++ )
                       state[ offsets[i] + j ] = content[ j ];
              }
          }
     }
  }
}