package clarion.common;
import java.util.Arrays;
import clarion.system.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the concept of chunk in CLARION model.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void calcBLA( long currentTime ): is to calculate the chunk BLA.
 * input.
 * <li> public void increaseUsage( long currentTime ): is to increase number of
 * chunk usages.
 * <li> public void append( int type, String label, Chunk[] chunks ): is to append
 * an array of chunks into one chunk.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is a base class for concrete chunk class to extend such as the GKSChunk,
 * GoalChunk and so on.
 * <li> Usage of this class. <br>
 * It is instantiated inside a concrete chunk object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class Chunk implements Cloneable {

   /** gloabl settings. */
   protected Global global;
   /** the feature list composing the chunk.  */
   protected Feature[] featList;
   /** the chunk type. */
   protected int    chunkType;
   /** the chunk label. */
   protected String chunkLabel;
   /** the chunk BLA option. */
   protected int    chunkBlaOption;  // constent or varied.
   /** maximum number of usages of a chunk. */
   protected int    MAX_USAGE_NUM = 1000;
   /** the number of usages of a chunk.*/
   protected int    usageNum;
   /** used for recording the times when the chunk was used. */
   protected long   usageRecords[];
   /** the three parameters for calculating BLA: c, d, iBLA. */
   protected double c, d, iBLA;
   /** base level activation. */
   protected double BLA;
   /** chunk strength used for conclusion chunks. */
   protected double strength;

   public Chunk( int type, String label, Feature[] list, Global g )
   {
       chunkType = type;
       chunkLabel = label;
       global = g;
       chunkBlaOption = global.getChunkBlaOption();
       if( chunkBlaOption == global.CONSTANT_OPTION )
           BLA = global.getChunkBla();
       else{
           c = global.getChunkBlaParas()[ global.BLA_C ];
           d = global.getChunkBlaParas()[ global.BLA_D ];
           iBLA = global.getChunkBlaParas()[ global.BLA_I ];
           usageRecords = new long[ MAX_USAGE_NUM ];
       }
       featList = new Feature[ list.length ];
       for( int i = 0; i < list.length; i++ )
            featList[ i ] = (Feature)(list[ i ].clone());
   }

   public Chunk( int type, String label, Chunk[] chunks, Global g )
   {
       chunkType = type;
       chunkLabel = label;
       global = g;
       chunkBlaOption = global.getChunkBlaOption();
       if( chunkBlaOption == global.CONSTANT_OPTION )
           BLA = global.getChunkBla();
       else{
           c = global.getChunkBlaParas()[ global.BLA_C ];
           d = global.getChunkBlaParas()[ global.BLA_D ];
           iBLA = global.getChunkBlaParas()[ global.BLA_I ];
           usageRecords = new long[ MAX_USAGE_NUM ];
       }
       int featLen = 0;
       for( int i = 0; i < chunks.length; i++ )
            featLen += chunks[i].getDimNum();
       featList = new Feature[ featLen ];
       int count = 0;
       for( int i = 0; i < chunks.length; i++ )
         for( int j = 0; j < chunks[i].getDimNum(); j++ )
              featList[ count++ ] = (Feature)(chunks[i].getFeature(j).clone());
   }

   /**
    * appends an array of chunks into one chunk.
    * @param type the chunk type.
    * @param label the chunk label.
    * @param chunks the array of chunks to form a new chunk.
    */
   public void append( int type, String label, Chunk[] chunks )
   {
       chunkType = type;
       chunkLabel = label;
       chunkBlaOption = global.getChunkBlaOption();
       if( chunkBlaOption == global.CONSTANT_OPTION )
           BLA = global.getChunkBla();
       else{
           c = global.getChunkBlaParas()[ global.BLA_C ];
           d = global.getChunkBlaParas()[ global.BLA_D ];
           iBLA = global.getChunkBlaParas()[ global.BLA_I ];
           usageRecords = new long[ MAX_USAGE_NUM ];
       }
       Feature[] list = new Feature[ featList.length ];
       for( int i = 0; i < featList.length; i++ )
            list[ i ] = (Feature)(featList[ i ].clone());
       int featLen = featList.length;
       for( int i = 0; i < chunks.length; i++ )
            featLen += chunks[i].getDimNum();
       featList = new Feature[ featLen ];
       int count = 0;
       for( int i = 0; i < list.length; i++ )
            featList[ count++ ] = (Feature)(list[ i ].clone());
       for( int i = 0; i < chunks.length; i++ )
         for( int j = 0; j < chunks[i].getDimNum(); j++ )
              featList[ count++ ] = (Feature)(chunks[i].getFeature(j).clone());
   }

   /** reinit the chunk */
   public void reinit()
   {
       resetUsage();
   }

    /**
     * returns the chunk type.
     * @return the chunk type. */
   public int getChunkType()
   {
       return chunkType;
   }

    /**
     * returns the number of dimesnions in this chunk.
     * @return the number of dimesnions in this chunk. */
   public int getDimNum()
   {
       return featList.length;
   }

    /**
     * returns the chunk label.
     * @return the chunk label. */
   public String getLabel( )
   {
       return chunkLabel;
   }

    /**
     * returns the chunk length in term of the sum of the lengths of the features in list.
     * @return the chunk length in term of the sum of the lengths of the features in list. */
   public int getLength()
   {
      int length = 0;
      for( int i = 0; i < featList.length; i++ )
           length += featList[i].getLength();
      return length;
   }

    /**
     * returns the chunk BLA option: constant or variable.
     * @return the chunk BLA option. */
   public int getBlaOption()
   {
       return chunkBlaOption;
   }

    /**
     * returns the number of usages of this chunk.
     * @return the number of usages of this chunk. */
   public int getUsageNum()
   {
       return usageNum;
   }

    /**
     * returns the info of usages of this chunk.
     * @return the info of usages of this chunk. */
   public long[] getUsageRecords()
   {
       return usageRecords;
   }

    /**
     * return the info of usages of this chunk.
     * @param arr the array to store the usage info. */
   public void getUsageRecords( long[] arr )
   {
       for( int i = 0; i < arr.length; i++ )
            arr[ i ] = usageRecords[ i ];
   }

    /**
     * returns the chunk BLA .
     * @return the chunk BLA. */
   public double getBLA()
   {
       return BLA;
   }

    /**
     * returns the chunk strength.
     * @return the chunk strength. */
   public double getStrength()
   {
       return strength;
   }

    /**
     * returns the number of feasures in this chunk.
     * @return the number of feasures in this chunk. */
   public int getFeatureNum()
   {
       return featList.length;
   }

    /**
     * returns the list of feasures in this chunk.
     * @return the list of feasures in this chunk. */
   public Feature[] getFeatures()
   {
       return featList;
   }

    /**
     * returns the list of feasures in this chunk.
     * @param other the feature array to store the feature list of this chunk. */
   public void getFeatures( Feature[] other )
   {
       other = new Feature[ featList.length ];
       for( int i = 0; i < featList.length; i++ )
            other[ i ] = (Feature)(featList[ i ].clone());
   }

    /**
     * returns a specified feasure in this chunk.
     * @return a specified feasure in this chunk. */
   public Feature getFeature( int idx )
   {
       return (Feature)(featList[ idx ].clone());
   }

    /**
     * returns the content of this chunk in term of active values.
     * @param arr the array to store the content of this chunk. */
   public void getContent( short[] arr )
   {
      int offset = 0;
      for( int i = 0; i < featList.length; i++ ){
           short[] content = new short[ featList[i].getLength() ];
           featList[i].getContent( content );
           for( int j = 0; j < content.length; j++ )
                arr[ offset + j ] = content[ j ];
           offset += content.length;
      }
   }

    /**
     * returns the content of this chunk in term of activations.
     * @param arr the array to store the content of this chunk. */
   public void getContent( double[] arr )
   {
      int offset = 0;
      for( int i = 0; i < featList.length; i++ ){
           short[] content = new short[ featList[i].getLength() ];
           featList[i].getContent( content );
           for( int j = 0; j < content.length; j++ )
                arr[ offset + j ] = content[ j ];
           offset += content.length;
      }
   }

    /**
     * set chunk label.
     * @param str the label to set. */
   public void setLabel( String str )
   {
       chunkLabel = str;
   }

    /**
     * set chunk BLA option.
     * @param option the option to set. */
   public void setBlaOption( int option )
   {
       chunkBlaOption = option;
   }

    /**
     * set chunk BLA. used for constent BLA option.
     * @param bla the BLA to set. */
   public void setBla( double bla )
   {
      BLA = bla;
   }

    /**
     * set chunk BLA parameter. used for variable BLA option.
     * @param val the value to set. */
   public void setBLA_C( double val )
   {
       c = val;
   }

    /**
     * set chunk BLA parameter. used for variable BLA option.
     * @param val the value to set. */
   public void setBLA_D( double val )
   {
       d = val;
   }

    /**
     * set chunk BLA parameter. used for variable BLA option.
     * @param val the value to set. */
   public void setBLA_INIT( double val )
   {
       iBLA = val;
   }

    /**
     * set chunk BLA parameters. used for variable BLA option.
     * @param blaParams the parameters to set. */
   public void setBlaParams( double[] blaParams )
   {
       c = blaParams[ global.BLA_C ];
       d = blaParams[ global.BLA_D ];
       iBLA = blaParams[ global.BLA_I ];
   }

    /**
     * set number of chunk usages.
     * @param num the value to set. */
   public void setUsageNum( int num )
   {
       usageNum = (short)num;
   }

    /**
     * set the info of chunk usages.
     * @param records the value to set. */
   public void setUsageRecords( long[] records )
   {
       for( int i = 0; i < usageRecords.length; i++ )
            usageRecords[ i ] = records[ i ];
   }

    /**
     * set chunk strength.
     * @param val the value to set. */
   public void setStrength( double val )
   {
       strength = val;
   }

    /**
     * increase number of chunk usages.
     * @param currentTime current time to record as one time of this chunk is used. */
   public void increaseUsage( long currentTime )
   {
       if( usageNum > 0 && currentTime == usageRecords[ usageNum-1 ] ){
           //System.out.println("usage has already update in Chunk.java");
           return;
       }
       if( usageNum < MAX_USAGE_NUM-1 )
           usageRecords[ usageNum++ ] = currentTime;
       else
           System.out.println("error usage update in Chunk.java");
   }

    /**
     * reset the usage of this chunk. */
   public void resetUsage()
   {
       usageNum = 0;
       for( int i = 0; i < usageRecords.length; i++ )
            usageRecords[ i ] = 0;
   }

    /**
     * calculate chunk BLA. */
   public void calcBLA( long currentTime )
   {
       if( chunkBlaOption == global.CONSTANT_OPTION )
           return;
       double val = iBLA;
       for( int j = 0; j < usageNum; j++ ){
            long timeLaps = currentTime - usageRecords[j];
            if( timeLaps <= 0 ){
                System.out.println( "error usage recording in Chunk.java" );
                break;
            }
            val += c * Math.pow( timeLaps, -d );
       }
       BLA = val;
       if( BLA < 0.01 )
           BLA = 0.01;
   }

   /**
    * search a specified feature labeled <em>label</em> in this chunk.
    * @param label the label of the feasure to search.
    */
   public int searchFeature( String label )
   {
       if( label == "" )
           return -1;

       int idx = -1;
       for( int i = 0; i < featList.length; i++ )
            if( label.compareTo( featList[ i ].getLabel() ) == 0 )
            {
                idx = i;
                break;
            }

      return idx;
   }

   /**
    * compare two feature list is equal or not.
    */
   public boolean isEqual( Feature[] features )
   {
       if( featList.length != features.length )
           return false;
       for( int i = 0; i < featList.length; i++ ) {
            if( !featList[ i ].equals( features[ i ] ) )
                return false;
       }
       return true;
   }

   /** methods from class <em>Object</em> */
    public String toString()
    {
       String str = Global.TABLE_STR + "Chunk = \n";
       String prefix = Global.TABLE_STR + Global.TABLE_STR;
       str += prefix + "feature list = ";
       for( int i = 0; i < featList.length; i++ )
            //str += featList[ i ] + "\n";
            str += featList[ i ] + ", ";
       str += "\n";
       str += prefix + "strength = " + String.valueOf( strength ) + "\n";
       return str;
    }

   /** methods from class <em>Object</em> */
    public int hashCode()
    {
       int sum = 0;
       for( int i = 0; i < featList.length; i++ )
            sum += featList[ i ].hashCode();

       return sum;
    }

   /** methods from class <em>Object</em> */
    public boolean equals( Object other )
    {
       if( other == null || !(other instanceof Chunk ) )
           return false;
       Chunk otherChunk = (Chunk)other;
       if( chunkType != otherChunk.getChunkType() )
           return false;
       Feature[] otherFeatures = otherChunk.getFeatures();
       return isEqual( otherFeatures );
    }

    /** methods from class <em>Object</em> */
    public Object clone() throws CloneNotSupportedException
    {
       Chunk chunk = new Chunk( chunkType, chunkLabel, featList, global );
       chunk.setStrength( strength );
       chunk.setBlaOption( chunkBlaOption );
       if( chunkBlaOption == global.CONSTANT_OPTION )
           chunk.setBla( BLA );
       else{
           chunk.setBLA_C( c );
           chunk.setBLA_D( d );
           chunk.setBLA_INIT( iBLA );
           chunk.setUsageNum( usageNum );
           chunk.setUsageRecords( usageRecords );
       }
       return chunk;
    }
}