package clarion.common;
import clarion.system.*;
import clarion.acs.net.*;
import clarion.acs.*;
import clarion.tools.*;
import clarion.coordinate.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is the base class to represent a general action-decision network.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void decideAction( int compIdx ): decide an action.
 * <li> public void reasoning(): the reasoning routine defined in this network.
 * <li> public void update( double reward ): update the network, needs overriding.
 * <li> public void formState(): compose current state given external, GS and WM states.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class is on the top in the structure of action-decision network which contains
 * two levels : the bottom level (BP network) and the top level (rule network).
 * <li> Usage of this class. <br>
 * It can be extended by the concrete action-decision networks of ACS or MCS for
 * specific purpose. And it is instantiated inside a concrete action-decision network
 * object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class GeneralNet{

  /** gloabl settings. */
  protected Global global;
  /** the user-defined CLARION settings. */
  protected TaskClarion taskClarion;
  /** subsystem: ACS or MCS. */
  protected int subsysType;
  /** the network indice in ACS or MCS. */
  protected int netIdx;
  /**  number of knowledge types. */
  protected int COMP_NUM;
  /** current chosen component. */
  protected int chosenComp;
  /** current chosen action,
   *  the first dim:  NORMAL or NACS-CONTROL action.
   *  the second dim: all of the action dimensions, ( # of bottom output dims ).
   *  the third indicates the active values inside that dimension.
   *  the actual action is a dim-value pairs. */
  protected short[][][] chosenFullAction;
  /** the numbers of dimensions of input and output. */
  protected int fullOutputDimNum, fullOutputNum;
  /** the unit numbers of total input, external input, goal structure input and
   *  working memory input. */
  protected int inputNum, exInputNum, gsInputNum, wmInputNum;
  /** current states. state = EX + GS + WM. */
  protected double[] state, exState, gsState, wmState;
  /** current sensory input. */
  protected Feature sensoryInput[];
  /** flags to indicate knowledge type is used or not. */
  protected boolean  netCompFlags[];
  /** the knowledge components composing this network. */
  protected GeneralNetComp netComps[];
  /** the action strengths supported from all of the components.
   *  the first dim: network components.
   *  the second dim: ACTION_NORMAL OR ACTION_CONTROL.
   *  the third dim: the activations. full length of the outputs (bottom output length). */
  protected double[][][] outputActivations;
  /** arrays of VCGSet to store the active values in each dimension from each
   *  net component such as IDN, RER, IRL and FR.
   *  the first dim: network components.
   *  the second dim: ACTION_NORMAL OR ACTION_CONTROL.
   *  the third dim: VCGSet for each dim. full # of the output dims (# of bottom output dims). */
  protected VCGSet compVcgSets[][][];
  /** arrays of VCGSet to store the active values in each NORMAL action dimension
   *  from each net component such as IDN, RER, IRL and FR.
   *  the first dim: network components.
   *  the second dim: VCGSet for each dim. full # of the output dims (# of bottom output dims). */
  protected VCGSet normalVcgSets[][];
  /** arrays of VCGSet to store the active values in each NACS-CONTROL action dimension
   *  from each net component such as IDN, RER, IRL and FR.
   *  the first dim: network components.
   *  the second dim: VCGSet for each dim. full # of the output dims (# of bottom output dims). */
  protected VCGSet controlVcgSets[][];
  /** current reinforcement. */
  protected double     reinforcement;
  /** the mode of integrating outputs from different componnents: fixed or variable.  */
  protected int  combMode1;
  /** the mode of integrating outputs from different componnents: stochastic or
   *  percentage. */
  protected int  combMode2;
  /** success rate of each component, for variable mode.  */
  protected double[] succRates;
  /** used for variable mode. */
  protected double[] c3s, c4s;
  /** Stochastic Combination. */
  protected StochasticComb stochComb;
  /** Weight Sum Combination. */
  protected WeightSumComb wsComb;
  /** correction Combination. */
  protected CorrectionComb ctComb;

  public GeneralNet( int subsysType )
  {
     this.subsysType = subsysType;
  }

  /**
   * returns the external state.
   * @param arr the array to store the external state. */
  public void getExState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          arr[i] = exState[i];
  }

  /**
   * returns the goal structure state.
   * @param arr the array to store the goal structure state. */
  public void getGsState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          arr[i] = gsState[i];
  }

  /**
   * returns the working memory state.
   * @param arr the array to store the working memory state. */
  public void getWmState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          arr[i] = wmState[i];
  }

  /**
   * returns the RER rule set in this network in terms of its representation.
   * @return the RER rule set. */
  public String[] getRerRules()
  {
     if( netComps[global.RER] != null )
         return ((RerRuleSet)netComps[ global.RER ]).getRules();
     else
         return (String[])null;
  }

  /**
   * returns the IRL rule set in this network in terms of its representation.
   * @return the IRL rule set. */
  public String[] getIrlRules()
  {
     if( netComps[global.IRL] != null )
         return ((IrlRuleSet)netComps[ global.IRL ]).getRules();
     else
        return (String[])null;
  }

  /**
   * returns the Fixed rule set in this network in terms of its representation.
   * @return the Fixed rule set. */
  public String[] getFixRules()
  {
     if( netComps[global.FR] != null )
         return ((FixRuleSet)netComps[ global.FR ]).getRules();
     else
         return (String[])null;
  }

  /**
   * returns the chosen action.
   * @return the chosen action. */
  public short[][][] getChosenFullAction()
  {
     return CommonMethods.copy( chosenFullAction );
  }

  /**
   * returns the number of positive match.
   * @return the number of positive match. */
  public double getPM( int idx )
  {
     if( netCompFlags[ idx ] )
         return netComps[ idx ].getPM();
     else
         return 0.0;
  }

  /**
   * returns the number of negative match.
   * @return the number of negative match. */
  public double getNM( int idx )
  {
     if( netCompFlags[ idx ] )
         return netComps[ idx ].getNM();
     else
         return 0.0;
  }

  /**
   * returns the response time in case of no NACS RT involved.
   * @return the response time. */
  public int getRT()
  {
     int RT = 0;
     int RT_BL = 0;
     int RT_TL = 0;
     if( combMode1 == Global.STOCH_COMB_MODE ){
         RT = netComps[ chosenComp ].getResponseTime( 0 );
         if( chosenComp == Global.BL )
             RT_BL = RT;
         else
             RT_TL = RT;
     }
     else{
        for( int i = 1; i < COMP_NUM; i++ )
             if( netCompFlags[i] )
                 RT_TL = Math.max(RT_TL, netComps[i].getResponseTime( 0 ));
        RT_BL = netComps[Global.BL].getResponseTime( 0 );
     }
     return taskClarion.getOverallRT( RT_BL, RT_TL );
  }

  /**
   * returns the response time in case of NACS RT involved.
   * @return the response time. */
  public int getRT( int nacsRT )
  {
     int RT = 0;
     int RT_BL = 0;
     int RT_TL = 0;
     if( combMode1 == Global.STOCH_COMB_MODE ){
         RT = netComps[ chosenComp ].getResponseTime( nacsRT );
         if( chosenComp == Global.BL )
             RT_BL = RT;
         else
             RT_TL = RT;
     }
     else{
        for( int i = 1; i < COMP_NUM; i++ )
             if( netCompFlags[i] )
                 RT_TL = Math.max(RT_TL, netComps[i].getResponseTime( nacsRT ));
        RT_BL = netComps[Global.BL].getResponseTime( nacsRT );
     }
     return taskClarion.getOverallRT( RT_BL, RT_TL );
  }

  /**
   * divide the chunk into EX, GS and WM states if exists. maybe revised further.
   * @param chunk the chunk to decompose. */
  public void setStates( Chunk chunk )
  {
     int dimNum = chunk.getDimNum();
     int offset = 0;
     for( int i = 0; i < dimNum; i++ ) {
          short[] arr = chunk.getFeature(i).getContent();
          for( int j = 0; j < arr.length; j++ )
               state[ offset + j ] = arr[ j ];
          offset += arr.length;
     }
  }

  /**
   * set current sensory input.
   * @param sensoryInput the senory input to set. */
  public void setSensoryInput( Feature[] sensoryInput )
  {
     this.sensoryInput = sensoryInput;
  }

  /**
   * set current goal structure state.
   * @param arr the goal structure state to set. */
  public void setGsState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          gsState[i] = arr[i];
  }

  /**
   * set current working memory state.
   * @param arr the working memory state to set. */
  public void setWmState( double[] arr )
  {
     for( int i = 0; i < arr.length; i++ )
          wmState[i] = arr[i];
  }

  /**
   * set current desired output.
   * @param desiredOutput the desired output to set. */
  public void setCurDesiredOutput( Feature[] desiredOutput )
  {
     ((IDN)netComps[ Global.BL ]).setCurDesiredOutput( desiredOutput );
  }

  /**
   * set current time.
   * @param curTime the time to set. */
  public void setCurrentTime( long curTime )
  {
     for( int i = 0; i < netCompFlags.length; i++ )
          if( netCompFlags[ i ] )
              netComps[ i ].setCurrentTime( (int)curTime );
  }

  /**
   * set reinforcement.
   * @param rwward the reinforcement to set. */
  public void setReinforcement( double reward )
  {
     reinforcement = reward;
  }

  /** compose current state given external, GS and WM states. needs overriding. */
  public void formState()
  {}

  /** Decides action by specific knowledge type. this means stochastic combination
   *  used.
   *  @param compIdx the specific knowledge type. */
  public void decideAction( int compIdx )
  {
     // decide an action.
     chosenComp = compIdx;
     if( netCompFlags[ chosenComp ] )
         chosenFullAction = netComps[ chosenComp ].decideAction();
     else
         chosenFullAction = null;
     if( !global.isOverrideByIdn() )  // added on 06/23/03
         return;
     // if the chosen action is empty, let IDN determine the action.
     if( chosenFullAction == null ){
         chosenComp = global.BL;
         chosenFullAction = netComps[ chosenComp ].decideAction();
     }

     //if( chosenComp == 0 )
     //    System.out.println( "In GeneralNet.java, ACS chosen comp at each step = " + chosenComp);

     // the chosen action is still empty, randomly select an action.
  }

  /** reasoning for determining an action to perform given current input. */
  public void reasoning()
  {
     // get active net comps.
     //netCompFlags[Global.BL] = global.isIdnOn( netIdx );
     netCompFlags[Global.BL] = true;
     netCompFlags[Global.RER] = global.isRerOn( netIdx );
     netCompFlags[Global.IRL] = global.isIrlOn( netIdx );
     netCompFlags[Global.FR] = global.isFrOn( netIdx );
     for( int i = 0; i < COMP_NUM; i++ ){
          if( netCompFlags[ i ] ){
              netComps[ i ].setStates( state, exState, gsState, wmState );
              netComps[ i ].reasoning();
          }
          if( combMode1 > Global.STOCH_COMB_MODE && combMode2 < Global.CT_COMB_MODE ){
              // if it is weight-sum combination.
              if( netCompFlags[ i ] )
                  netComps[ i ].getOutputActivations( outputActivations[ i ] );
              else
                  for( int j = 0; j < 2; j++ ){
                       if( outputActivations[ i ][ j ] != null )
                            for( int k = 0; k < outputActivations[ i ][ j ].length; k++ )
                                 outputActivations[ i ][ j ][ k ] = 0.0;
                  }
          }
     }
     // combine the results.
     if( combMode1 == Global.STOCH_COMB_MODE )
         stochCombine();
     else{ // bottom-up verification or top-guide
         if( combMode2 < Global.CT_COMB_MODE )
             weightSumCombine();
         else
             correctionCombine();
     }
     // set the chosen action.
     deliverAction();
  }

  /** update the network, needs overriding.*/
  public void update( double reward )
  {
  }

  /** fill in the VCGSets for action decision making. */
  public void fillInVcgSets( int fillType )
  {
     for( int i = 0; i < COMP_NUM; i++ ){
       int offset = 0;
       for( int j = 0; j < compVcgSets[i][fillType].length; i++ ){
            int length = compVcgSets[i][fillType][j].getFullLength();
            double[] activations = new double[ length ];
            System.arraycopy( outputActivations[i][fillType], offset, activations, 0, length );
            compVcgSets[i][fillType][j].setActivations( activations );
            offset += length;
       }
     }
  }

  /** deliver currently chosen action to each component. */
  protected void deliverAction()
  {
     for( int i = 0; i < COMP_NUM; i++ ){
          if( netCompFlags[ i ] )
              netComps[ i ].setChosenAction( chosenFullAction );
     }
  }

  /** stochastic combination method. */
  protected void stochCombine()
  {
      // percentage combination, only one set of measuses needed.
      if( combMode2 == global.VAR_COMB_MODE ) {
          double[] PMs = new double[ COMP_NUM ];
          double[] NMs = new double[ COMP_NUM ];
          for( int i = 0; i < COMP_NUM; i++ ) {
               PMs[ i ] = getPM( i );
               NMs[ i ] = getNM( i );
               succRates[i] = (c3s[i] + PMs[i]) / (c4s[i] + PMs[i] + NMs[i]);
          }
          stochComb.setSuccRate( succRates );
          stochComb.calcProbabilities();
      }
      // select which part: BL, RER, IRL or FR to decide action.
      chosenComp = stochComb.selectComponent();

      //System.out.println( "the chosen comp = " + chosenComp );

      // decide an action.
      decideAction( chosenComp );
   }

  /** weight sum combination method. */
   protected void weightSumCombine()
   {
      if( combMode2 == global.VAR_COMB_MODE ){
          // calculate PM and NM.
          double[] PMs = new double[ COMP_NUM ];
          double[] NMs = new double[ COMP_NUM ];
          for( int i = 0; i < COMP_NUM; i++ ) {
               PMs[ i ] = getPM( i );
               NMs[ i ] = getNM( i );
               succRates[i] = (c3s[i] + PMs[i])/(c4s[i] + PMs[i] + NMs[i]);
          }
          wsComb.setSuccRate( succRates );
          wsComb.calcWeights();
      }
      // decide action.
      /// fill in the VCGSets for action decision making. */
      fillInVcgSets( Global.ACTION_NORMAL );
      wsComb.setVCGSets( normalVcgSets );
      chosenFullAction[Global.ACTION_NORMAL] = wsComb.combination();
  }

  /** the correction combination method should be overrided by user's code. */
  protected void correctionCombine()
  {
     //ctComb.correctionComb();
     if( combMode1 == Global.BU_COMB_MODE )
         taskClarion.combineByBottomUpRectification();
     if( combMode1 == Global.TG_COMB_MODE )
         taskClarion.combineByTopDownGuidance();
  }
}