// implements Q learning over a backprop neural net.
// input is a state, and output is Q values of each action
package clarion.common;
import clarion.system.*;
import clarion.tools.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the Q-Learning algorithm using a backpropagation network.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void update( short[] action, double feedback, double[] qValue ):
 * update routine used for simplified Q-Learning.
 * <li> public void update( short[] action, double feedback, double[] qValue1, double[] qVaule2 ):
 * update routine used for Q-Learning.
 * <li> public void backwardPass( double[] qValue1, double[] qValue2 ): update all of the
 * weights inside the neural network (Q-Learning version).
 * <li> public void backwardPass( double[] qValue ): update all of the  weights inside the
 * neural network (simplified Q-Learning version).
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class extends the base class <em>BPNet</em> and implements Q-Learning algorithm
 * which is the major learning algorithm in the implicit process in Action-Centered
 * network.
 * <li> Usage of this class. <br>
 * It is instantiated inside the class <em>IDN</em> implementing the implicit decision
 * network, the bottom level in Action-Centered network.
 * </ul>
 * @author Xi Zhang
 *
 */

public class QBPNet extends BPNet {

  /** subsystem: ACS or MCS. */
  protected int subsysIdx;
  /** network type : EX, GS or WM. */
  protected int netType;
  /** the net index */
  protected int netIdx;
  /* # of output dims  */
  protected int OUTPUT_DIM_NUM;
  /** the # of NACS-control dims. */
  protected int CONTROL_OUTPUT_DIM_NUM;
  /** D-V info of bottom (IDN) outputs. */
  protected short[] fullOutputDVs;
  /** start position of each (bottom) output dim in one dimensional aray. */
  protected short[] fullOutputOffsets;
  /** the # of each type of output:  EX, WM, GS, NACS-CONTROL if any. */
  /** the dim-val info of NACS-control outputs. */
  protected short[] fullControlOutputDVs;
  /** the start offsets of each NACS-control dim in an one-dimenisonal array. */
  protected short[] fullControlOutputOffsets;
  /** #s of each type of output in its physical full length. */
  protected short[] outputNums;
  /** used to locate each type of output in the overall outputs.
   *  EX, WM, GS, NACS-CONTROL if any. */
  protected short[] outputLocations;
  /** action performed. */
  protected short  performedAction[][][];
  /** the discount and reinforcement. */
  protected double discount, reinforcement;
  /** current state and previous state */
  protected double state[], preState[];

  /** constructor.
   *  @param inputNum the number of input unit.
   *  @param nIdx the net index.
   *  @g the associative global settings.
   */
  public QBPNet( int inputNum, int nIdx, Global g )
  {
      super( inputNum, nIdx, g, Global.ACS );
      netIdx = nIdx;
      global = g;
      discount = global.getIdnQDiscountRate( netIdx );
      state = new double[ nInput ];
      if( global.getIdnLearningOption( netIdx ) == global.QL )
          preState = new double[ nInput ];

      if( subsysIdx == Global.ACS ){
          netType = global.isExNet(netIdx) ? Global.EX_TYPE :(global.isWmNet(netIdx) ? Global.WM_TYPE : Global.GS_TYPE );
          OUTPUT_DIM_NUM = global.getIdnOutputDimNum( netIdx );
          CONTROL_OUTPUT_DIM_NUM = Global.ACT_DIM_NUM_NACS_CONTROL;
          fullControlOutputDVs = global.getNacsControlPhysicOutputDimVals();
          fullControlOutputOffsets = new short[CONTROL_OUTPUT_DIM_NUM];
          CommonMethods.calcOffsets( fullControlOutputOffsets, fullControlOutputDVs );
          fullOutputDVs = global.getAcsPhysicOutputDimVals( netIdx );
          fullOutputOffsets = new short[fullOutputDVs.length];
          CommonMethods.calcOffsets( fullOutputOffsets, fullOutputDVs );
          // initialize variables for locations of each type of output in overall outputs.
          outputNums = global.getAcsPhysicOutputNums();
          outputLocations = new short[ outputNums.length ];
          CommonMethods.calcOffsets( outputLocations, outputNums );
      }
      else{  // MCS.
      }
  }

  /** reinitialize this network. */
  public void reinit()
  {
      resetMatches();
      reinitWeights();
  }

 /**
  * returns current state.
  * @param fillMe the array to store current state. */
  public void getState( int[] fillMe )
  {
      for(int i = 0; i < nInput; i++ )
           fillMe[i] = (int) state[i];
  }

 /**
  * returns the discount.
  * @return the discount. */
  public double getQDiscount()
  {
      return discount;
  }

 /**
  * returns the reinforcement.
  * @return the reinforcement. */
  public double getReinforcement()
  {
      return reinforcement;
  }

 /**
  * returns the Q-value of a particular output unit.
  * @param index index on the particular output unit.
  * @return the Q-value of a particular output unit. */
  public double getQVal( int index )
  {
     return Output[ index ];
  }

 /**
  * returns all of the Q-values (the outputs).
  * @param fillMe the array to store the outputs. */
  public void getQVals( double[] fillMe )
  {
      for (int i = 0; i < nOutput; i++ )
            fillMe[i] = Output[i];
  }

 /** Returns the max Q-values of each output dimension.
  *  @param fillMe the array to store the maximal Q-values. */
  public void getMaxQVals( double[] fillMe )
  {
  }

 /**
  * sets current state using an array of active dimensional values.
  * @param curState the array used to set current state. */
  public void setState( int[] curState )
  {
      if( preState != null )  // normal Q-Learning.
      {
          for( int i = 0; i < nInput; i++ )
               preState[ i ] = state[ i ];
      }
      for( int i = 0; i < nInput; i++ )
           state[ i ] = curState[ i ];

      setInput();
  }

 /**
  * sets current state using an array of activations of all of the dimensional values.
  * @param curState the array used to set current state. */
  public void setState( double[] curState )
  {
      if( preState != null )  // normal Q-Learning.
      {
          for( int i = 0; i < nInput; i++ )
               preState[ i ] = state[ i ];
      }
      for( int i = 0; i < nInput; i++ )
           state[ i ] = curState[ i ];

      setInput();
  }

  /** Sets the previous state as input to BPNet. used for normal Q-Learning.*/
  public void setPreState()
  {
      for( int i = 0; i < nInput; i++ )
           state[ i ] = preState[ i ];

      setInput();
  }

  /** Sets the state as input to BPNet. used for normal Q-Learning.*/
  public void setInput()
  {
      setInput( state );
  }

 /**
  * sets currently received reinforcement.
  * @param value the value used to set reinforcement. */
  public void setReinforcement( double value )
  {
      reinforcement = value;
  }

  /** Sets the performed action.
   *  @param action the performed action to set. */
  public void setPerformedAction( short[][][] action )
  {
      performedAction = CommonMethods.copy( action );
  }

  /** update routine used for simplified Q-Learning.
   *  @param action the performed action.
   *  @param feedback the currently received reinforcement.
   *  @param qValue the array of Q-values the action has. */
  public void update( short[][][] action, double feedback, double[][][] qValue )
  {
      //let QBP know what action was performed
      setPerformedAction( action );
      // update the reinforcement
      setReinforcement( feedback );
      // update qbp net
      backwardPass( qValue );
  }

  /** update routine used for Q-Learning.
   *  @param action the performed action.
   *  @param feedback the currently received reinforcement.
   *  @param qVal1 the array of maximal Q-values after the action is performed.
   *  @param qVal2 the array of Q-values the action has before the action is
   *  performed. */
  public void update( short[][][] action, double feedback, double[][][] qVal1, double[][][] qVal2 )
  {
      //let QBP know what action was performed
      setPerformedAction( action );
      //update the reinforcement
      setReinforcement( feedback );
      // update qbp net
      setPreState();
      forwardPass();
      backwardPass( qVal1, qVal2 );
  }

 /** Updates the neural network using QBP version.
   *  @param qVal1 the array of maximal Q-values after the action is performed.
   *  @param qVal2 the array of Q-values the action has before the action is
   *  performed. */
  public void backwardPass( double[][][] qValue1, double[][][] qValue2 )
  {
      computeErrors( qValue1, qValue2 );
      computeHiddenErrors();
      modifyHiddenToOutput();
      modifyInputToHidden();
  }

  /** Updates the neural network using simplified QBP version.
   *  @param action the performed action.
   *  @param feedback the currently received reinforcement.
   *  @param qValue the array of Q-values the action has. */
  public void backwardPass( double[][][] qValue )
  {
      computeErrors( qValue );
      computeHiddenErrors();
      modifyHiddenToOutput();
      modifyInputToHidden();
  }

/************************** PROTECTED METHODS *********************************/
  /** computes the errors of output layer of QBP network.
   *  @param qVal1 the array of maximal Q-values after the action is performed.
   *  @param qVal2 the array of Q-values the action has before the action is
   *  performed. */
  protected void computeErrors( double[][][] qValue1, double[][][] qValue2 )
  {
     computeErrors( netType, performedAction[Global.ACTION_NORMAL],
                    qValue1[Global.ACTION_NORMAL], qValue1[Global.ACTION_NORMAL] );
     if( subsysIdx == Global.ACS && global.isNacsOn() )
         computeErrors( Global.CONTROL_TYPE, performedAction[Global.ACTION_CONTROL],
                       qValue1[Global.ACTION_CONTROL], qValue1[Global.ACTION_CONTROL] );
  }

  /** computes the errors of output layer of simplified QBP network.
   *  @param qValue the array of Q-values the action has. */
  protected void computeErrors( double[][][] qValue )
  {
     computeErrors( netType, performedAction[Global.ACTION_NORMAL],
                    qValue[Global.ACTION_NORMAL], null );
     if( subsysIdx == Global.ACS && global.isNacsOn() )
         computeErrors( Global.CONTROL_TYPE, performedAction[Global.ACTION_CONTROL],
                        qValue[Global.ACTION_CONTROL], null );
  }

 /** outputType: EX_TYPE, WM_TYPE, GS_TYPE or CONTROL_TYPE. */
  private void computeErrors( int outputType, short[][] action, double[][] qValue1, double[][] qValue2 )
  {
     int dimNum = (outputType != Global.CONTROL_TYPE) ? OUTPUT_DIM_NUM : CONTROL_OUTPUT_DIM_NUM;
     short[] oDVs = (outputType != Global.CONTROL_TYPE) ? fullOutputDVs  : fullControlOutputDVs;
     short[] offsets = (outputType != Global.CONTROL_TYPE) ? fullOutputOffsets  : fullControlOutputOffsets;
     short[][] dimInfo = (outputType != Global.CONTROL_TYPE) ? global.getIdnOutputFeatureInfo( netIdx ) :
                          global.getNacsControlFeatureInfo();

     for( int i = 0; i < dimNum; i++ ){
          int idx = dimInfo[i][1];
          if( subsysIdx == Global.ACS && outputType != Global.EX_TYPE )
              idx = i;
          if( global.getOutputMultiValueProperty( outputType )[idx] == Global.SINGLE_ACTIVE_VALUED ){
              for( int j = 0; j < qValue1[i].length; j++ ){
                   int offset = outputLocations[outputType] + offsets[idx] + ((action[i]==null)? 0 : action[i][0]);
                   if( j == 1 )// SINGLE_ACTIVE_VALUED.
                       offset = outputLocations[outputType] + offsets[idx] + oDVs[idx] - 2;
                   double err = ( qValue2 != null )? reinforcement + discount * qValue2[i][j] - qValue1[i][j]:
                                reinforcement - qValue1[i][j];
                   Errors[ offset ] = OutputDeriv[ offset ] * err;
              }
          }
          else{
              int length = qValue1[i].length;
              short[] flags = new short[length-1];
              CommonMethods.resetArray( flags, 0 );
              if( action[i] != null )
                  for( int j = 0; j < action[i].length; j++ )
                       flags[action[i][j]] = 1;
              for( int j = 0; j < length-1; j++ ){
                   int offset = outputLocations[outputType] + offsets[idx] + 2*j;
                   if( flags[j] == 1 )
                       offset += Global.ELEM_ACTIVE;
                   double err = ( qValue2 != null )? reinforcement + discount * qValue2[i][j] - qValue1[i][j]:
                                reinforcement - qValue1[i][j];
                   Errors[ offset ] = OutputDeriv[ offset ] * err;
              }
              int offset = outputLocations[outputType] + offsets[idx] + 2*length-2;
              if( action[i] != null && action[i].length > 1 )
                  offset += Global.MULTI_ACTIVE_VALUED;
              double err = ( qValue2 != null )? reinforcement + discount * qValue2[i][length-1] - qValue1[i][length-1]:
                           reinforcement - qValue1[i][length-1];
              Errors[ offset ] = OutputDeriv[ offset ] * err;
          }
     }
  }
}