package clarion.common;
import clarion.system.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to impliment the rule attributes associated with a rule such as
 * the BLA, utility and response time.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void calcUtility( double cost ): calcluate the utility.
 * <li> public void calcBLA( int curTime ): calcluate the BLA.
 * <li> public void calcTLRT( int TL_AT ): calculate the response time given the
 * top-level actuation time.
 * <li> public void updateUsages( int curTime ): update the usage info.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the measures relevant to rule knowledge type.
 * <li> Usage of this class. <br>
 * It is instantiated either inside its associated concrete rule object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class RuleAttributes {

   /** global settings. */
   protected Global global;
   /** task-specific clarion settings. */
   protected TaskClarion taskClarion;
   /** ACS rule or NACS rule. */
   protected boolean acsRuleFlag;
   /** the net index. */
   protected int netIdx;
   /** the rule type. */
   protected int ruleType;
   /** the rule set the rule belongs to. */
   protected int ruleSet;
   /** age from last application of this rule. */
   protected int age;
   /** RT variables. */
   protected int TLRT, TLPT, TLDT, actionTime;
   /** associative application time. */
   protected int assocAppTime;
   /** parameters for chunk retrieval time. */
   protected double t0, t1, chunkBLA;
   /** base-level activation option: constant or variable. */
   protected int blaOption;
   /** maxmal number of usage. */
   protected int MAX_USAGE_NUM = 2000;
   /** the variables for calculating the BLA. */
   protected int usedNum, usages[];
   /** the parameters for BLA. */
   protected double BLA, c, d, iBLA;
   /** utility option. */
   protected int utlOption;
   //double utility, benefit, cost, v, c7, c8, PM, NM;
   /** parameters for utility. */
   protected double utility, PM, NM;

   /** constructor. used for NACS rules.
    *  @param global the associated global settings. */
   public RuleAttributes( Global global )
   {
       this.global = global;
       taskClarion = global.getTaskClarion();
       acsRuleFlag = false;
       actionTime = global.getDT_TL_actionTime();
       t0 = global.getDT_TL_t0();
       t1 = global.getDT_TL_t1();
       if( global.isAssocBlaOn() ){
           blaOption = global.getAssocBlaOption();
           if( blaOption == global.CONSTANT_OPTION )
               BLA = global.getAssocBla();
           else{
               c = global.getAssocBlaParas()[global.BLA_C];
               d = global.getAssocBlaParas()[global.BLA_D];
               iBLA = global.getAssocBlaParas()[global.BLA_I];
           }
       }
       else
           blaOption = global.EMPTY;
       //MAX_USAGE_NUM = global.getMaxUsageNum();
       usages = new int[ MAX_USAGE_NUM ];
       //utlOption = global.getAssocUtilityOption();
       //benefit = global.getAssocUtilityParas()[global.UTL_BENE];
       //cost = global.getAssocUtilityParas()[global.UTL_COST];
       //v = global.getAssocUtilityParas()[global.UTL_V];
       //c7 = global.getAssocUtilityParas()[global.UTL_C7];
       //c8 = global.getAssocUtilityParas()[global.UTL_C8];
   }

   /** constructor. used for ACS rules.
    *  @param netIdx the network index this rule belongs to.
    *  @param ruleType rule type.
    *  @param ruleSet rule set this rule belongs to.
    *  @param global the associated global settings. */
   public RuleAttributes( int netIdx, int ruleType, int ruleSet, Global global )
   {
       this.netIdx = netIdx;
       this.ruleType = ruleType;
       this.ruleSet = ruleSet;
       this.global = global;
       taskClarion = global.getTaskClarion();
       acsRuleFlag = true;
       TLPT = global.getPT_TL();
       actionTime = global.getDT_TL_actionTime();
       t0 = global.getDT_TL_t0();
       t1 = global.getDT_TL_t1();
       if( global.isRuleBlaOn( netIdx, ruleType, ruleSet ) ){
           blaOption = global.getActionRuleBlaOption( netIdx, ruleType, ruleSet );
           if( blaOption == global.CONSTANT_OPTION )
               BLA = global.getActionRuleBla( netIdx, ruleType, ruleSet );
           else{
               c = global.getActionRuleBlaParas( netIdx, ruleType, ruleSet )[global.BLA_C];
               d = global.getActionRuleBlaParas( netIdx, ruleType, ruleSet )[global.BLA_D];
               iBLA = global.getActionRuleBlaParas( netIdx, ruleType, ruleSet )[global.BLA_I];
           }
       }
       else
           blaOption = global.EMPTY;
       //MAX_USAGE_NUM = global.getMaxUsageNum();
       usages = new int[ MAX_USAGE_NUM ];
       if( global.isRuleUtilityOn( netIdx, ruleType, ruleSet ) ){
           utlOption = global.getRuleUtilityOption( netIdx, ruleType, ruleSet );
           //benefit = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_BENE];
           //cost = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_COST];
           //v = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_V];
           //c7 = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_C7];
           //c8 = global.getRuleUtilityParas( netIdx, ruleType, ruleSet )[global.UTL_C8];
       }
   }

   /** reinitalization routine. */
   public void reinit()
   {
       PM = 0.0;
       NM = 0.0;
       usedNum = 0;
       for( int i = 0; i < usages.length; i++ )
            usages[i] = 0;
   }

   /** returns age from last application of this rule.
    *  @return age of the rule. */
   public int getAge()
   {
       return age;
   }

   /** returns number of positive matches of this rule.
    *  @return number of positive matches. */
   public double getPM()
   {
       return PM;
   }

   /** returns number of negative matches of this rule.
    *  @return number of negative matches. */
   public double getNM()
   {
       return NM;
   }

   /** returns the utility of this rule.
    *  @return the utility. */
   public double getUtility()
   {
       return utility;
   }

   /** returns the BLA of this rule.
    *  @return the BLA. */
   public double getBLA()
   {
       return BLA;
   }

   /** returns current response time.
    *  @return current response time. */
   public int getResponseTime()
   {
       return TLRT;
   }

   /** returns associative rule application time.
    *  @return associative rule application time. */
   public int getAssocRuleAppTime()
   {
      return assocAppTime;
   }

   /** set age of this rule.
    *  @param age the value to set. */
   public void setAge( int age )
   {
       this.age = age;
   }

   /** set positive matches.
    *  @param PM the value to set. */
   public void setPM( double PM )
   {
       this.PM = PM;
   }

   /** set negative matches.
    *  @param PM the value to set. */
   public void setNM( double NM )
   {
       this.NM = NM;
   }

   /** set number of usages of this rule.
    *  @param usedNum the value to set. */
   public void setUsedNum( int usedNum )
   {
      this.usedNum = usedNum;
   }

   /** set info of usages of this rule.
    *  @param arr the value to set. */
   public void setUsages( int[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           usages[ i ] = arr[ i ];
   }

   /*
   public void setCost( double c )
   {
       cost = c;
   }
   */

   /** set chunk BLA.
    *  @param val the value to set. */
   public void setChunkBLA( double val )
   {
       chunkBLA = val;
   }

   /** reset the matches. */
   public void resetMatches()
   {
       PM = 0.0;
       NM = 0.0;
   }

   /** discount the matches using the given discount value. */
   public void discountMatches( double discount )
   {
       PM *= discount;
       NM *= discount;
   }

   /** calculate the response time given the top-level actuation time.
    *  @param TL_AT the top-level acuation time. */
   public void calcTLRT( int TL_AT )
   {
      TLPT = taskClarion.getPT_TL( global.getPT_BL() );
      int TLAT = TL_AT;
      TLRT = taskClarion.getTLOverallRT( TLPT, TLDT, TLAT );
   }

   /** calculate the TL Decision Time given chunk BLA.
    *  @param chunkBLA the relevant chunk BLA. */
   public void calcTLDT( double chunkBLA )
   {
      this.chunkBLA = chunkBLA;
      //TLDT = (int)( actionTime + t0 / BLA + t1 / chunkBLA );
      TLDT = taskClarion.getDT_TL( global.getDT_BL(), actionTime, BLA, chunkBLA );
   }

   /** calculate the TL Decision Time given chunk BLA.
       @param flag the flag indicates involving NACS or not.
    *  @param chunkRetrTime the relevant chunk retrieval time. */
   public void calcTLAT( boolean flag, int chunkRetrTime )
   {
      calcAssocRuleAppTime();
      int TLAT = taskClarion.getAT_TL( flag, assocAppTime, chunkRetrTime );
   }

   /** calculate the associative rule application time for one iteration. needs
    *  user definition. */
   public void calcAssocRuleAppTime()
   {
      /*
      double t4 = global.getAssocAppTime_t4();
      double t5 = global.getAssocAppTime_t5();
      assocRuleAppTime = (int)(t4 + t5 / BLA);
      */
      assocAppTime = taskClarion.getAssocAppTime( BLA );
   }

   /** calcluate the utility.
    *  @param cost the cost of using this rule.*/
   public void calcUtility( double cost )
   {
      /*
       if( utlOption == global.VARIED_OPTION ){
           benefit = ( c7 + PM ) / ( c8 + PM + NM );
           cost = c;
       }
       utility = benefit - v * cost;
       */
       //this.cost = cost;
       if( acsRuleFlag ){
           switch( ruleType ){
              case Global.RER:
                   utility = taskClarion.getRerUtility( netIdx, ruleSet, PM, NM );
                   break;
              case Global.IRL:
                   utility = taskClarion.getIrlUtility( netIdx, ruleSet, PM, NM );
                   break;
              case Global.FR:
                   utility = taskClarion.getFrUtility( netIdx, ruleSet, PM, NM );
                   break;
          }
       }
   }

   /** calcluate the BLA.
    *  @param curTime current time.*/
   public void calcBLA( int curTime )
   {
       if( blaOption != Global.VARIED_OPTION )
           return;
       else{
           double sum = iBLA;
           for( int i = 0; i < usedNum; i++ ) {
                int timeLaps = curTime - usages[ i ];
                if( timeLaps <= 0 ){
                    break;
                }
                sum += c * Math.pow( (double)timeLaps, -d );
           }
           BLA = sum;
       }
       if( BLA < 0.01 )
           BLA = 0.01;
   }

   /** update the age. */
   public void updateAge()
   {
       age++;
   }

   /** update the matches.
    *  @param flag indicates positive or negative matches.
    */
   public void updateMatches( boolean flag )
   {
      if( flag )
          PM += 1.0;
      else
          NM += 1.0;
   }

   /** update the matches.
    *  @param matchType indicates positive or negative matches.
    */
   public void updateMatches( int matchType )
   {
      if( matchType == global.POS_MATCH_TYPE )
          PM += 1.0;
      if( matchType == global.NEG_MATCH_TYPE )
          NM += 1.0;
   }

   /** update the usage info.
    *  @param curTime current time when this rule is applied. */
   public void updateUsages( int curTime )
   {
       age = 0;
       if( usedNum > 0 && curTime <= usages[ usedNum-1 ] ){
           return;
       }
       if( usedNum < MAX_USAGE_NUM-1 )
           usages[ usedNum++ ] = curTime;
       else
           System.out.println( "error usage update in RuleAttributes.java" );
   }

    /** methods from class <em>Object</em> */
    public Object clone()
   {
      RuleAttributes feature = new RuleAttributes( netIdx, ruleType, ruleSet, global );
      feature.setAge( age );
      feature.setUsedNum( usedNum );
      feature.setUsages( usages );
      feature.setPM( PM );
      feature.setNM( NM );
      feature.setChunkBLA( chunkBLA );
      return feature;
   }
}