package clarion.mcs;
import clarion.tools.*;
import clarion.system.*;
import clarion.acs.*;
import clarion.nacs.*;
import clarion.common.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the MCS subsystem of CLARION model.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> protected void getOutputs( GeneralNet net ): computes the outputs from a
 * specific BP decision network.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements MCS subsystem, the Meta-Cognitive Subsystem, and on the
 * top of MCS implementation structure and all the other classes relevant to MCS
 * are declared inside it.
 * <li> Usage of this class. <br>
 * It is instantiated inside a <em>Clarion</em> object.
 * </ul>
 * @author Xi Zhang
 *
 */

public class MCS {

   final int ACS_COMP_NUM = Global.ACS_COMP_NUM;
   /** global CLARION settings. */
   protected Global global;
   /** the number of ACS action-decision networks. */
   protected int  acsNetNum;
   /** the index of goal action. */
   protected int  goalAction;
   /** the reinforcement. */
   protected double  reinforcement;
   /** the info of the filtered ACS input dimensions. */
   protected short[] acsInputDims[][];
   /** the info of the filtered ACS output dimensions. */
   protected short[] acsOutputDims[][];
   /** the info of the filtered NACS input dimensions. */
   protected short[] nacsInputDims[];
   /** the info of the filtered NACS output dimensions. */
   protected short[] nacsOutputDims[];
   /** the ACS reasoning methods. */
   protected short[] acsReasonMethods[];
   /** the NACS reasoning methods. */
   protected short[] nacsReasonMethods;
   /** the ACS learning methods. */
   protected short[] acsLearnMethods[];
   /** the NACS learning methods. */
   protected short[] nacsLearnMethods;
   protected short[] offsets = new short[ 5 ];
   /** overall input. */
   protected double[] input;
   protected double[] output;
   /** sensory input. */
   protected double[] state;
   /** goal input. */
   protected double[] goal;
   /** working memory input. */
   protected double[] wm;
   /** drive input. */
   protected double[] drives;
   /** monitor buffer input. */
   protected double[] buffer;
   /** the BP net for evaluation. */
   protected GeneralNet evalNet;
   /** the BP net for goal action. */
   protected GeneralNet gsNet;
   /** the BP nets for ACS filters. */
   protected GeneralNet acsInputFilterNets[][], acsOutputFilterNets[][];
   /** the BP nets for NACS filters. */
   protected GeneralNet nacsInputFilterNets[], nacsOutputFilterNets[];
   /** the BP nets for ACS reasoning. */
   protected GeneralNet acsReasonNets[][];
   /** the BP nets for NACS reasoning. */
   protected GeneralNet nacsReasonNets[];
   /** the BP nets for ACS learning. */
   protected GeneralNet acsLearnNets[][];
   /** the BP nets for NACS learning. */
   protected GeneralNet nacsLearnNets[];
   /** the monitor buffer. */
   protected MonitorBuf monitorBuf;
   /** the action decider. */
   protected StochasticDecider decider;

   /** constructor.
    *  @param acs the ACS subsystem.
    *  @param nacs the NACS subsystem.
    *  @param drives the used drives.
    *  @param global the global settings. */
   public MCS( ACS acs, NACS nacs, Drives drives, Global global )
   {
      this.global = global;
      global.setMCS( this );
      acsNetNum = global.getAcsNetNum();
      // the BP nets for filters.
      acsInputFilterNets = new GeneralNet[acsNetNum][ACS_COMP_NUM];
      acsOutputFilterNets = new GeneralNet[acsNetNum][ACS_COMP_NUM];
      nacsInputFilterNets = new GeneralNet[2];
      nacsOutputFilterNets = new GeneralNet[2];
      // the BP nets for reasoning.
      acsReasonNets = new GeneralNet[acsNetNum][ACS_COMP_NUM];
      nacsReasonNets = new GeneralNet[2];
      // the BP nets for learning.
      acsLearnNets = new GeneralNet[acsNetNum][ACS_COMP_NUM];
      nacsLearnNets = new GeneralNet[2];
      monitorBuf = new MonitorBuf( global );
      offsets[0] = 0;
      offsets[1] = (short)(offsets[0] + acs.getSensoryLength());
      offsets[2] = (short)(offsets[1] + acs.getGsLength());
      offsets[3] = (short)(offsets[2] + acs.getWmLength());
      offsets[4] = (short)(offsets[3] + drives.getLength());
      int len = offsets[4] + monitorBuf.getLength();
      input = new double[ len ];
   }

   /** Sets the sensory input.
    *  @param arr the sensory input to set. */
   public void setState( short[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           state[i] = arr[i];
   }

   /** Sets the goal input.
    *  @param arr the goal input to set. */
   public void setGoal( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           goal[i] = arr[i];
   }

   /** Sets the working memory input.
    *  @param arr the working input to set. */
   public void setWM( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           wm[i] = arr[i];
   }

   /** Sets the drive input.
    *  @param arr the drive input to set. */
   public void setDrvies( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           drives[i] = arr[i];
   }

   /** Sets the monitor buffer input.
    *  @param arr the monitor buffer input to set. */
   public void setBuffer( double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           buffer[i] = arr[i];
   }

   /** Returns the monitor buffer.
    *  @return the monitor buffer. */
   public MonitorBuf getMonitorBuffer()
   {
      return monitorBuf;
   }

   /** Returns the reinforcement.
    *  @return the reinforcement. */
   public double getReinforcement()
   {
      getOutputs( evalNet );
      reinforcement = decider.getRecommendation( output, output.length );
      return reinforcement;
   }

   /** Returns the goal action.
    *  @return the goal action. */
   public int getGoalAction()
   {
      getOutputs( gsNet );
      goalAction = decider.getRecommendation( output, output.length );
      return goalAction;
   }

   /** Returns the info of the filtered ACS input dimensions with the specific
    *  network and the component in that network.
    *  @param netIdx the ACS network index.
    *  @param compIdx the component index.
    *  @return the info of the filtered ACS input dimensions. */
   public short[] getAcsInputDims( int netIdx, int compIdx )
   {
      getOutputs( acsInputFilterNets[netIdx][compIdx] );
      return acsInputDims[netIdx][compIdx];
   }

   /** Returns the info of the filtered ACS output dimensions with the specific
    *  network and the component in that network.
    *  @param netIdx the ACS network index.
    *  @param compIdx the component index.
    *  @return the info of the filtered ACS output dimensions. */
   public short[] getAcsOutputDims( int netIdx, int compIdx )
   {
      getOutputs( acsOutputFilterNets[netIdx][compIdx] );
      return acsOutputDims[netIdx][compIdx];
   }

   /** Returns the info of the filtered NACS input dimensions with the specific
    *  component.
    *  @param compIdx the component index.
    *  @return the info of the filtered NACS input dimensions. */
   public short[] getNacsInputDims( int compIdx )
   {
      getOutputs( nacsInputFilterNets[compIdx] );
      return nacsInputDims[compIdx];
   }

   /** Returns the info of the filtered NACS output dimensions with the specific
    *  component.
    *  @param compIdx the component index.
    *  @return the info of the filtered NACS output dimensions. */
   public short[] getNacsOutputDims( int compIdx )
   {
      getOutputs( nacsOutputFilterNets[compIdx] );
      return nacsOutputDims[compIdx];
   }

   /** Returns the reasoning method of the specific network and the component in
    *  that ACS network.
    *  @param netIdx the ACS network index.
    *  @param compIdx the component index.
    *  @return the reasoning method. */
   public int getAcsReasonMethod( int netIdx, int compIdx )
   {
      getOutputs( acsReasonNets[netIdx][compIdx] );
      return acsReasonMethods[netIdx][compIdx];
   }

   /** Returns the reasoning method of the specific component in NACS subsystem.
    *  @param compIdx the component index.
    *  @return the reasoning method. */
   public int getNacsReasonMethod( int compIdx )
   {
      getOutputs( nacsReasonNets[compIdx] );
      return nacsReasonMethods[compIdx];
   }

   /** Returns the learning method of the specific network and the component in
    *  that ACS network.
    *  @param netIdx the ACS network index.
    *  @param compIdx the component index.
    *  @return the learning method. */
   public int getAcsLearnMethod( int netIdx, int compIdx )
   {
      getOutputs( acsLearnNets[netIdx][compIdx] );
      return acsLearnMethods[netIdx][compIdx];
   }

   /** Returns the learning method of the specific component in NACS subsystem.
    *  @param compIdx the component index.
    *  @return the learning method. */
   public int getNacsLearnMethod( int compIdx )
   {
      getOutputs( nacsLearnNets[compIdx] );
      return nacsLearnMethods[compIdx];
   }

   /** Computes the outputs from a specific BP decision network.
    *  @param net the specific decision network. */
   protected void getOutputs( GeneralNet net )
   {
      fillInput();
      //net.setInput( input );
      net.reasoning();
      //output = net.getChosenAction();
   }

   /** fill in the input. */
   private void fillInput()
   {
      fillInput( 0, state );
      fillInput( 1, goal );
      fillInput( 2, wm );
      fillInput( 3, drives );
      fillInput( 4, buffer );
   }

   /** fill in the input. */
   private void fillInput( int idx, double[] arr )
   {
      for( int i = 0; i < arr.length; i++ )
           input[ offsets[idx] + i ] = arr[i];
   }
}