package clarion.nacs;
import java.util.Vector;
import java.util.Arrays;
import clarion.system.*;
import clarion.common.*;
import clarion.tools.*;
import clarion.tools.linkedlist.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement an associative rule in GKS.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void calculateSupport( Vector con, double[] weights ): computes
 * the rule support.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements associative rule which is one of the two basic knowledge
 * representations in GKS.
 * <li> Usage of this class. <br>
 * It is instantiated either inside a <em>AssocRuleGrp</em> object or wherever an
 * associative rule is needed.
 * </ul>
 * @author Xi Zhang
 */

public class AssocRule implements CloneableListElement {

  /** the global settings. */
  Global global;
  /** the task-specific CLARION settings. */
  TaskClarion taskClarion;
  /** GKS reasoning method. */
  int reasoningMethod;
  /** age of the rule since last use. */
  int   age;
  /** condition in terms of chunk indices on the chunk list in GKS. */
  short condition[];
  /** conclusion in terms of chunk index on the chunk list in GKS. */
  short conclusion;
  /** used for calculating the support by intermixed reasoning type. */
  double c14, c15;
  /** rule support. */
  double support;
  /** the rule attributes. */
  RuleAttributes attrib;
 /*********************** rule feature need be added.**************************/

  /** constructor.
   *  @param con the rule condition in terms of chunk indices on the chunk list in GKS.
   *  @param conc the rule conclusion in terms of chunk index on the chunk list in GKS.
   *  @param global the associated global settings. */
  public AssocRule( short[] con, short conc, Global global )
  {
     condition = new short[ con.length ];
     for( int i = 0; i < con.length; i++ )
          condition[ i ] = con[ i ];
     conclusion = conc;
     this.global = global;
     this.taskClarion = global.getTaskClarion();
     c14 = global.getRBR_BalConstant();
     c15 = global.getSBR_BalConstant();
     attrib = new RuleAttributes( global );
  }

  /** Check if the specific chunk is in the rule condition or not.
   *  @param cond the specific chunk.
   *  @return <tt>true</tt> if it is in the rule condition, <tt>false</tt> otherwise. */
  public boolean isInCondition( int cond )
  {
    for( int i = 0; i < condition.length; i++ )
         if( condition[i] == (int)cond )
             return true;
     return false;
  }

  /** Check if the specific condition is the rule condition or not.
   *  @param con the specific condition.
   *  @return <tt>true</tt> if it is the rule condition, <tt>false</tt> otherwise. */
  public boolean isCondition( short[] cond )
  {  // con is sorted.
     if( condition.length != cond.length )
         return false;
     // new revision 11/01/04
     short[] cond1 = CommonMethods.copy( condition );
     short[] cond2 = CommonMethods.copy( cond );
     Arrays.sort( cond1 );
     Arrays.sort( cond2 );
     //for( int i = 0; i < condition.length; i++ )
     //     if( condition[i] != con[i] )
     for( int i = 0; i < cond1.length; i++ )
          if( cond1[i] != cond2[i] )
              return false;
     return true;
  }

  /** Return the age of the rule.
   *  @return the age of the rule. */
  public int getAge()
  {
     return age;
  }

  /** Returns the conclusion in terms of chunk index.
   *  @return the conclusion. */
  public int getConclusion()
  {
     return conclusion;
  }

  /** Returns the number of chunks in the condition.
   *  @return the number of chunks in the condition. */
  public int getConditionNum()
  {
     return condition.length;
  }

  /** Returns the condition in terms of chunk indices on the chunk list in GKS.
   *  @return the condition. */
  public short[] getCondition()
  {
     return CommonMethods.copy( condition );
  }

  /** Returns the condition in terms of chunk indices on the chunk list in GKS.
   *  @param con the array to store the rule condition. */
  public void getCondition( short[] con )
  {
     for( int i = 0; i < con.length; i++ )
          con[ i ] = condition[ i ];
  }

  /** Returns the rule BLA.
   *  @return the rule BLA. */
  public double getBLA()
  {
     return attrib.getBLA();
  }

  /** Return the rule support.
   *  @return the rule support. */
  public double getSupport()
  {
     return support;
  }

  /* reset the rule age. */
  public void resetAge()
  {
     age = 0;
  }

  /** increase the rule age. */
  public void increaseAge()
  {
     age++;
  }

  /** Sets GKS reasoning method.
    *  @param method the reasoning method to set. */
  public void setReasoningMethod( int method )
  {
     reasoningMethod = method;
  }

  /** Sets the rule conclusion in terms of chunk index.
   *  @conc the chunk index used as new conclusion. */
  public void setConclusion( short conc )
  {
     conclusion = conc;
  }

  /** Sets the rule condition in terms of chunk indices.
   *  @con the chunk indices used as new condition. */
  public void setCondition( short[] con )
  {
     condition = new short[ con.length ];
     for( int i = 0; i < con.length; i++ )
          condition[ i ] = con[ i ];
  }

  /** Computes the rule support.
   *  @param con a vector of chunks used to compare the condition to compute
   *  the rule support.
   *  @param weights the weights of subcondition for computing the rule support. */
  public void calculateSupport( Vector con, double[] weights )
  {
     support = 0.0;
     double ruleSupport = 0.0;
     // calculate rule support.
     double[] conditionStrengths = new double[ condition.length ];
     for( int i = 0; i < condition.length; i++ )
          conditionStrengths[ i ] = 0.0;
     for( int i = 0; i < con.size(); i++ ){
          int idx = ( (Integer)con.elementAt(i) ).intValue();
          for( int j = 0; j < condition.length; j++ ){
               if( idx == condition[ j ] ){
                   GKSChunk chunk = global.getNACS().getGks().getChunk(idx);
                   conditionStrengths[ j ] = chunk.getStrength();
                   break;
               }
          }
     }
     for( int i = 0; i < weights.length; i++ )
          ruleSupport += weights[ i ] * conditionStrengths[ i ];
     // calculate similarity support.
     double simiSupport = 0.0;
     if( reasoningMethod == Global.FORWARD_CHAIN_WITH_SBR ){
         double[] similarities = new double[ condition.length ];
         for( int i = 0; i < condition.length; i++ ){
              similarities[i] = 0.0;
              GKSChunk destChunk = global.getNACS().getGks().getChunk( condition[i] );
              for( int j = 0; j < con.size(); j++ ){
                   int idx = ( (Integer)con.elementAt(j) ).intValue();
                   GKSChunk souChunk = global.getNACS().getGks().getChunk( idx );
                   double simi = souChunk.getStrength() * taskClarion.getSimilarity( souChunk, destChunk );
                   if( simi > similarities[i] )
                       similarities[i] = simi;
              }
         }
         simiSupport = 1.0;
         for( int i = 0; i < similarities.length; i++ )
              if( similarities[i] < simiSupport )
                  simiSupport = similarities[i];
         // combine the supports from both rule-based and similarity-based.

         //System.out.println("rule support = " + ruleSupport + ", simi support = " + simiSupport );
         support = Math.max( c14 * ruleSupport, c15 * simiSupport );
     }
     else
         support = ruleSupport;
  }

  /** update the rule usage.
   *  @param currentTime current time when this rule is used. */
  public void increaseUsage( long currentTime )
  {
     attrib.updateUsages( (int)currentTime );
  }

  /** Compute rule BLA.
   *  @param currentTime current time. */
  public double calcBLA( long currentTime )
  {
     attrib.calcBLA( (int)currentTime );
     return attrib.getBLA();
  }

  /** Compute the rule application time. */
  public int calcAssocRuleAppTime()
  {
     attrib.calcAssocRuleAppTime();
     return attrib.getAssocRuleAppTime();
  }

/************************** OVERRIDED METHODS ********************************/
  /** methods from class <em>Object</em> */
  public String toString()
  {
     String str = "( ";
     for( int i = 0; i < condition.length; i++ )
          str += "chunk_" + String.valueOf( condition[ i ] ) + " ";
     str += ") -----> chunk_" + String.valueOf( conclusion ) + "\n";

     return str;
  }

  /** methods from class <em>Object</em> */
  public int hashCode()
  {
     int sum = 0;
     for( int i = 0; i < condition.length; i++ )
          sum += condition[ i ];
     sum += conclusion;

     return sum;
  }

  /** methods from class <em>Object</em> */
  public boolean equals( Object other )
  {
     if( other == null || !( other instanceof AssocRule ) )
         return false;

     AssocRule otherRule = (AssocRule)other;
     if( conclusion != otherRule.getConclusion() )
         return false;
     return isCondition( otherRule.getCondition() );
  }

  /** methods from class <em>Object</em> */
  public Object clone() throws CloneNotSupportedException
  {
     AssocRule rule = new AssocRule( condition, conclusion, global );
     return rule;
  }
}
