package clarion.nacs;
import java.util.Vector;
import java.util.Arrays;
import clarion.common.*;
import clarion.tools.*;
import clarion.system.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the concept of chunk in GKS.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public boolean activateChunk( Feature[] cue, int whichLevel ): Checks if
 * this chunk is activated or not and computes the chunk strength.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * This class implements the chunk in GKS which is one of the two basic knowledge
 * representations in GKS.
 * <li> Usage of this class. <br>
 * It is instantiated either inside a <em>GKS</em> object or wherever a GKS chunk
 * is needed.
 * </ul>
 * @author Xi Zhang
 */

public class GKSChunk extends Chunk {

   final int AMN_IDX = 0, GKS_IDX = 1;

   /** the global settings. */
   Global global;
   /** the index to an IDN the chunk is from. */
   int    idnIdx = -1;
   /** the index to an AMN the chunk is in. */
   int    amnIdx = -1;
   /** id of the chunk. */
   int    chunkID;
   /** age from last usage. */
   int    age;
   /** chunk retrieval time. */
   int    retrievalTime;
   /** current time. */
   long   currentTime;
   /** the conclusion dimension. */
   short  CONC_IDX;
   /** an associative rule group with this chunk as the conclusion. */
   AssocRuleGrp rules;
   /** mode of activation: minimum or average. */
   int    activationMode;
   /** record the matched dimensions with the cue. */
   short  dimMatches[];
   /** the dimensional weights. */
   double weights[];      //
   /** the dimensional strength. */
   double dimStrengths[];
   /** the chunk is in EM or not. */
   boolean inEmOn;
   /** this variable is not used. */
   boolean availableFlag;

   /** new constructor.
    *  @param cId chunk ID.
    *  @param type chunk type.
    *  @param label chunk label.
    *  @param features chunk features.
    *  @param idnIdx the index to an IDN this chunk is from.
    *  @param amnIdx the index to an AMN this chunk is from.
    *  @param g the global settings. */
   public GKSChunk( int cId, int type, String label, Feature[] features,
                    int idnIdx, int amnIdx, Global g )
   {
      super( type, label, features, g );
      chunkID = cId;
      this.idnIdx = idnIdx;
      this.amnIdx = amnIdx;
      global = g;
      rules = new AssocRuleGrp( chunkID, global );
      activationMode = global.getActivationMeasure();
      dimMatches = new short[ getDimNum() ];
      weights = global.getTaskClarion().getAssocDimWeights( getDimNum() );
      dimStrengths = new double[ getDimNum() ];
   }

   /** new constructor for a combined chunk.
    *  @param cId chunk ID.
    *  @param type chunk type.
    *  @param label chunk label.
    *  @param chunks chunks to compose this chunk.
    *  @param idnIdx the index to an IDN this chunk is from.
    *  @param amnIdx the index to an AMN this chunk is from.
    *  @param g the global settings. */
   public GKSChunk( int cId, int type, String label, Chunk[] chunks,
                    int idnIdx, int amnIdx, Global g )
   {
      super( type, label, chunks, g );
      chunkID = cId;
      this.idnIdx = idnIdx;
      this.amnIdx = amnIdx;
      global = g;
      rules = new AssocRuleGrp( chunkID, global );
      activationMode = global.getActivationMeasure();
      dimMatches = new short[ getDimNum() ];
      dimStrengths = new double[ getDimNum() ];
   }

/***************************** GET AND SET METHODS ****************************/
   /** Returns the index to an IDN this chunk is from.
    *  @return the index to an IDN this chunk is from. */
   public int getIdnIdx()
   {
      return idnIdx;
   }

   /** Returns the index to an AMN this chunk is from.
    *  @return the index to an AMN this chunk is from. */
   public int getAmnIdx()
   {
      return amnIdx;
   }

   /** Returns the chunk age.
    *  @return the chunk age. */
   public int getAge()
   {
      return age;
   }

   /** Returns the number of rules with this chunk as conclusion.
    *  @return the number of rules. */
   public int getRuleNum()
   {
      return rules.getRuleNum();
   }

   /** Returns the rules with this chunk as conclusion.
    *  @return the rules. */
   public AssocRuleGrp getRules()
   {
      return rules;
   }

   /** Returns the conditions of the rules with this chunk as conclusion. The
    *  conditions are in terms of chunk indices to the chunk list in GKS.
    *  @return the conditions of the rules. */
   public short[][] getRuleConditions()
   {
      return rules.getRuleConditions();
   }

   /** Returns the conditions of the activated rules with this chunk as conclusion.
    *  The conditions are in terms of chunk indices to the chunk list in GKS.
    *  @return the conditions of the activated rules. */
   public short[][] getActiveRuleConditions()
   {
      return rules.getActiveRuleConditions();
   }

   /** Returns the index to the conclusion dimension.
    *  @return the index to the conclusion dimension.
    *  @deprecated it is not used any more. */
   public int getConclusionDimIdx( )
   {
      return CONC_IDX;
   }

   /** not used any more. */
   public boolean getAvailableFlag()
   {
      return availableFlag;
   }

   /** Returns if chunk is in EM or not. */
   public boolean isInEmOn()
   {
      return inEmOn;
   }

   /** Returns the conclusion dimension.
    *  @return the conclusion dimension.
    *  @deprecated it is not used any more. */
   public Feature getConclusion()
   {
      return featList[ CONC_IDX ];
   }

   /** Returns the retrieval time.
    *  @return the retrieval time. */
   public int getRetrievalTime()
   {
      return retrievalTime;
   }

   /** Returns the associated rule application time.
    *  @return the associated rule application time. */
   public int getAssocRuleAppTime()
   {
      return rules.getAssocRuleAppTime();
   }

   /** reset the chunk age. */
   public void resetAge()
   {
      age = 0;
   }

   /** Sets GKS reasoning method.
    *  @param method the reasoning method to set. */
   public void setReasoningMethod( int method )
   {
      //reasoningMethod = method;
      rules.setReasoningMethod( method );
   }

   /** Sets current time.
    *  @param curTime the time to set. */
   public void setCurrentTime( long curTime )
   {
      currentTime = curTime;
      rules.setCurrentTime( currentTime );
   }

   /** Sets the assoicated rules.
    *  @param ruleGrp the rule group to set. */
   public void setRules( AssocRuleGrp ruleGrp )
   {
      try{
          if( ruleGrp != null )
              rules = (AssocRuleGrp)(ruleGrp.clone());
      }
      catch(  CloneNotSupportedException e ) {
         System.out.println( "clone AssocRuleGrp error from GKSChunk.java" );
      }
   }

   /** not used any more. */
   public void setAvailableFlag( boolean flag )
   {
      availableFlag = flag;
   }

   /** Sets flag indicating this chunk is in EM or not.
    *  @param flag the value to set. */
   public void setInEmOn( boolean flag )
   {
      inEmOn = flag;
   }

   /** Sets the index to the conclusion dimension.
    *  @param idx the index to set.
    *  @deprecated it is not used any more. */
   public void setConclusionDimIdx( short idx )
   {
       CONC_IDX = idx;
   }

/*************************** REASONING METHODS ********************************/
   /** Checks if this chunk is activated or not and computes the chunk strength.
    *  To say a chunk is activated, it means that all or some of its dimensions are
    *  activated to a strength level.
       A dimension of the chunk is activated to a strength level means one of
       allowable values in the dimension is activated with a strength( only for
       the case that the value # cue has in that dimension is 1. )
       Generally, To say a dimension of a chunk is activated, all of the values
       cue has in the dimension should be contained in the dimension of the chunk.
       @param cue the cue to activate the reasoning.
       @param whichLevel the level (GKS or AMN) this reasoning process is in.
       @return <tt>true</tt> if the reasoning is similarity-based and some dimensions
       are matched or the reasoning is rule-based and all dimensions are matched.
       <tt>false</tt>, otherwise.  */
   public boolean activateChunk( Feature[] cue, int whichLevel )
   {
       for( int i = 0; i < getDimNum(); i++ ){
            dimMatches[ i ] = 0;
            dimStrengths[ i ] = 0.0;
       }
       // fill in dimensional strengths.
       int matchCount = 0;
       if( featList.length > 0 ){
           for( int i = 0; i < cue.length; i++ ){
                int idx = searchFeature( cue[ i ].getLabel() );
                if( idx == -1 || cue[ i ].getType() != featList[ idx ].getType() )
                    continue;
                Feature f1 = featList[ idx ];
                Feature f2 = cue[ i ];
                dimMatches[ idx ] = CommonMethods.contain( f1.getContent(), f2.getContent() );
                if( dimMatches[ idx ] == 1 ){
                    matchCount++;
                    if( whichLevel == GKS_IDX )
                        dimStrengths[ idx ] = 1.0;
                    else
                        dimStrengths[ idx ] = calcDimStrength( f1, f2 );
                }
                else
                    dimStrengths[ idx ] = 0.0;
            }
       }
       // calculate strength.
       strength = 0.0;
       if( matchCount == 0 || (!global.isSimilarityOn() && matchCount < getDimNum()) )
           return false;
       if( whichLevel == GKS_IDX )
           for( int i = 0; i < getDimNum(); i++ )
                strength += dimStrengths[i] * weights[i];
       else
           for( int i = 0; i < getDimNum(); i++ )
                strength += dimStrengths[i] / getDimNum() ;
       return true;
   }

   /** Calculates the support from the associative rules pointing to this chunk.
    *  @param con the vector of chunks to trigger the proccess of computing rule support. */
   public void calculateStrength( Vector con )
   {
       if( getRuleNum() > 0 )
           strength = rules.drawConclusion( con );
   }

   /** Search a rule in this chunk with the specific condition in terms of chunk
    *  indices on the chunk list in GKS. */
   public boolean searchRule( short[] conditions )
   {
       return rules.searchRule( conditions );
   }

/****************************** BLA METHODS ***********************************/
   /** Updates the usage. */
   public void increaseUsage()
   {
       super.increaseUsage( currentTime );
   }

   /** Computes the retrieval time. */
   public void calcRetrievalTime()
   {
      /*
      double t2 = global.getChunkRetTime_t2();
      double t3 = global.getChunkRetTime_t3();
      super.calcBLA( currentTime );
      retrievalTime = (int)(t2 + t3 / BLA);
      */
      super.calcBLA( currentTime );
      retrievalTime = global.getTaskClarion().getChunkRetrTime( BLA );
   }

/************************ INSERT AND REMOVE METHODS ***************************/
  /** insert a rule with the specific condition and conclusion in terms of the
   *  chunk indices on the chunk list in GKS.
   *  @param condition the rule condition.
   *  @param conclusion the rule conclusion. */
   public boolean insertRule( short[] condition, short conclusion )
   {
       return rules.insertRule( condition, conclusion );
   }

  /** Remove a rule with the specific condition and conclusion in terms of the
   *  chunk indices on the chunk list in GKS.
   *  @param condition the rule condition.
   *  @param conclusion the rule conclusion. */
   public boolean removeRule( short[] condition, short conclusion )
   {
       return rules.removeRule( condition, conclusion );
   }

  /** Removes any rule which condition contains the specific chunk index.
   *  @param chunkIdx the chunk index.
   *  @return true if such rules exist, false otherwise. */
   public boolean removeCondition( int chunkIdx )
   {
       return rules.removeCondition( chunkIdx );
   }

/*************************** CHUNK UPDATE METHODS *****************************/
   /** Update routine. */
   public boolean update()
   {
       age++;
       int MAX_CHUNK_AGE = (int)( 1.0 / global.getChunkDensity() );
       if( age >= MAX_CHUNK_AGE ){
           //System.out.println( "age = " + age );
           return true;
       }

       int MAX_RULE_AGE = (int)( 1.0 / global.getAssocDensity() );
       rules.setMaxRuleAge( MAX_RULE_AGE );
       rules.update();

       return false;
   }

/************************** OVERRIDED METHODS ********************************/
   /** methods from class <em>Object</em> */
   public String toString()
   {
      String str = super.toString();
      //str += "CHUNK AGE = " + String.valueOf( age ) + "\n";
      String prefix = Global.TABLE_STR + Global.TABLE_STR;
      str += prefix + "Rules pointing to this chunk = \n" + rules;
      return str;
   }

   /** methods from class <em>Object</em> */
   public Object clone() throws CloneNotSupportedException
   {
       GKSChunk chunk = new GKSChunk( chunkID, chunkType, chunkLabel, featList, idnIdx, amnIdx, global );
       // from super class.
       chunk.setStrength( strength );
       chunk.setBlaOption( chunkBlaOption );
       if( chunkBlaOption == global.CONSTANT_OPTION )
           chunk.setBla( BLA );
       else{
           chunk.setBLA_C( c );
           chunk.setBLA_D( d );
           chunk.setBLA_INIT( iBLA );
           chunk.setUsageNum( usageNum );
           chunk.setUsageRecords( usageRecords );
       }
       // from this class
       // chunk.setConclusionDimIdx( CONC_IDX );
       chunk.setAvailableFlag( availableFlag );
       chunk.setRules( rules );
       //chunk.setAge( age );
       return chunk;
   }

/************************* PRIVATE METHODS ************************************/
   /** Calculates f1 dimension strength of the chunk activated by AMN f2 dimension.*/
   private double calcDimStrength( Feature f1, Feature f2 )
   {
       double actVal = 0.0;
       short[] values = f1.getValues();
       if( activationMode == 0 ){ // minimum mode.
           actVal = 1.0;
           for( int i = 0; i < values.length; i++ ){
                double val = f2.getActivations()[ values[ i ] ];
                if( val < actVal )
                    actVal = val;
           }
       }
       else{  // average mode.
           for( int i = 0; i < values.length; i++ ){
                double val = f2.getActivations()[ values[ i ] ];
                actVal += val;
           }
           actVal /= values.length;
       }
       return actVal;
   }
}