package clarion.options.transfer.subsys;
import java.io.*;
import clarion.system.*;
import clarion.options.*;
import clarion.options.transfer.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the process of transferring ACS configuration data
 * between GUI and <tt>Global</tt> object or persistent storage.
 * <li> important methods <br>
 * <ol>
 * <li> public void load(): implement the process of loading options from persistent
 * storage.
 * <li> public void save(): implement the process of save the options to persistent
 * storage or to the associated <tt>Global</tt> object.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model. <br>
 * This class is part of data transfer subsystem and is not related directly to
 * CLARION model.
 * <li> Usage of this class. <br>
 * It extends the base <tt>Transfer</t>> class and is instantiated in the class
 * <em>TransferManager</em>.
 * </ul>
 * @author Xi Zhang
 *
 **/

public class AcsTransfer extends Transfer {

  /** constants indicate different components in ACS subsystem. */
  protected final int ACS_COMP_NUM = 4;
  protected final int ACS_COMP = 0, ACS_NET = 1, ACS_WM = 2, ACS_GS = 3;
  protected final int ACS_NET_OPT_NUM = 4;
  protected final int EX_NET = 0, WM_NET = 1, GS_NET = 2, ACS_COR = 3;
  protected final int ACS_NET_COMP_NUM = 5;
  protected final int IDN = 0, RER = 1, IRL = 2, FR = 3, COMP = 4;
  String[] acsNetCompNames = { "IDN", "RER", "IRL", "FR" };
  String[] acsNetFileNames = { "idn", "rer", "irl", "fr", "components" };

  /** constructor. instantiated with the associated <tt>TransferManager</tt> object
   * and <tt>Global</tt> object. */
  public AcsTransfer( TransferManager manager, Global global )
  {
     super( manager, global );
  }

  /** Get the Goal Structure type.
   *  0: goal stack, 1: goal list, 2: not selected.
   *  @return the Goal Structure type. */
  public int getGsOption()
  {
     return userOptions[ACS_COMP][0][0][0][0][0].getUserChoice();
  }

  /** Get the Working Memory option to check if it is selected or not.
   *  @return the Working Memory option. */
  public boolean getWmOption()
  {
     return userOptions[ACS_COMP][0][0][0][0][1].getUserChoice()==ON;
  }

  /** Get the number of ACS external networks.
   *  @return the number of ACS external networks. */
  public int getAcsExNetNum()
  {
     return userOptions[ACS_COMP][0][0][0][0][2].getOptionItem(0).getParameter(0).getIntValue();
  }

  /** Get the Goal Structure IDN option to check if the IDN is selected or not.
   *  @return the Goal Structure IDN option. */
  public boolean isGsIdnOn()
  {
     return userOptions[ACS_COMP][0][0][0][0][3].getUserChoice()==ON;
  }

  /** Get the Working Memory IDN option to check if the IDN is selected or not.
   *  @return the Goal Structure IDN option. */
  public boolean isWmIdnOn()
  {
     return userOptions[ACS_COMP][0][0][0][0][4].getUserChoice()==ON;
  }

  /** Get ACS component option in one ACS network to check if the component is
   *  selected or not.
   *  @param netType : external, GS or WM network.
   *  @param netidx : the index on one type of networks.
   *  @param type: the component type IDN, RER, IRL or FR.
   *  @return ACS component option. */
  public boolean isAcsCompOn( int netType, int netIdx, int type )
  {
     return userOptions[ACS_NET][netType][netIdx][COMP][0][type].getUserChoice()==ON;
  }

  /** Get the number of a specified component allowed in one ACS network.
   *  @param netType : external, GS or WM network.
   *  @param netidx : the index on one type of networks.
   *  @param type: the component type IDN, RER, IRL or FR.
   *  @return the number of a specified component allowed in one ACS network. */
  public int getAcsCompFileNum( int netType, int netIdx, int type )
  {
     return userOptions[ACS_NET][netType][netIdx][COMP][0][type].getOptionItem(0).getParameter(ON).getIntValue();
  }

  /** the process of loading configurations from persistent storage. */
  public void load()
  {
     sysOptions = new Option[ ACS_COMP_NUM ][][][][][];
     sysOptions[ACS_NET] = new Option[ ACS_NET_OPT_NUM ][][][][];
     userOptions = new Option[ ACS_COMP_NUM ][][][][][];
     userOptions[ACS_NET] = new Option[ ACS_NET_OPT_NUM ][][][][];
     subsysPath = this.searchFile( "acs" );
     loadAcsCompSettings();
     for( int i = 0; i < ACS_NET_OPT_NUM-1; i++ )
          loadAcsNetSettings( i );
     loadAcsCorSettings();
     loadAcsGSSettings();
     loadAcsWMSettings();
  }

  /** the process of saving the options to persistent storage or to the
   *  associated <tt>Global</tt> object. */
  public void save()
  {
     if( saveCode == SAVE_TO_FILE ){
         // create ACS setting directory.
         String path = global.getSettingDir() + "/" + global.getTaskName()
                          + "/" + global.getTaskSettingName() + "/acs";
         new File( path ).mkdir();
         path = global.getSettingDir() + "/" + global.getTaskName() + "/" + global.getTaskSettingName()
                                             + "/acs/nets/";
         new File( path ).mkdir();
     }
     // save ACS component settings.
     saveAcsCompSettings();

     //System.out.println( "ACS: after comp");

     // save ACS net setitngs.
     int netNum = manager.getAcsExNetNum();
     if( manager.getWmOption() )
         netNum++;
     if( manager.getGsOption() != global.GS_NON )
         netNum++;
     if( saveCode == SAVE_TO_GLOBAL ){
         global.setAcsNetNum( netNum );
         global.initAcsNets();
     }

     //System.out.println( "ACS: after init acs nets");

     saveAcsNetSettings( EX_NET );

     //System.out.println( "ACS: after ex net");

     //System.out.println( "after acs net settings" );
     //System.out.println( "irl set num = " + global.getIrlRuleNums(0).length );

     if( manager.getGsOption() != global.GS_NON ){
         saveAcsGSSettings();
         if( manager.isGsIdnOn() )
             saveAcsNetSettings( GS_NET );
     }

     //System.out.println( "ACS: after gs net");

     if( manager.getWmOption() ){
         saveAcsWMSettings();
         if( manager.isWmIdnOn() )
             saveAcsNetSettings( WM_NET );
     }

     //System.out.println( "ACS: after wm net");

     // save coordinating.
     saveAcsCoordinate();

     //System.out.println( "ACS: after coordi" );
  }

  public void update()
  {

  }

/************************* PRIVATE METHODS ************************************/
  /** loading configuration for selection of ACS components. */
  private void loadAcsCompSettings()
  {
     loadedFileName = "components.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/acs/";
     sysOptions[ACS_COMP] = new Option[1][1][1][1][];
     sysOptions[ACS_COMP][0][0][0][0] = loadOptions( path );
     path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/acs/";
     userOptions[ACS_COMP] = new Option[1][1][1][1][];
     userOptions[ACS_COMP][0][0][0][0] = loadOptions( path );
  }

  /** loading configuration for coordination of ACS networks.  */
  private void loadAcsCorSettings()
  {
     loadedFileName = "coordinate.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/acs/nets/";
     sysOptions[ACS_NET][ACS_COR] = new Option[1][1][1][];
     sysOptions[ACS_NET][ACS_COR][0][0][0] = loadOptions( path );
     path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/acs/nets/";
     userOptions[ACS_NET][ACS_COR] = new Option[1][1][1][];
     userOptions[ACS_NET][ACS_COR][0][0][0] = loadOptions( path );
  }

  /** loading configuration for ACS network with particular type.
   * @param netType: external, GS or WM network. */
  private void loadAcsNetSettings( int netType )
  {
     String[] subNetName = { "ex", "wm", "gs" };
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/acs/nets/" + subNetName[netType];

     String [] a = new File( path ).list();
     int netNum = a.length;
     //Handle .DS_STORE on Apple
     for(int b = 0; b < a.length; b++)
     {
    	 if(a[b].indexOf(".DS_Store") != -1)
    		 netNum--;
    		 
     }
     int MAX_NET_NUM = 1;
     if( netType == EX_NET )
         MAX_NET_NUM = Global.MAX_EX_NET_NUM;
     sysOptions[ACS_NET][netType] = new Option[netNum][ACS_NET_COMP_NUM][][];

     //System.out.println( "before fill in sysOptions, netType = " + netType );

     // fill in system options.
     for( int i = 0; i < netNum; i++ )
        for( int j = 0; j < ACS_NET_COMP_NUM; j++ ){
             if( j == COMP ){
                 loadedFileName = acsNetFileNames[ j ] + ".txt";
                 sysOptions[ACS_NET][netType][i][j] = new Option[1][];
                 sysOptions[ACS_NET][netType][i][j][0] = loadOptions( path +
                            "/net" + i + "/" );
                 continue;
             }
             File[] files = new File( path + "/net" + i + "/" + acsNetCompNames[j] ).listFiles();
             int b = files.length;
             for(int l = 0; l < files.length; l++)
             {
            	 if(files[l].getName().indexOf(".DS_Store") != -1)
            		 b--;
             }
             sysOptions[ACS_NET][netType][i][j] = new Option[b][];
             for( int k = 0; k < b; k++ ){
                  loadedFileName = acsNetFileNames[ j ] + k + ".txt";
                  sysOptions[ACS_NET][netType][i][j][k] = loadOptions( path + "/net" + i +
                      "/" + acsNetCompNames[j] + "/" );
             }
        }

     //System.out.println( "before fill in user options, netType = " + netType );

     // fill in user options.
     userOptions[ACS_NET][netType] = new Option[MAX_NET_NUM][ACS_NET_COMP_NUM][][];
     path = global.getSettingDir() + "/" + global.getTaskName() + "/" +
            global.getTaskSettingName() + "/acs/nets/" + subNetName[netType];
     if( !new File( path ).exists() )
     { // the type of net doesn't exist, copy from system opitons.
       //System.out.println( "the net type doesn't exist, copy from DEFAULT opitons." );
       for( int i = 0; i < MAX_NET_NUM; i++ ){
         userOptions[ACS_NET][netType][i] = new Option[sysOptions[ACS_NET][netType][0].length][][];
         for( int j = 0; j < ACS_NET_COMP_NUM; j++ ){
           userOptions[ACS_NET][netType][i][j] = new Option[sysOptions[ACS_NET][netType][0][j].length][];
           for( int k = 0; k < sysOptions[ACS_NET][netType][0][j].length; k++ ){
             userOptions[ACS_NET][netType][i][j][k] = new Option[sysOptions[ACS_NET][netType][0][j][k].length];
             for( int m = 0; m < sysOptions[ACS_NET][netType][0][j][k].length; m++ )
                  userOptions[ACS_NET][netType][i][j][k][m] =
                        (Option)(sysOptions[ACS_NET][netType][0][j][k][m].clone());
             }
         }
       }
       return;
     }

     //System.out.println( "before the type of net exists, netType = " + netType );

     // the type of net exists.
     for( int i = 0; i < MAX_NET_NUM; i++ ){
          if( i < netNum ){
              File netDir = new File( path + "/net" + i );
              File[] compDirs = netDir.listFiles();
              for( int j = 0; j < Global.ACS_NET_COMP_NUM; j++ ){
                  if( j == COMP ){
                      // the userOptions for components.
                      userOptions[ACS_NET][netType][i][j] = new Option[1][];
                      loadedFileName = acsNetFileNames[ j ] + ".txt";
                      userOptions[ACS_NET][netType][i][j][0] = loadOptions( path +
                            "/net" + i + "/" );
                      continue;
                  }
                  int maxFileNum = Global.MAX_RULE_NUMS[j];
                  int fileNum = 1;
                  int b = compDirs.length;
                  for(int l = 0; l < compDirs.length; l++)
                  {
                 	 if(compDirs[l].getName().indexOf(".DS_Store") != -1)
                 		 b--;
                  }
                  userOptions[ACS_NET][netType][i][j] = new Option[maxFileNum][];
                  boolean compOn = false;
                  for( int k = 0; k < compDirs.length; k++ )
                       if(acsNetCompNames[j].equals(compDirs[k].getName())){
                          fileNum = b;
                          compOn = true;
                          break;
                       }
                  if( compOn )
                  {  // existing component.
                     for( int m = 0; m < maxFileNum; m++ ){
                          if( m < fileNum ){
                              loadedFileName = acsNetFileNames[ j ] + m + ".txt";
                              userOptions[ACS_NET][netType][i][j][m] = loadOptions( path +
                                 "/net" + i + "/" + acsNetCompNames[j] + "/" );
                          }
                          else{
                              loadedFileName = acsNetFileNames[ j ] + 0 + ".txt";
                              userOptions[ACS_NET][netType][i][j][m] = loadOptions( path +
                                 "/net" + i + "/" + acsNetCompNames[j] + "/" );
                          }
                     }
                 }
                 else{ // the component doesn't exist, copy from corresponding system options.
                     //System.out.println( "the component doesn't exist, copy from corresponding DEFAULT options." );
                     fileNum = sysOptions[ACS_NET][netType][0][j].length;
                     for( int m = 0; m < maxFileNum; m++ )
                          if( m < fileNum ){
                              int optNum = sysOptions[ACS_NET][netType][0][j][m].length;
                              userOptions[ACS_NET][netType][i][j][m] = new Option[optNum];
                              for( int n = 0; n < optNum; n++ )
                                   userOptions[ACS_NET][netType][i][j][m][n] =
                                      (Option)(sysOptions[ACS_NET][netType][0][j][m][n].clone());
                          }
                          else{
                             int optNum = sysOptions[ACS_NET][netType][0][j][0].length;
                             userOptions[ACS_NET][netType][i][j][m] = new Option[optNum];
                             for( int n = 0; n < optNum; n++ )
                                  userOptions[ACS_NET][netType][i][j][m][n] =
                                     (Option)(sysOptions[ACS_NET][netType][0][j][0][n].clone());
                          }
                 }
             }
          }
          else{  // copy from other networks.
             //System.out.println( "the component doesn't exist, copy from other networks" );
             userOptions[ACS_NET][netType][i] = new Option[userOptions[ACS_NET][netType][0].length][][];
             for( int j = 0; j < ACS_NET_COMP_NUM; j++ ){
               userOptions[ACS_NET][netType][i][j] = new Option[userOptions[ACS_NET][netType][0][j].length][];
               for( int k = 0; k < userOptions[ACS_NET][netType][0][j].length; k++ ){
                 userOptions[ACS_NET][netType][i][j][k] = new Option[userOptions[ACS_NET][netType][0][j][k].length];
                 for( int m = 0; m < userOptions[ACS_NET][netType][0][j][k].length; m++ )
                      userOptions[ACS_NET][netType][i][j][k][m] =
                            (Option)(userOptions[ACS_NET][netType][0][j][k][m].clone());
                 }
             }
          }
     }
  }

  /** loading configuration for Goal Structure. */
  private void loadAcsGSSettings()
  {
     loadedFileName = "goalstruct.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/acs/";
     sysOptions[ACS_GS] = new Option[1][1][1][1][];
     sysOptions[ACS_GS][0][0][0][0] = loadOptions( path );

     /*
     path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/acs/";
     if( new File( path + loadedFileName ).exists() ){
         userOptions[ACS_GS] = new Option[1][1][1][1][];
         userOptions[ACS_GS][0][0][0][0] = loadOptions( path );
     }
     else{
         //System.out.println( "the Goal Structure doesn't exist, copy from corresponding DEFAULT options." );
         int len = sysOptions[ACS_GS][0][0][0][0].length;
         userOptions[ACS_GS] = new Option[1][1][1][1][len];
         for( int i = 0; i < len; i++ )
              userOptions[ACS_GS][0][0][0][0][i] = (Option)sysOptions[ACS_GS][0][0][0][0][i].clone();
     }
     */
     userOptions[ACS_GS] = new Option[1][1][1][1][];
     path = searchFile( "acs", loadedFileName );
     userOptions[ACS_GS][0][0][0][0] = loadOptions( path );

     //System.out.println( "path = " + path );

  }

  /** loading configuration for Working Memory. */
  private void loadAcsWMSettings()
  {
     loadedFileName = "workmem.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/acs/";
     sysOptions[ACS_WM] = new Option[1][1][1][1][];
     sysOptions[ACS_WM][0][0][0][0] = loadOptions( path );
     userOptions[ACS_WM] = new Option[1][1][1][1][];
     path = searchFile( "acs", loadedFileName );
     userOptions[ACS_WM][0][0][0][0] = loadOptions( path );
  }

  /** save the configuration for selection of ACS components. */
  private void saveAcsCompSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalAcsComponents();
         return;
     }
     // create ACS component setting file.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/acs/";
     savedFileName = "components.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save ACS component setting.
     saveOptions( path, userOptions[ACS_COMP][0][0][0][0] );
  }

  /** save the configuration for coordinating of ACS networks. */
  private void saveAcsCoordinate()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalAcsCoordinate();
         return;
     }
     // create ACS component setting file.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/acs/nets/";
     savedFileName = "coordinate.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save ACS component setting.
     saveOptions( path, userOptions[ACS_NET][ACS_COR][0][0][0] );
  }

  /** save configuration for ACS network with particular type.
   * @param netType: external, GS or WM network. */
  private void saveAcsNetSettings( int netType )
  {
     String[] subNetName = { "ex", "wm", "gs" };
     String path = global.getSettingDir() + "/" + global.getTaskName() + "/" + global.getTaskSettingName()
                    + "/acs/nets/" + subNetName[netType] +"/";
     if( saveCode == SAVE_TO_FILE ){
         // create a directory /acs/nets
         new File( path ).mkdir();
     }
     int netNum = 1;
     if( netType == EX_NET )
         netNum = manager.getAcsExNetNum();
     try{
       for( int i = 0; i < netNum; i++ ){
            if( saveCode == SAVE_TO_FILE ){
                File netDir = new File( path + "/net" + i );
                // create dir /acs/nets/neti
                netDir.mkdir();
            }
            for( int j = 0; j < ACS_NET_COMP_NUM; j++ ){
                 if( j == COMP ){
                    if( saveCode == SAVE_TO_FILE ){
                        savedFileName = acsNetFileNames[j] + ".txt";
                        saveOptions( path + "/net" + i + "/", userOptions[ACS_NET][netType][i][j][0] );
                     }
                     else {

                      //System.out.println( "before init net basics" );

                        initGlobalAcsNet( netType, i, j, 0, 1 );

                       // System.out.println( "after init net basics" );

                     }
                  }
                 else if( manager.isAcsCompOn( netType, i, j ) ){ // if the component exists
                        if( saveCode == SAVE_TO_FILE ){
                            File compDir = new File( path + "/net" + i + "/" + acsNetCompNames[j] );
                            // create dir: /acs/nets/neti/IDN, RER,IRL,or FR.
                            boolean bVal = compDir.mkdir();
                         }
                         int fileNum = manager.getAcsCompFileNum( netType, i, j );
                         // convert netIdx.
                         int netIdx = 0;
                         if( netType == this.EX_NET )
                             netIdx = i;
                         if( netType == this.GS_NET )
                             netIdx = global.getGsNetIdx();
                         if( netType == this.WM_NET )
                             netIdx = global.getWmNetIdx();
                         if( saveCode == SAVE_TO_GLOBAL )
                             global.initAcsRules( netIdx, j, fileNum );
                         for( int k = 0; k < fileNum; k++ ){
                              if( saveCode == SAVE_TO_FILE ){
                                  savedFileName = acsNetFileNames[j] + k + ".txt";
                                  saveOptions( path + "/net" + i + "/" + acsNetCompNames[j] +"/",
                                               userOptions[ACS_NET][netType][i][j][k] );
                              }
                              else {
                                  initGlobalAcsNet( netType, i, j, k, fileNum );
                                  //if( j == IRL )
                                   //System.out.println( "knowledge = " +  j + "file num = " + fileNum );
                              }
                         }
                 }
                 //if( j >= IRL )
                 //    System.out.println( "irl set num = " + global.getIrlRuleNums(0).length );
            }
       }
     }
     catch( Exception e ){
       System.out.println( e );
     }
  }

  /** save configuration for Goal Structure. */
  private void saveAcsGSSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalAcsGS();
         return;
      }
     // create ACS GS setting file
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/acs/";
     savedFileName = "goalstruct.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save ACS GS setting.
     saveOptions( path, userOptions[ACS_GS][0][0][0][0] );
  }

  /** save configuration for Working Memory. */
  private void saveAcsWMSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalAcsWM();
         return;
      }
     // create ACS WM setting file
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/acs/";
     savedFileName = "workmem.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save ACS WM setting.
     saveOptions( path, userOptions[ACS_WM][0][0][0][0] );
  }

  /** initialize the variables relevant to selection of ACS components in the
   * associated <tt>Global</tt> object  */
  private void initGlobalAcsComponents()
  {
     final int GS_OPT = 0, WM_OPT = 1, EX_NET_NUM = 2, GS_IDN_ON = 3, WM_IDN_ON = 4;
     Option[] opts = userOptions[ACS_COMP][0][0][0][0];
     global.setGsOption( opts[GS_OPT].getUserChoice() );
     global.setWmOption( opts[WM_OPT].getUserChoice()==ON );
     global.setExNetNum( opts[EX_NET_NUM].getOptionItem(0).getParameter(0).getIntValue() );
     global.setGsIdnOn(opts[GS_IDN_ON].getUserChoice() == ON);
     global.setWmIdnOn(opts[WM_IDN_ON].getUserChoice() == ON);
  }

  /** initialize the variables relevant to coordinating of ACS networks in the
   * associated <tt>Global</tt> object.  */
  private void initGlobalAcsCoordinate()
  {
     final int ELIG = 0, REIN = 1, REIN_ALL = 2, COR_EX = 3, COR_EX_GS = 4;
     Option[] opts = userOptions[ACS_NET][ACS_COR][0][0][0];
     global.setCorMultiExtMode( opts[COR_EX].getUserChoice());
     global.setExternCorConstant(opts[COR_EX_GS].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setGoalCorConstant(opts[COR_EX_GS].getOptionItem(0).getParameter(1).getDoubleValue());
  }

  /** initialize the variables relevant to one ACS network in the associated
   * <tt>Global</tt> object.
   * @param netType external, GS or WM network.
   * @param netIdx the index on the type of networks.
   * @param compIdx the component index on the ACS network.
   * @param setIdx the index on the set of particular component in the ACS network.
   * @param setNum the number of particular component in the ACS network. */
  private void initGlobalAcsNet( int netType, int netIdx, int compIdx, int setIdx, int setNum )
  {
     switch( compIdx ){
       case IDN:
            initGlobalIdn( netType, netIdx );
            break;
       case RER:
            initGlobalRer( netType, netIdx );
            break;
       case IRL:
            initGlobalIrl( netType, netIdx, setIdx, setNum );
            break;
       case FR:
            initGlobalFR( netType, netIdx, setIdx, setNum );
            break;
       default:
            initGlobalAcsNetBasics( netType, netIdx );
            break;
     }
  }

  /** initialize the variables relevant to Goal Structure in the associated
   * <tt>Global</tt> object.  */
  private void initGlobalAcsGS()
  {
     final int GOAL_SET = 0, SIZE_GC = 1, SIZE_G = 2;
     final int TH_GS = 3, DEC_MODE = 4, BLA_OPT = 5;
     Option[] opts = userOptions[ACS_GS][0][0][0][0];
     //global.setGoals( goals );
     global.setGoalChunkSize( opts[SIZE_GC].getOptionItem(0).getParameter(0).getIntValue() );
     // goal structure size.
     global.setGoalStructSize( opts[SIZE_G].getOptionItem(0).getParameter(0).getIntValue() );
     // threshold when checking qualification for participating in the competition of the goal list.
     global.setGoalActivationThreshold( opts[TH_GS].getOptionItem(0).getParameter(0).getDoubleValue() );
     // goal decided by TL or both of levels.
     global.setGoalDecisionMode( opts[DEC_MODE].getUserChoice() );
     // goal BLA option
     global.setGoalBlaOption( opts[BLA_OPT].getUserChoice() );
     // goal BLA.
     if( global.getGoalBlaOption() == global.CONSTANT_OPTION )
         global.setGoalBla( opts[BLA_OPT].getOptionItem(0).getParameter(0).getDoubleValue() );
     else {
         double[] params = new double[ opts[BLA_OPT].getOptionItem(1).getParameters().length ];
         for( int i = 0; i < params.length; i++ )
              params[i] = opts[BLA_OPT].getOptionItem(1).getParameter(i).getDoubleValue();
         global.setGoalBlaParas( params );
     }

     // the following function will be implimented in one of goal structure class.
     /*
     // GS item dim info including GOAL dim.
     int dimNum = opts[ITEM_DIM].getOptionItem(0).getParameter(0).getCount()/2;
     short[][] dimInfos = new short[dimNum][2];
     short[] paramValue = opts[ITEM_DIM].getOptionItem(0).getParameter(0).getIntValues();
     for( int j = 0; j < dimNum; j++ ){
          dimInfos[j][0] = paramValue[2*j];
          dimInfos[j][1] = paramValue[2*j+1];
     }
     global.setGsItemDimInfo( dimInfos );
     */
  }

  /** initialize the variables relevant to Working Memory in the associated
   * <tt>Global</tt> object.  */
  private void initGlobalAcsWM()
  {
     final int SIZE_WM = 0, TH_WM = 1, FLAG_WM = 2, ACT_DEC_MODE = 3;
     final int FLAG_DEC_MODE = 4, BLA_OPT = 5;
     Option[] opts = userOptions[ACS_WM][0][0][0][0];
     int wmSize = opts[SIZE_WM].getOptionItem(0).getParameter(0).getIntValue();
     global.setWmSize( wmSize );
     // initialize the feature: Global.FT_SYS, Global.FT_SYS_WM_ACT_PARA.
     global.setFeatureLength( Global.FT_SYS, Global.FT_SYS_WM_ACT_PARA, wmSize );
     String[] valueNames = new String[ wmSize ];
     for( int i = 0; i < wmSize; i++ )
          valueNames[i] = "wm slot " + i;
     global.setFeatureValueNames( Global.FT_SYS, Global.FT_SYS_WM_ACT_PARA, valueNames );
     global.setWmActivationThreshold( opts[TH_WM].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setWmFlagsOn( opts[FLAG_WM].getUserChoice()==ON );
     if( global.isWmFlagsOn() ){
         int flagNum = opts[FLAG_WM].getOptionItem(ON).getParameter(0).getIntValue();
         global.setWmFlagNum( flagNum );
         // initialize the feature: Global.FT_SYS, Global.FT_SYS_WM_FLAG
         global.setFeatureLength( Global.FT_SYS, Global.FT_SYS_WM_FLAG, flagNum );
         valueNames = new String[ flagNum ];
         for( int i = 0; i < flagNum; i++ )
              valueNames[i] = "wm flag " + i;
         global.setFeatureValueNames( Global.FT_SYS, Global.FT_SYS_WM_FLAG, valueNames );
     }
     global.setWmActionDecisionMode( opts[ACT_DEC_MODE].getUserChoice() );
     global.setWmActionFlagDecisionMode( opts[FLAG_DEC_MODE].getUserChoice() );
     global.setWmBlaOption( opts[BLA_OPT].getUserChoice() );
     // WM BLA.
     if( global.getWmBlaOption() == global.CONSTANT_OPTION )
         global.setWmBla( opts[BLA_OPT].getOptionItem(0).getParameter(0).getDoubleValue() );
     else {
         double[] params = new double[ opts[BLA_OPT].getOptionItem(1).getParameters().length ];
         for( int i = 0; i < params.length; i++ )
              params[i] = opts[BLA_OPT].getOptionItem(1).getParameter(i).getDoubleValue();
         global.setWmBlaParas( params );
     }
  }

  /************* init Global ACS knowledge types **************/
  // 14 userOptions.
  private void initGlobalAcsNetBasics( int netType, int nIdx )
  {
     final int IDN_ON = 0, RER_ON = 1, IRL_ON = 2, FR_ON = 3;
     final int EX_INSTR = 4, IM_LEARN = 5, PLAN = 6, MCS = 7;
     final int DETER = 8, COMB = 9, STOCH = 10, BU = 11, TG = 12;
     final int ELIG = 13;
     Option[] opts = userOptions[ACS_NET][netType][nIdx][COMP][0];
     // convert nIdx.
     int netIdx = this.convertNetIdx( netType, nIdx );
     global.setRerOn( netIdx, opts[RER_ON].getUserChoice()==ON );
     global.setIrlOn( netIdx, opts[IRL_ON].getUserChoice()==ON );
     if( global.isIrlOn( netIdx ) )
         global.setIrlSetNum( netIdx, opts[IRL_ON].getOptionItem(0).getParameter(0).getIntValue());
     global.setFrOn( netIdx, opts[FR_ON].getUserChoice()==ON );
     if( global.isFrOn( netIdx ) )
         global.setFixRuleNum( netIdx, opts[FR_ON].getOptionItem(0).getParameter(0).getIntValue());
     global.setExtInstructOn( netIdx, opts[EX_INSTR].getUserChoice()==ON );
     global.setILOn( netIdx, opts[IM_LEARN].getUserChoice()==ON );
     global.setCombineDeterMode( opts[DETER].getUserChoice() );
     int combMethod1 = opts[COMB].getUserChoice();
     global.setAcsCrossCombineMode( netIdx, 0, combMethod1 );
     int combMethod2 = opts[COMB + 1 + combMethod1].getUserChoice();
     global.setAcsCrossCombineMode( netIdx, 1, combMethod2 );
     int paraNum = opts[COMB + 1 + combMethod1].getOptionItem(combMethod2).getParameters().length;
     if( combMethod1 != Global.STOCH_COMB_MODE )
         paraNum--;
     if( combMethod2 == global.FIX_COMB_MODE ){
         double[] params = new double[ paraNum ];
         for( int i = 0; i < paraNum; i++ )
              params[i] = opts[COMB + 1 + combMethod1].getOptionItem(combMethod2).getParameter(i).getDoubleValue();
         global.setAcsCrossCombineParas( netIdx, params );
       }


     //System.out.println( "in the middle of init net basics");


     if( combMethod2 == global.VAR_COMB_MODE ){
         double[] params = new double[ paraNum-2 ];
         for( int i = 0; i < paraNum-2; i++ )
              params[i] = opts[COMB + 1 + combMethod1].getOptionItem(combMethod2).getParameter(i).getDoubleValue();
         global.setAcsCrossCombineParas( netIdx, params );
         double[] aboves = new double[ Global.ACS_KNOW_TYPE_NUM ];
         for( int i = 0; i < aboves.length; i++ )
              aboves[i] = opts[COMB+1+combMethod1].getOptionItem(combMethod2).getParameter(paraNum-2).getDoubleValue();
         global.setAcsCrossCombineVarParaAboves( netIdx, aboves );
         double[] unders = new double[ Global.ACS_KNOW_TYPE_NUM ];
         for( int i = 0; i < unders.length; i++ )
              unders[i] = opts[COMB+1+combMethod1].getOptionItem(combMethod2).getParameter(paraNum-1).getDoubleValue();
         global.setAcsCrossCombineVarParaUnders( netIdx, unders );
     }
     if( combMethod2 == global.CT_COMB_MODE ){
     }
     if( combMethod1 != Global.STOCH_COMB_MODE )
         global.setOverallActionSelectionTemp( netIdx,
              opts[COMB + 1 + combMethod1].getOptionItem(combMethod2).getParameter(paraNum).getDoubleValue() );
  }

  // 10 userOptions.
  private void initGlobalIdn( int netType, int nIdx )
  {
     final int IN_DIM_INFO = 0, OUT_DIM_INFO = 1;
     final int UNITS = 2, LEARN = 3, L_RATE = 4, MOM = 5, Q_RATE = 6;
     final int TEMP = 7, REIN_EACH = 8, MCS = 9;
     //final int ACT_NUM = 10, ACT_REP = 11, RWD_NUM = 12, RWD_DIM = 13, RWD = 14;

     Option[] opts = userOptions[ACS_NET][netType][nIdx][IDN][0];
     // convert nIdx.
     int netIdx = this.convertNetIdx( netType, nIdx );
     // fill input and output dimension info.
     fillDimInfo( netType, netIdx, IDN, opts );
     // other options.
     global.setIdnHiddenNum(netIdx, opts[UNITS].getOptionItem(0).getParameter(1).getIntValue());
     //global.setExIdnEligibleCondiitons( int netIdx, String value );
     global.setIdnLearningOption( netIdx, opts[LEARN].getUserChoice() );
     global.setIdnBPLearningRate( netIdx,
            opts[L_RATE].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setIdnBPMomentum( netIdx,
            opts[MOM].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setIdnQDiscountRate( netIdx,
            opts[Q_RATE].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setIdnActionSelectionTemp( netIdx,
            opts[TEMP].getOptionItem(0).getParameter(0).getDoubleValue());
     //global.setIdnRLFunction( int netIdx, String value );
     global.setIdnRLFuncByMCS( netIdx, opts[MCS].getUserChoice()==ON );
  }

  // 21 userOptions.
  private void initGlobalRer( int netType, int nIdx )
  {
     final int DENSITY = 13, EXT_PROB = 14, GS_MODE = 15, POS = 16;
     final int IG = 17, TH = 18, RANGE = 19, MERGE = 20;
     Option[] opts = userOptions[ACS_NET][netType][nIdx][RER][0];
     // convert nIdx.
     int netIdx = this.convertNetIdx( netType, nIdx );
     // init ACS rule common options.
     initAcsRules( netType, netIdx, RER, 0, opts );
     // init RER specific options.
     global.setRerDensity( netIdx, opts[DENSITY].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setRerExtractionProb( netIdx, opts[EXT_PROB].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setRerGenSpecMode( netIdx, opts[GS_MODE].getUserChoice());
     //global.setRerPosCriterion( netIdx, String value);
     //global.setRerIGMeasurement( netIdx, String value);
     global.setRerIGParaAbove( netIdx, opts[IG].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setRerIGParaUnder( netIdx, opts[IG].getOptionItem(0).getParameter(1).getDoubleValue());
     global.setRerPosThreshold( netIdx, opts[TH].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setRerGenThreshold( netIdx, opts[TH].getOptionItem(0).getParameter(1).getDoubleValue());
     global.setRerSpecThreshold( netIdx, opts[TH].getOptionItem(0).getParameter(2).getDoubleValue());
     global.setRerGenSpecValueRange( netIdx, opts[RANGE].getUserChoice() );
     global.setRuleMergeInterval( netIdx, opts[MERGE].getOptionItem(0).getParameter(0).getIntValue() );
  }

  // 20 userOptions
  private void initGlobalIrl( int netType, int nIdx, int setIdx, int setNum )
  {
     final int RULE_NUM = 13, RULES = 14, DENSITY = 15;
     final int POS = 16, IG = 17, TH = 18, RANGE = 19;
     Option[] opts = userOptions[ACS_NET][netType][nIdx][IRL][setIdx];
     // convert nIdx.
     int netIdx = this.convertNetIdx( netType, nIdx );
   // init ACS rule common options.
     initAcsRules( netType, netIdx, IRL, setIdx, opts );
     // init IRL specific options.
     global.setIrlRuleNum(netIdx,setIdx,opts[RULE_NUM].getOptionItem(0).getParameter(0).getIntValue());
     //global.setIrlRuleParameters( netIdx, setIdx, params[netIdx][setIdx] );
     global.setIrlDensity( netIdx, opts[DENSITY].getOptionItem(0).getParameter(0).getDoubleValue());
     //global.setIrlPosCriterion( netIdx, String value);
     //global.setIrlIGMeasurement( netIdx, String value);
     global.setIrlIGParaAbove( netIdx, opts[IG].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setIrlIGParaUnder( netIdx, opts[IG].getOptionItem(0).getParameter(1).getDoubleValue());
     // thresholeds
     global.setIrlPosThreshold( netIdx, opts[TH].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setIrlGenThreshold( netIdx, opts[TH].getOptionItem(0).getParameter(1).getDoubleValue());
     global.setIrlSpecThreshold( netIdx, opts[TH].getOptionItem(0).getParameter(2).getDoubleValue());
     global.setIrlDeleteThreshold( netIdx, opts[TH].getOptionItem(0).getParameter(3).getDoubleValue());
     global.setRerGenSpecValueRange( netIdx, opts[RANGE].getUserChoice() );
  }

  // 15 userOptions
  private void initGlobalFR( int netType, int nIdx, int setIdx, int setNum )
  {
     final int RULE_COND = 13, RULE_ACT = 14;
     Option[] opts = userOptions[ACS_NET][netType][nIdx][FR][setIdx];
     // convert nIdx.
     int netIdx = this.convertNetIdx( netType, nIdx );
     // init ACS rule common options.
     initAcsRules( netType, netIdx, FR, setIdx, opts );
     // init FR specific options.
     // global.setFixRuleConditions( netIdx, setIdx, conditions[netIdx][setIdx] );
     // global.setFixRuleAction( netIdx, setIdx, actions[netIdx][setIdx] );
  }

  // 13 OPTIONS.
  public void initAcsRules( int netType, int netIdx, int ruleType, int setIdx, Option[] opts )
  {
     final int IN_DIM_INFO = 0, OUT_DIM_INFO = 1, TEMP = 2;
     final int REASON = 3, UTL_ON = 4, PM_ON = 5, RBLA_ON = 6, CBLA_ON = 7;
     final int WGT_C = 8, WGT_R = 9, RBLA = 10, CBLA = 11, UTL = 12;
     // fill input and output dimension info.
     fillDimInfo( netType, netIdx, ruleType, opts );
     // fill in other options.
     global.setARActionSelectionTemp( netIdx, ruleType, opts[TEMP].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setARReasoningMethod( netIdx, ruleType, opts[REASON].getUserChoice() );
     if( global.getARReasoningMethod( netIdx, ruleType ) == 1 )  // Partial Match
         global.setPMThreshold( netIdx, opts[REASON].getOptionItem(1).getParameter(0).getDoubleValue());
     global.setRuleUtilityOn( netIdx, ruleType, setIdx, opts[UTL_ON].getUserChoice()==ON );
     global.setRulePartMatchOn( netIdx, ruleType, setIdx, opts[PM_ON].getUserChoice()==ON );
     global.setRuleBlaOn( netIdx, ruleType, setIdx, opts[RBLA_ON].getUserChoice()==ON );
     global.setActionChunkBlaOn( netIdx, ruleType, setIdx, opts[CBLA_ON].getUserChoice()==ON );
     global.setRuleConditionWeightMode( netIdx, ruleType, setIdx, opts[WGT_C].getUserChoice() );
     global.setRuleWeightMode( netIdx, ruleType, setIdx, opts[WGT_R].getUserChoice() );

     //if( ruleType == FR )
      //   System.out.println( "before bla" );

     // BLA settings.
     if( global.isRuleBlaOn( netIdx, ruleType, setIdx ) ){
         global.setActionRuleBlaOption( netIdx, ruleType, setIdx, opts[RBLA].getUserChoice());
         if( global.getActionRuleBlaOption( netIdx, ruleType, setIdx ) == global.CONSTANT_OPTION )
             global.setActionRuleBla( netIdx, ruleType, setIdx,
                opts[RBLA].getOptionItem(0).getParameter(0).getDoubleValue());
         else{
             double[] blaParams = new double[ opts[RBLA].getOptionItem(1).getParameters().length ];
             for( int j = 0; j < blaParams.length; j++ )
                  blaParams[j] = opts[RBLA].getOptionItem(1).getParameter(j).getDoubleValue();
             global.setActionRuleBlaParas( netIdx, ruleType, setIdx, blaParams );
         }
     }

     //if( ruleType == FR )
     //    System.out.println( "after RULE bla" );

     if( global.isActionChunkBlaOn( netIdx, ruleType, setIdx ) ){
         global.setAcsChunkBlaOption( netIdx, ruleType, setIdx,
                opts[CBLA].getUserChoice() );
         if( global.getAcsChunkBlaOption( netIdx, ruleType, setIdx ) == global.CONSTANT_OPTION )
             global.setAcsChunkBla( netIdx, ruleType, setIdx,
                opts[CBLA].getOptionItem(0).getParameter(0).getDoubleValue());
         else{
             double[] blaParams = new double[ opts[CBLA].getOptionItem(1).getParameters().length ];
             for( int j = 0; j < blaParams.length; j++ )
                  blaParams[j] = opts[CBLA].getOptionItem(1).getParameter(j).getDoubleValue();
             global.setAcsChunkBlaParas( netIdx, ruleType, setIdx, blaParams );
         }
     }

     //if( ruleType == FR )
     //    System.out.println( "after CHUNK bla" );

     // utility settings.
     if( global.isRuleUtilityOn( netIdx, ruleType, setIdx ) ){
         int optIdx = opts[UTL].getUserChoice();
         global.setRuleUtilityOption( netIdx, ruleType, setIdx, optIdx );
         if( optIdx < 2 ){ // not the user defined
             double[] utlParams = new double[ opts[UTL].getOptionItem(optIdx).getParameters().length ];
             for( int j = 0; j < utlParams.length; j++ ){
                  utlParams[j] = opts[UTL].getOptionItem(optIdx).getParameter(j).getDoubleValue();
                  //System.out.println( "param value = " + utlParams[j] );
             }
             global.setRuleUtilityParas( netIdx, ruleType, setIdx, utlParams );
         }
     }
  }

  private int convertNetIdx( int netType, int nIdx )
  {
     // convert nIdx.
     int netIdx = 0;
     if( netType == this.EX_NET )
         netIdx = nIdx;
     if( netType == this.GS_NET )
         netIdx = global.getGsNetIdx();
     if( netType == this.WM_NET )
         netIdx = global.getWmNetIdx();
     return netIdx;
  }

  private int calcAcsNetCompType( String name )
  {
     int type = COMP;
     for( int i = 0; i < acsNetCompNames.length; i++ )
          if( acsNetCompNames[ i ].equals( name ) ){
              type = i;
              break;
           }
     return type;
  }

  private void fillDimInfo( int netType, int netIdx, int knowType, Option[] opts )
  {
     final int IN_DIM_INFO = 0, OUT_DIM_INFO = 1;
     // input dimension values.
     short dimNum = 0, dimTypes[] = null, dimIdxs[] = null, dimVals[] = null, featInfo[][] = null;
     boolean specBySysFlag = opts[IN_DIM_INFO].getOptionItem(0).getParameter(Global.DIM_SPEC_SYS).getBoolValue();
     if( !specBySysFlag ){
         dimNum = (short)opts[IN_DIM_INFO].getOptionItem(0).getParameter(Global.DIM_SPEC_INDEX).getCount();
         short[] dTypes = opts[IN_DIM_INFO].getOptionItem(0).getParameter(Global.DIM_SPEC_TYPE).getIntValues();
         dimTypes = new short[dimNum];
         System.arraycopy( dTypes, 0, dimTypes, 0, Math.min(dimTypes.length, dTypes.length) );
         if(dTypes.length < dimNum) {
              for(int i = dTypes.length; i < dimTypes.length; i++ )
                    dimTypes[i] = dTypes[dTypes.length-1];
         }
         dimIdxs = opts[IN_DIM_INFO].getOptionItem(0).getParameter(Global.DIM_SPEC_INDEX).getIntValues();
         dimVals = new short[ dimNum ];
         featInfo = new short[ dimNum ][2];
         for( int i = 0; i < dimNum; i++ ){
              dimVals[i] = (short)global.getFeatureLength(dimTypes[i], dimIdxs[i]);
              featInfo[i][0] = dimTypes[i];
              featInfo[i][1] = dimIdxs[i];
         }
         if( knowType == IDN ){
             global.setIdnInputDimVals( netIdx, dimVals );
             global.setIdnInputFeatureInfo( netIdx, featInfo );
         }
         else{
             global.setARInputDimVals( netIdx, knowType, dimVals );
             global.setARInputFeatureInfo( netIdx, knowType, featInfo );
         }
     }
     else{
         if( netType == Global.GS_NET ){
             dimNum = global.getFeatureNums()[Global.FT_GS];
             dimVals = new short[ dimNum ];
             featInfo = new short[ dimNum ][2];
             for( int i = 0; i < dimNum; i++ ){
                  dimVals[i] = (short)global.getFeatureLength(Global.FT_GS, i);
                  featInfo[i][0] = Global.FT_GS;
                  featInfo[i][1] = (short)i;
             }
             if( knowType == IDN ){
                 global.setIdnInputDimVals( netIdx, dimVals );
                 global.setIdnInputFeatureInfo( netIdx, featInfo );
             }
             else{
                global.setARInputDimVals( netIdx, knowType, dimVals );
                global.setARInputFeatureInfo( netIdx, knowType, featInfo );
             }
         }
         if( netType == Global.WM_NET )
             setupGenericFormat( WM, INPUT, netIdx, opts[IN_DIM_INFO] );
     }
     // output dimension values.
     specBySysFlag = opts[OUT_DIM_INFO].getOptionItem(0).getParameter(Global.DIM_SPEC_SYS).getBoolValue();
     if( !specBySysFlag ){
         dimNum = (short)opts[OUT_DIM_INFO].getOptionItem(0).getParameter(Global.DIM_SPEC_INDEX).getCount();
         short[] dTypes = opts[OUT_DIM_INFO].getOptionItem(0).getParameter(Global.DIM_SPEC_TYPE).getIntValues();
         dimTypes = new short[dimNum];
         System.arraycopy( dTypes, 0, dimTypes, 0, Math.min(dimTypes.length, dTypes.length) );
         if(dTypes.length < dimNum) {
              for(int i = dTypes.length; i < dimTypes.length; i++ )
                    dimTypes[i] = dTypes[dTypes.length-1];
         }
         dimIdxs = opts[OUT_DIM_INFO].getOptionItem(0).getParameter(Global.DIM_SPEC_INDEX).getIntValues();
         dimVals = new short[ dimNum ];
         featInfo = new short[ dimNum ][2];
     }
     else{
         if( netType == Global.GS_NET ){
             dimNum = (short)(global.getFeatureNums()[Global.FT_GS] + 1);
             dimTypes = new short[ dimNum ];
             dimIdxs = new short[ dimNum ];
             dimVals = new short[ dimNum ];
             featInfo = new short[ dimNum ][2];
             dimTypes[0] = Global.FT_SYS;
             dimIdxs[0] = Global.FT_SYS_GS_ACT;
             for( short i = 1; i < dimNum; i++ ){
                  dimTypes[i] = Global.FT_GS;
                  dimIdxs[i] = (short)(i-1);
             }
         }
         if( netType == Global.WM_NET ){
             dimNum = Global.ACT_DIM_NUM_WM;
             dimTypes = new short[ dimNum ];
             dimIdxs = new short[ dimNum ];
             dimVals = new short[ dimNum ];
             featInfo = new short[ dimNum ][2];
             for( short i = 0; i < dimNum; i++ ){
                  dimTypes[i] = Global.FT_SYS;
                  dimIdxs[i] = (short)(Global.FT_SYS_WM_ACT + i);
             }
         }
     }
     for( int i = 0; i < dimNum; i++ ){
          dimVals[i] = (short)global.getFeatureLength(dimTypes[i], dimIdxs[i]);
          featInfo[i][0] = dimTypes[i];
          featInfo[i][1] = dimIdxs[i];
     }
     if( knowType == IDN ){
         global.setIdnOutputDimVals( netIdx, dimVals );
         global.setIdnOutputFeatureInfo( netIdx, featInfo );
     }
     else{
         global.setAROutputDimVals( netIdx, knowType, dimVals );
         global.setAROutputFeatureInfo( netIdx, knowType, featInfo );
     }
  }
}
