package clarion.options.transfer.subsys;
import java.io.*;
import clarion.system.*;
import clarion.options.*;
import clarion.options.transfer.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the process of transferring NACS configuration data
 * between GUI and <tt>Global</tt> object or persistent storage.
 * <li> important methods <br>
 * <ol>
 * <li> public void load(): implement the process of loading options from persistent
 * storage.
 * <li> public void save(): implement the process of save the options to persistent
 * storage or to the associated <tt>Global</tt> object.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model. <br>
 * This class is part of data transfer subsystem and is not related directly to
 * CLARION model.
 * <li> Usage of this class. <br>
 * It extends the base <tt>Transfer</t>> class and is instantiated in the class
 * <em>TransferManager</em>.
 * </ul>
 * @author Xi Zhang
 *
 **/

public class NacsTransfer extends Transfer {

  /** constants indicate different components in NACS subsystem. */
  protected final int NACS_COMP_NUM = 6;
  protected final int NACS_COMP = 0, NACS_GKS = 1, NACS_AMN = 2, NACS_EM = 3, NACS_AEM = 4, NACS_COR = 5;

  /** constructor. instantiated with the associated <tt>TransferManager</tt> object
   * and <tt>Global</tt> object. */
  public NacsTransfer( TransferManager manager, Global global )
  {
     super( manager, global );
  }

  /** Get the GKS option to check if the GKS component is selected or not.
   *  @return the GKS option. */
  public boolean isGksOn()
  {
     return userOptions[NACS_COMP][0][0][0][0][0].getUserChoice()==ON;
  }

  /** Get the number of AMN networks in the NACS subystem.
   *  @return the number of AMN networks in the NACS subystem. */
  public int getAmnNum()
  {
     return userOptions[NACS_COMP][0][0][0][0][1].getOptionItem(0).getParameter(0).getIntValue();
  }

  /** Get the EM option to check if the EM component is selected or not.
   *  @return the EM option. */
  public boolean isEmOn()
  {
     return userOptions[NACS_COMP][0][0][0][0][2].getUserChoice()==ON;
  }

  /** Get the AEM option to check if the AEM component is selected or not.
   *  @return the AEM option. */
  public boolean isAemOn()
  {
     return userOptions[NACS_COMP][0][0][0][0][3].getUserChoice()==ON;
  }

  /** Get the type of an AMN network in the NACS subystem.
   *  @return the type of an AMN network in the NACS subystem.
   *  @deprecated */
  public int getAmnType( int netIdx )
  {
     return userOptions[NACS_AMN][netIdx][0][0][0][0].getUserChoice();
  }

  /** Get the GKS reasoning method.
   *  @return the GKS reasoning method. */
  public int getGksReasoningType()
  {
     return userOptions[NACS_COR][0][0][0][0][1].getUserChoice();
  }

  /** the process of loading configurations from persistent storage. */
  public void load()
  {
     // load NACS subsystem options.
     sysOptions = new Option[ NACS_COMP_NUM ][][][][][];
     userOptions = new Option[ NACS_COMP_NUM ][][][][][];
     subsysPath = this.searchFile( "nacs" );
     loadNacsCompSettings();
     loadNacsAmnSettings();
     loadNacsGksSettings();
     loadNacsEmSettings();
     loadNacsAemSettings();
     loadNacsCoordinateSettings();
  }

  /** the process of saving the options to persistent storage or to the
   *  associated <tt>Global</tt> object. */
  public void save()
  {
     // create NACS setting directory.
     if( saveCode == SAVE_TO_FILE ){
         String path = global.getSettingDir() + "/" + global.getTaskName()
                               + "/" + global.getTaskSettingName() + "/nacs";
         File nacsDir = new File( path );
         nacsDir.mkdir();
     }
     saveNacsCompSettings();
     saveNacsAmnSettings();
     saveNacsGksSettings();
     if( manager.isEmOn() )
         saveNacsEmSettings();
     if( manager.isAemOn() )
         saveNacsAemSettings();
     //saveNacsCoordinateSettings();
  }

  /** load configuration for selection of NACS components. */
  private void loadNacsCompSettings()
  {
     loadedFileName = "components.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/nacs/";
     sysOptions[NACS_COMP] = new Option[1][1][1][1][];
     sysOptions[NACS_COMP][0][0][0][0] = loadOptions( path );
     //path = global.getSettingDir() + "/" + global.getTaskName()
     //              + "/" + global.getTaskSettingName() + "/nacs/";
     userOptions[NACS_COMP] = new Option[1][1][1][1][];
     userOptions[NACS_COMP][0][0][0][0] = loadOptions( subsysPath );
  }

  /** load configuration for GKS. */
  private void loadNacsGksSettings()
  {
     loadedFileName = "gks.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/nacs/";
     sysOptions[NACS_GKS] = new Option[1][1][1][1][];
     sysOptions[NACS_GKS][0][0][0][0] = loadOptions( path );
     //path = global.getSettingDir() + "/" + global.getTaskName()
     //              + "/" + global.getTaskSettingName() + "/nacs/";
     userOptions[NACS_GKS] = new Option[1][1][1][1][];
     userOptions[NACS_GKS][0][0][0][0] = loadOptions( subsysPath );
  }

  /** load configuration for AMN networks. */
  private void loadNacsAmnSettings()
  {
     loadedFileName = "amn.txt";
     // fill in system options
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/nacs/amns";
     int netNum = new File( path ).list().length;
     sysOptions[NACS_AMN] = new Option[netNum][1][1][1][];
     for( int i = 0; i < netNum; i++ )
          sysOptions[NACS_AMN][i][0][0][0] = loadOptions( path + "/amn" + i + "/" );
     // fill in user options
     path = subsysPath + "amns";
     netNum = new File( path ).list().length;
     userOptions[NACS_AMN] = new Option[Global.MAX_AMN_NUM][1][1][1][];
     for( int i = 0; i < Global.MAX_AMN_NUM; i++ ){
          if( i < netNum )
              userOptions[NACS_AMN][i][0][0][0] = loadOptions( path + "/amn" + i + "/" );
          else{
              userOptions[NACS_AMN][i] = new Option[1][1][1][userOptions[NACS_AMN][0][0][0][0].length];
              for( int j = 0; j < userOptions[NACS_AMN][0][0][0][0].length; j++ )
                   userOptions[NACS_AMN][i][0][0][0][j] =
                       (Option)(userOptions[NACS_AMN][0][0][0][0][j].clone());
          }
     }
  }

  /** load configuration for EM component. */
  private void loadNacsEmSettings()
  {
     loadedFileName = "em.txt";
     // load EM settings.
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/nacs/";
     sysOptions[NACS_EM] = new Option[1][1][1][1][];
     sysOptions[NACS_EM][0][0][0][0] = loadOptions( path );
     path = searchFile( "nacs", loadedFileName );
     userOptions[NACS_EM] = new Option[1][1][1][1][];
     userOptions[NACS_EM][0][0][0][0] = loadOptions( path );

     //System.out.println( "path = " + path );
  }

  /** load configuration for AEM component. */
  private void loadNacsAemSettings()
  {
     loadedFileName = "aem.txt";
     // load AEM settings.
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/nacs/";
     sysOptions[NACS_AEM] = new Option[1][1][1][1][];
     sysOptions[NACS_AEM][0][0][0][0] = loadOptions( path );
     path = searchFile( "nacs", loadedFileName );
     userOptions[NACS_AEM] = new Option[1][1][1][1][];
     userOptions[NACS_AEM][0][0][0][0] = loadOptions( path );

     //System.out.println( "path = " + path );

  }

  /** load configuration for coordinating ACS and NACS. */
  private void loadNacsCoordinateSettings()
  {
     loadedFileName = "coordinate.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/nacs/";
     sysOptions[NACS_COR] = new Option[1][1][1][1][];
     sysOptions[NACS_COR][0][0][0][0] = loadOptions( path );
     //path = global.getSettingDir() + "/" + global.getTaskName()
     //              + "/" + global.getTaskSettingName() + "/nacs/";
     userOptions[NACS_COR] = new Option[1][1][1][1][];
     userOptions[NACS_COR][0][0][0][0] = loadOptions( subsysPath );
  }

  /** save configuration for selection of NACS components. */
  private void saveNacsCompSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalNacsComponents();
         return;
      }
     // create NACS components setting file.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/nacs/";
     savedFileName = "components.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save NACS components setting.
     saveOptions( path, userOptions[NACS_COMP][0][0][0][0] );
  }

  /** save configuration for AMN networks. */
  private void saveNacsAmnSettings()
  {
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/nacs/amns";
     if( saveCode == SAVE_TO_FILE ){
         new File( path ).mkdir();
     }
     int netNum = manager.getAmnNum();
     // initialize global.java with netNum info.
     if( saveCode == SAVE_TO_GLOBAL ){
         global.setAmnNum( netNum );
         global.initNacsNets();
     }
     try{
       for( int i = 0; i < netNum; i++ ){
            if( saveCode == SAVE_TO_FILE ){
                File file1 = new File( path + "/amn" + i );
                file1.mkdir();
                File file2 = new File( path + "/amn" + i + "/" + "amn.txt" );
                file2.createNewFile();
                savedFileName = "amn.txt";
                saveOptions( path + "/amn" + i + "/", userOptions[NACS_AMN][i][0][0][0] );
            }
            else
                initGlobalAmn( i );
       }
     }
     catch( Exception e ){
       System.out.println( e );
     }
  }

  /** save configuration for GKS component. */
  private void saveNacsGksSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalGks();
         return;
     }
     // create NACS gks setting file.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/nacs/";
     savedFileName = "gks.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save NACS gks setting.
     saveOptions( path, userOptions[NACS_GKS][0][0][0][0] );
  }

  /** save configuration for EM component. */
  private void saveNacsEmSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalEM();
         return;
     }
     // create NACS EM setting file.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/nacs/";
     savedFileName = "em.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save NACS gks setting.
     saveOptions( path, userOptions[NACS_EM][0][0][0][0] );
  }

  /** save configuration for AEM component. */
  private void saveNacsAemSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalAem();
         return;
     }
     // create NACS aem setting file.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/nacs/";
     savedFileName = "aem.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save NACS gks setting.
     saveOptions( path, userOptions[NACS_AEM][0][0][0][0] );
  }

  /** save configuration for coordinating of ACS and NACS. */
  private void saveNacsCoordinateSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalNacsCoordinate();
         return;
     }
     // create NACS coordinate setting file.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/nacs/";
     savedFileName = "coordinate.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save NACS coordinate setting.
     saveOptions( path, userOptions[NACS_COR][0][0][0][0] );
  }

  /** initialize the variables relevant to selection of NACS components in the
   * associated <tt>Global</tt> object  */
  private void initGlobalNacsComponents()
  {
     final int GKS = 0, AMN = 1, EM = 2, AEM = 3;
     Option[] opts = userOptions[NACS_COMP][0][0][0][0];
     global.setGKSon( opts[GKS].getUserChoice()==ON );
     int amnNum = opts[AMN].getOptionItem(0).getParameter(0).getIntValue();
     global.setAmnNum( amnNum );
     // initialize the feature: Global.FT_SYS, Global.FT_SYS_NACS_AMNS
     global.setFeatureLength( Global.FT_SYS, Global.FT_SYS_NACS_AMNS, amnNum );
     String[] valueNames = new String[ amnNum ];
     for( int i = 0; i < amnNum; i++ )
          valueNames[i] = "GKS chunk " + i;
     global.setFeatureValueNames( Global.FT_SYS, Global.FT_SYS_NACS_AMNS, valueNames );
     global.setEMon( opts[EM].getUserChoice()==ON );
     global.setAEMon( opts[AEM].getUserChoice()==ON );
  }

  /** initialize the variables relevant to GKS in the associated <tt>Global</tt> object  */
  private void initGlobalGks()
  {
     final int IN_DIM_INFO = 0, OUT_DIM_INFO = 1, TEMP = 2, ESC = 3, SIM_ON = 4;
     final int RBLA_ON = 5, CBLA_ON = 6, WGT_C = 7, WGT_R = 8, RBLA = 9, CBLA = 10;
     final int SBR = 11, SBR_TH = 12, SBR_PARAM = 13, RENC_ON = 14, CENC_ON = 15;
     final int REXT_ON = 16, CEXT_ON = 17, RENC_PROB = 18, CENC_PROB = 19, RDEN = 20;
     final int CDEN = 21, REXT_TH = 22, CEXT_TH = 23, REXT_PROB = 24, CEXT_PROB = 25;
     final int REASON_GKS = 26, REASON_ITER = 27, REASON_TH = 28;
     Option[] opts = userOptions[NACS_GKS][0][0][0][0];
     // input dimension values.
     this.setupGenericFormat( GKS, INPUT, 0, opts[IN_DIM_INFO] );
     // output dimension values.
     this.setupGenericFormat( GKS, OUTPUT, 0, opts[OUT_DIM_INFO] );
     // maximum # of chunks in GKS.
     // global.setGksMaxChunkNum( 200 );
     // initialize the feature: Global.FT_SYS, Global.FT_SYS_NACS_CHKS
     int maxChunkNum = global.getGksMaxChunkNum();
     global.setFeatureLength( Global.FT_SYS, Global.FT_SYS_NACS_CHKS, maxChunkNum );
     String[] valueNames = new String[ maxChunkNum ];
     for( int i = 0; i < maxChunkNum; i++ )
          valueNames[i] = "GKS chunk " + i;
     global.setFeatureValueNames( Global.FT_SYS, Global.FT_SYS_NACS_CHKS, valueNames );
     // temperature
     global.setGksOutputTemperature( opts[TEMP].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setESChunkOn( opts[ESC].getUserChoice()==ON );
     global.setSimilarityOn( opts[SIM_ON].getUserChoice()==ON );
     global.setAssocBlaOn( opts[RBLA_ON].getUserChoice()==ON);
     global.setNacsChunkBlaOn( opts[CBLA_ON].getUserChoice()==ON );
     global.setAssocConditionWeightMode( opts[WGT_C].getUserChoice() );
     global.setAssocDimensionWeightMode( opts[WGT_R].getUserChoice() );
     /*
     // utility settings.
     if( global.isAssocUtilityOn() ){
         int optIdx = opts[UTL].getUserChoice();
         global.setAssocUtilityOption( optIdx );
         if( optIdx < 2 ){ // not the user defined
             double[] utlParams = new double[ opts[UTL].getOptionItem(optIdx).getParameters().length ];
             for( int j = 0; j < utlParams.length; j++ )
                  utlParams[j] = opts[UTL].getOptionItem(optIdx).getParameter(j).getDoubleValue();
             global.setAssocUtilityParas( utlParams );
        }
     }
     */
     // BLA settings.
     if( global.isAssocBlaOn() ){
         global.setAssocBlaOption( opts[RBLA].getUserChoice() );
         if( global.getAssocBlaOption() == global.CONSTANT_OPTION )
             global.setAssocBla( opts[RBLA].getOptionItem(0).getParameter(0).getDoubleValue() );
         else {
           double[] params = new double[ opts[RBLA].getOptionItem(1).getParameters().length ];
           for( int i = 0; i < params.length; i++ )
                params[i] = opts[RBLA].getOptionItem(1).getParameter(i).getDoubleValue();
           global.setAssocBlaParas( params );
         }
     }
     if( global.isNacsChunkBlaOn() ){
         global.setChunkBlaOption( opts[CBLA].getUserChoice() );
         if( global.getChunkBlaOption() == global.CONSTANT_OPTION )
             global.setChunkBla( opts[CBLA].getOptionItem(0).getParameter(0).getDoubleValue() );
         else {
           double[] params = new double[ opts[CBLA].getOptionItem(1).getParameters().length ];
           for( int i = 0; i < params.length; i++ )
                params[i] = opts[CBLA].getOptionItem(1).getParameter(i).getDoubleValue();
           global.setChunkBlaParas( params );
         }
     }
     global.setThreshold_sim( opts[SBR_TH].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setRBR_BalConstant( opts[SBR_PARAM].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setSBR_BalConstant( opts[SBR_PARAM].getOptionItem(0).getParameter(1).getDoubleValue() );
    /*********************** SETTINGS ABOUT GKS LEARNINGS ************************/
     global.setAssocExtractOn( opts[REXT_ON].getUserChoice()==ON );
     global.setAssocEncodeExternOn( opts[RENC_ON].getUserChoice()==ON );
     global.setChunkEncodeExternOn( opts[CENC_ON].getUserChoice()==ON );
     global.setChunkFromAmnOn( opts[CEXT_ON].getUserChoice()==ON );
     global.setChunkEncodeProb( opts[CENC_PROB].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setAssocEncodeProb( opts[RENC_PROB].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setChunkDensity( opts[CDEN].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setAssocDensity( opts[RDEN].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setAssocExtThreshold( opts[REXT_TH].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setChunkExtThreshold( opts[CEXT_TH].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setAssocExtProb( opts[REXT_PROB].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setChunkExtProb( opts[CEXT_PROB].getOptionItem(0).getParameter(0).getDoubleValue() );
     //global.setActivationMeasure( opts[TH_MEASURE].getUserChoice() );
     global.setGksReasoningMethod( opts[REASON_GKS].getUserChoice() );
     global.setGksReasonIterNum( opts[REASON_ITER].getOptionItem(0).getParameter(0).getIntValue() );
     global.setReasonThreshold( opts[REASON_TH].getOptionItem(0).getParameter(0).getDoubleValue() );
  }

  /** initialize the variables relevant to one AMN network in the associated
   * <tt>Global</tt> object
   * @param amnIdx the network index. */
  private void initGlobalAmn( int amnIdx )
  {
     final int IN_DIM_INFO = 0, OUT_DIM_INFO = 1, UNITS = 2, PASS = 3, L_RATE = 4;
     final int MOM = 5, TRAIN_PROB = 6, ITEM_NUM = 7, ITER_NUM = 8, ITEM_TYPE = 9;

     Option[] opts = userOptions[NACS_AMN][amnIdx][0][0][0];
     // input dimension values.
     this.setupGenericFormat( AMN, INPUT, amnIdx, opts[IN_DIM_INFO] );
     // output dimension values.
     this.setupGenericFormat( AMN, OUTPUT, amnIdx, opts[OUT_DIM_INFO] );
     // network unit numbers.
     global.setAmnHiddenNum(amnIdx, opts[UNITS].getOptionItem(0).getParameter(1).getIntValue());
     global.setAmnMultiPassOn( amnIdx, opts[PASS].getUserChoice()==ON );
     if( global.isAmnMultiPassOn(amnIdx) )
         global.setAmnPassNum( amnIdx, opts[PASS].getOptionItem(0).getParameter(0).getIntValue());
     else
         global.setAmnPassNum( amnIdx, 1 );
     // learning rate and momentum settings.
     global.setAmnBPLearningRate( amnIdx, opts[L_RATE].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setAmnBPMomentum( amnIdx,
            opts[MOM].getOptionItem(0).getParameter(0).getDoubleValue() );
     // settings for AMN training.
     global.setAmnTrainedByEmProb(opts[TRAIN_PROB].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setAmnTrainedByEmItemNum(opts[ITEM_NUM].getOptionItem(0).getParameter(0).getIntValue());
     global.setAmnTrainedByEmIterNum(opts[ITER_NUM].getOptionItem(0).getParameter(0).getIntValue());
     global.setAmnTrainedByEmItemType(opts[ITEM_TYPE].getUserChoice());
  }

  /** initialize the variables relevant to EM in the associated <tt>Global</tt>
   * object  */
  private void initGlobalEM()
  {
     final int EM_TH = 0, EM_PROB = 1, EM_ITEM = 2, EM_ITER = 3, ABLA = 4, NBLA = 5;

     Option[] opts = userOptions[NACS_EM][0][0][0][0];
     // EM BLA settings.
     short[] blaOptValues = new short[ 2 ];
     blaOptValues[0] = (short)opts[ABLA].getUserChoice();
     blaOptValues[1] = (short)opts[NBLA].getUserChoice();
     global.setEmBlaOptions( blaOptValues );
     if( global.getEmBlaOption(0) == global.CONSTANT_OPTION &&
         global.getEmBlaOption(1) == global.CONSTANT_OPTION ){
         double[] blas = new double[ 2 ];
         blas[0] = opts[ABLA].getOptionItem(0).getParameter(0).getDoubleValue();
         blas[1] = opts[NBLA].getOptionItem(0).getParameter(0).getDoubleValue();
         global.setEmBlas( blas );
     }
     else{
         double[][] blaParams = new double[2][];
         if( global.getEmBlaOption(0) != global.CONSTANT_OPTION ){
             blaParams[0] = new double[ opts[ABLA].getOptionItem(1).getParameters().length ];
             for( int i = 0; i < blaParams[0].length; i++ )
                  blaParams[0][i] = opts[ABLA].getOptionItem(1).getParameter(i).getDoubleValue();
         }
         if( global.getEmBlaOption(1) != global.CONSTANT_OPTION ){
             blaParams[1] = new double[ opts[NBLA].getOptionItem(1).getParameters().length ];
             for( int i = 0; i < blaParams[1].length; i++ )
                  blaParams[1][i] = opts[NBLA].getOptionItem(1).getParameter(i).getDoubleValue();
         }
         global.setEmBlaParas( blaParams );
     }
     // EM other settings.
     global.setThreshold_EM( opts[EM_TH].getOptionItem(0).getParameter(0).getDoubleValue() );
     //global.setEmCapacity( int value );
     global.setEmTrainAcsProb( opts[EM_PROB].getOptionItem(0).getParameter(0).getDoubleValue() );
     global.setEmTrainAcsItemNum( opts[EM_ITEM].getOptionItem(0).getParameter(0).getIntValue() );
     global.setEmTrainAcsIterNum( opts[EM_ITER].getOptionItem(0).getParameter(0).getIntValue() );
  }

  /** initialize the variables relevant to AEM in the associated <tt>Global</tt>
   * object  */
  private void initGlobalAem()
  {
     final int AEM_STATE_BOUND = 0, AEM_ACTION_BOUND = 1, AEM_NS_NUM = 2, AEM_QUAN = 3;
     final int AEM_EM_PROB = 4, AEM_EM_ITEM = 5, AEM_EM_ITER = 6, AEM_EM_LR = 7;
     final int AEM_EM_MOM = 8, AEM_ACS_PROB = 9, AEM_ACS_ITEM = 10;
     final int AEM_ACS_ITER = 11, AEM_ACS_TEMP = 12;

     Option[] opts = userOptions[NACS_AEM][0][0][0][0];

     global.setAemStateNumBound(opts[AEM_STATE_BOUND].getOptionItem(0).getParameter(0).getIntValue());
     global.setAemActionNumBound(opts[AEM_ACTION_BOUND].getOptionItem(0).getParameter(0).getIntValue());
     global.setAemPotentialNextStateNum(opts[AEM_NS_NUM].getOptionItem(0).getParameter(0).getIntValue());

     global.setAemReinRangeLowerBound(opts[AEM_QUAN].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setAemReinRangeUpperBound(opts[AEM_QUAN].getOptionItem(0).getParameter(1).getDoubleValue());
     global.setAemReinIncrement(opts[AEM_QUAN].getOptionItem(0).getParameter(2).getDoubleValue());

     global.setAemTrainedByEmProb(opts[AEM_EM_PROB].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setAemTrainedByEmItemNum(opts[AEM_EM_ITEM].getOptionItem(0).getParameter(0).getIntValue());
     global.setAemTrainedByEmIterNum(opts[AEM_EM_ITER].getOptionItem(0).getParameter(0).getIntValue());

     global.setAemTrainedByEmBPLearningRate(opts[AEM_EM_LR].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setAemTrainedByEmBPMomentum(opts[AEM_EM_MOM].getOptionItem(0).getParameter(0).getDoubleValue());

     global.setAemTrainAcsProb(opts[AEM_ACS_PROB].getOptionItem(0).getParameter(0).getDoubleValue());
     global.setAemTrainAcsSampleNum(opts[AEM_ACS_ITEM].getOptionItem(0).getParameter(0).getIntValue());
     global.setAemTrainAcsIterNum(opts[AEM_ACS_ITER].getOptionItem(0).getParameter(0).getIntValue());
     global.setAemTrainAcsOutputTemp(opts[AEM_ACS_TEMP].getOptionItem(0).getParameter(0).getDoubleValue());
  }

  /** initialize the variables relevant to coordinating of ACS and NACS in the
   * associated <tt>Global</tt> object  */
  private void initGlobalNacsCoordinate()
  {
     //final int COR = 0, GKS = 1, ITER = 2, TH = 3;
     final int COR = 0;
     Option[] opts = userOptions[NACS_COR][0][0][0][0];
     global.setCorAcsNacsType( opts[COR].getUserChoice() );
  }
}
