package clarion.options.transfer.subsys;
import java.io.*;
import clarion.system.*;
import clarion.options.*;
import clarion.options.transfer.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement the process of transferring Supervisory Subsystem
*  configuration data between GUI and <tt>Global</tt> object or persistent storage.
 * <li> important methods <br>
 * <ol>
 * <li> public void load(): implement the process of loading options from persistent
 * storage.
 * <li> public void save(): implement the process of save the options to persistent
 * storage or to the associated <tt>Global</tt> object.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model. <br>
 * This class is part of data transfer subsystem and is not related directly to
 * CLARION model.
 * <li> Usage of this class. <br>
 * It extends the base <tt>Transfer</t>> class and is instantiated in the class
 * <em>TransferManager</em>.
 * </ul>
 * @author Xi Zhang
 *
 **/

public class SuperTransfer extends Transfer {

  protected final int SUPER_COMP_NUM = 3;
  protected final int SUPER_COMP = 0, SUPER_MS = 1, SUPER_MCS = 2;

  /** constructor. instantiated with the associated <tt>TransferManager</tt> object
   * and <tt>Global</tt> object. */
  public SuperTransfer( TransferManager manager, Global global )
  {
     super( manager, global );
  }

  /** Get the MS option to check if the MS component is selected or not in supervisory
   * subsystem.
   * @return the MS option. */
  public boolean isMsOn()
  {
     return userOptions[SUPER_COMP][0][0][0][0][0].getUserChoice() == ON;
  }

  /** Get the MCS option to check if the MCS component is selected or not in supervisory
   * subsystem.
   * @return the MCS option. */
  public boolean isMcsOn()
  {
     return userOptions[SUPER_COMP][0][0][0][0][1].getUserChoice() == ON;
  }

  /** implement the process of loading options from persistent storage. */
  public void load()
  {
     // load NACS subsystem options.
     sysOptions = new Option[ SUPER_COMP_NUM ][][][][][];
     userOptions = new Option[ SUPER_COMP_NUM ][][][][][];
     subsysPath = this.searchFile( "super" );
     loadSuperCompSettings();
     loadMsSettings();
     loadMcsSettings();
  }

  /** implement the process of save the options to persistent storage or to the
   * associated <tt>Global</tt> object. */
  public void save()
  {
     // create supervisory setting directory.
     if( saveCode == SAVE_TO_FILE ){
         String path = global.getSettingDir() + "/" + global.getTaskName()
                               + "/" + global.getTaskSettingName() + "/super";
         File superDir = new File( path );
         superDir.mkdir();
     }
     saveSuperCompSettings();
     if( manager.isMsOn() )
         saveMsSettings();
     if( manager.isMcsOn() )
         saveMcsSettings();
  }

 /** load configuration for selection of Supervisory Subsystem components. */
  private void loadSuperCompSettings()
  {
     loadedFileName = "components.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/super/";
     sysOptions[SUPER_COMP] = new Option[1][1][1][1][];
     sysOptions[SUPER_COMP][0][0][0][0] = loadOptions( path );
     userOptions[SUPER_COMP] = new Option[1][1][1][1][];
     userOptions[SUPER_COMP][0][0][0][0] = loadOptions( subsysPath );
  }

  /** load configuration for Motivational Subsystem component. */
  private void loadMsSettings()
  {
     loadedFileName = "ms.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/super/";
     sysOptions[SUPER_MS] = new Option[1][1][1][1][];
     sysOptions[SUPER_MS][0][0][0][0] = loadOptions( path );
     path = searchFile( "super", loadedFileName );
     userOptions[SUPER_MS] = new Option[1][1][1][1][];
     userOptions[SUPER_MS][0][0][0][0] = loadOptions( path );
  }

  /** load configuration for Meta-Cognitive Subsystem component. */
  private void loadMcsSettings()
  {
     loadedFileName = "mcs.txt";
     String path = global.getSettingDir() + "/DEFAULT/subjectGroup0/super/";
     sysOptions[SUPER_MCS] = new Option[1][1][1][1][];
     sysOptions[SUPER_MCS][0][0][0][0] = loadOptions( path );
     path = searchFile( "super", loadedFileName );
     userOptions[SUPER_MCS] = new Option[1][1][1][1][];
     userOptions[SUPER_MCS][0][0][0][0] = loadOptions( path );
  }

  /** save configuration for selection of Supervisory components. */
  private void saveSuperCompSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalSuperComp();
         return;
     }
     // create supervisory setting directory.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/super";
     savedFileName = "/supervisory.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save supervisory setting.
     saveOptions( path, userOptions[SUPER_COMP][0][0][0][0] );
  }

   /** save configuration for Motivational Subsystem. */
  private void saveMsSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalMS();
         return;
     }
     // create NACS gks setting file.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/super/";
     savedFileName = "ms.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save NACS gks setting.
     saveOptions( path, userOptions[SUPER_MS][0][0][0][0] );
  }

  /** save configuration for Meta-Cognitive Subsystem. */
  private void saveMcsSettings()
  {
     if( saveCode == SAVE_TO_GLOBAL ){
         initGlobalMCS();
         return;
     }
     // create NACS gks setting file.
     String path = global.getSettingDir() + "/" + global.getTaskName()
                   + "/" + global.getTaskSettingName() + "/super/";
     savedFileName = "mcs.txt";
     try{
        new File( path + savedFileName ).createNewFile();
     }
     catch( Exception e ){
        System.out.println( e );
     }
     // save NACS gks setting.
     saveOptions( path, userOptions[SUPER_MCS][0][0][0][0] );
  }

  /************ init Global Supervisory Subsystem *************/
  /** initialize the variables relevant to selection of Supervisory components in
   * the associated  <tt>Global</tt> object. */
  private void initGlobalSuperComp()
  {
    final int MS = 0, MCS = 1;
    Option[] opts = userOptions[SUPER_COMP][0][0][0][0];
    global.setMsOn( opts[MS].getUserChoice() == ON );
    global.setMcsOn( opts[MCS].getUserChoice() == ON );
  }

  /** initialize the variables relevant to Motivational Subsystem in the associated
   * <tt>Global</tt> object. */
  private void initGlobalMS()
  {
    final int DRIVE = 0, DRIVE_REP = 1, DRIVE_NET = 2;
    Option[] opts = userOptions[SUPER_MS][0][0][0][0];
    //global.setDrives( String value );
    //global.setDriveReps( String value );
  }

  /** initialize the variables relevant to Meta-Cognitive Subsystem in the associated
   * <tt>Global</tt> object. */
  private void initGlobalMCS()
  {
    final int REIN_FUNC = 0, GS_DRIVE = 1;
    final int ACS_IN_DIM = 2, NACS_IN_DIM = 3, ACS_OUT_DIM = 4, NACS_OUT_DIM = 5;
    final int ACS_IN_VAL = 6, NACS_IN_VAL = 7, ACS_OUT_VAL = 8, NACS_OUT_VAL = 9;
    final int ACS_LEARN = 10, NACS_LEARN = 11, ACS_REASON = 12, NACS_REASON = 13;
    final int ACS_CL = 14, ACS_PARAM = 15, NACS_PARAM = 16, ACS_REGU = 17, GS_TIME = 18;
    final int ACS_PERFORM = 19, NACS_PERFORM = 20, ACS_CONC_NUM = 21, NACS_CONC_NUM = 22;
    final int ACS_GAUGE = 23, NACS_GAUGE = 24, ACS_EPI_NUM = 25, NACS_EPI_NUM = 26;
    final int EPI_DEF = 27, ACS_INTEG = 28, NACS_INTEG = 29;
    Option[] opts = userOptions[SUPER_MCS][0][0][0][0];
    //global.setModuleGS( String value );
    //global.setModuleRein( String value );
    //global.setModuleAcsInputDims( String value );
    //global.setModuleNacsInputDims( String value );
    //global.setModuleAcsOutputDims( String value );
    //global.setModuleNacsOutputDims( String value );
    //global.setModuleAcsInputValues( String value );
    //global.setModuleNacsInputValues( String value );
    //global.setModuleAcsOutputValues( String value );
    //global.setModuleNacsOutputValues( String value );
    //global.setModuleAcsLearningMethod( String value );
    //global.setModuleNacsLearningMethod( String value );
    //global.setModuleAcsReasoningMethod( String value );
    //global.setModuleNacsReasoningMethod( String value );
    //global.setModuleAcsLevelSelection( String value );
    //global.setModuleNacsLevelSelection( String value );
    //global.setAcsRegulationCL( String value );
    //global.setTimingGS( String value );
    //global.setAcsPerformMeasure( String value );
    //global.setNacsPerformMeasure( String value );
    //global.setAcsGaugeLeaningMeasure( String value );
    //global.setNacsGaugeLearningMeasure( String value );
    global.setMonitorAcsTopConcNum(opts[ACS_CONC_NUM].getOptionItem(0).getParameter(TL).getIntValue());
    global.setMonitorAcsBottomConcNum(opts[ACS_CONC_NUM].getOptionItem(0).getParameter(BL).getIntValue());
    global.setMonitorAcsTopEpisodeNum(opts[ACS_EPI_NUM].getOptionItem(0).getParameter(TL).getIntValue());
    global.setMonitorAcsBottomEpisodeNum(opts[ACS_EPI_NUM].getOptionItem(0).getParameter(BL).getIntValue());
    global.setMonitorNacsTopConcNum(opts[NACS_CONC_NUM].getOptionItem(0).getParameter(TL).getIntValue());
    global.setMonitorNacsBottomConcNum(opts[NACS_CONC_NUM].getOptionItem(0).getParameter(BL).getIntValue());
    global.setMonitorNacsTopEpisodeNum(opts[NACS_EPI_NUM].getOptionItem(0).getParameter(TL).getIntValue());
    global.setMonitorNacsBottomEpisodeNum(opts[NACS_EPI_NUM].getOptionItem(0).getParameter(BL).getIntValue());
    //global.setEpisodeMeasure( String value );
    //global.setAcsParamCLMcs( String value );
    //global.setNacsParamCLMcs( String value );
  }

}
