package clarion.system;
import clarion.common.*;
import clarion.nacs.*;
import clarion.tools.*;
import clarion.options.*;
import clarion.options.transfer.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is the base class for all of the classes of concrete tasks to be simulated.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> public void init(): initializes the <tt>Global</tt> object associated with each agent
 * without using GUI.
 * <li> abstract public void start(): the start point of simulation of a concrete task and it
 * needs be extended by concrete methods.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * In the implemenation of CLARION model, there exists two parts, one part is implementation of
 * the CLARION model itself and the other part is task simulator in charge of specific task
 * simulation process. This class is a part of task simulator.
 * <li> Usage of this class. <br>
 * User needs extends <tt>this</tt> abstract class for a concrete task to run CLARION.
 * This class is instantiated inside a <tt>ClarionMain</tt> object by loading a specified task.
 * </ul>
 * @author Xi Zhang
 */

public abstract class Task {

  /** the # of agents running simultineously. */
  protected int runningAgentNum;
  /** the agents running simultineously. */
  protected TaskAgent[] agents;
  /** the global settings of agents. */
  protected Global[] globals;
  /** the setting names of agents. */
  protected String[] settingNames;
  /** the file used to store the simulation data. */
  protected WriteFile modelDataFile;

  /** Returns the number of agent running simultaneously.*/
  public int getRunningAgentNum()
  { return runningAgentNum; }

  /** Sets the number of agent running simultaneously.*/
  public void setRunningAgentNum( int num )
  { runningAgentNum = num; }

  /** Returns a specific agent. */
  public TaskAgent getAgent( int agentIdx )
  { return agents[agentIdx]; }

  /** Sets a specific agent.
   *  @param agentIdx the agent index on the agent group.
   *  @agent the value used to set the specific agent. */
  public void setAgent( int agentIdx, TaskAgent agent )
  { agents[agentIdx] = agent; }

  /** Returns the setting name for this agent.*/
   public String getSettingName( int agentIdx )
   { return settingNames[agentIdx]; }

   /** Sets setting names for running agents.*/
   public void setSettingNames( String[] settingNames )
   { this.settingNames = settingNames;  }

  /** Sets the global settings each agent is associated with. */
  public void setGlobals( Global[] globals )
  {  this.globals = globals; }

  /** Sets the model dats file to store the simulation data. */
  public void setDataFile( WriteFile file )
  {
     modelDataFile = file;
  }

  /** initialize the <tt>Global</tt> object associated with each agent without
   *  using GUI. 
   **/
  public void init()
  {
    initClarionOptions();
    initTaskSpecificOptions();
    initAgents();
  }
  
  /**Added 9/30/2006 by Nick Wilson
   * 
   * intialize the <tt>Global</tt> object associated with a single agent without
   *  using the GUI.
   * @param agent
   * @param setting
   */
  public void init(int agent, int setting)
  {
	  globals[agent] = new Global();
	  initClarionOptions(agent, setting);
	  initTaskSpecificOptions(agent, setting);
  }

  /** initialize the <tt>Global</tt> object associated with each agent without
   *  using GUI. 
   *  
   *  (note added 9/30/2006 by Nick Wilson)
   *  NOTE: IF YOU HAVE LESS GROUPS THAN AGENTS THIS WILL INIT ALL AGENTS AFTER
   *  GROUP # OF AGENTS TO BE IN THE LAST GROUP.
   *  EX: IF YOU HAVE 2 GROUPS AND 100 AGENTS, AGENT 1 WILL BE IN GROUP 1 AND
   *  AGENTS 2 - 100 WILL BE IN GROUP 2.*/
  private void initClarionOptions()
  {
     for( int i = 0; i < runningAgentNum; i++ ){
          globals[i].setTaskSettingName( settingNames[Math.min(i, settingNames.length-1)] );
          TransferManager transManager = new TransferManager( globals[i] );
          transManager.load( Global.CLARION_OPTIONS );
          transManager.setSaveCode( Global.CLARION_OPTIONS, Global.SAVE_TO_GLOBAL );
          transManager.save( Global.CLARION_OPTIONS );
     }
  }
  
  /**Added 9/30/2006 by Nick Wilson
   * 
   * intialize the <tt>Global</tt> object associated with a single agent without
   *  using the GUI.
   * @param agent
   * @param setting
   */
  private void initClarionOptions(int agent, int setting)
  {
	  globals[agent].setTaskSettingName( settingNames[setting] );
      TransferManager transManager = new TransferManager( globals[agent] );
      transManager.load( Global.CLARION_OPTIONS );
      transManager.setSaveCode( Global.CLARION_OPTIONS, Global.SAVE_TO_GLOBAL );
      transManager.save( Global.CLARION_OPTIONS );
  }

  /** initialize the <tt>Global</tt> object associated with each agent without
   *  using GUI. 
   * 
   *  (note added 9/30/2006 by Nick Wilson)
   *  NOTE: IF YOU HAVE LESS GROUPS THAN AGENTS THIS WILL INIT ALL AGENTS AFTER
   *  GROUP # OF AGENTS TO BE IN THE LAST GROUP.
   *  EX: IF YOU HAVE 2 GROUPS AND 100 AGENTS, AGENT 1 WILL BE IN GROUP 1 AND
   *  AGENTS 2 - 100 WILL BE IN GROUP 2.*/
  private void initTaskSpecificOptions()
  {
     for( int i = 0; i < runningAgentNum; i++ ){
          globals[i].setTaskSettingName( settingNames[Math.min(i, settingNames.length-1)] );
          TransferManager transManager = new TransferManager( globals[i] );
          transManager.load( Global.TASK_OPTIONS );
          transManager.setSaveCode( Global.TASK_OPTIONS, Global.SAVE_TO_GLOBAL );
          transManager.save( Global.TASK_OPTIONS );
     }
  }
  
  /**Added 9/30/2006 by Nick Wilson
   * 
   * intialize the <tt>Global</tt> object associated with a single agent without
   *  using the GUI.
   * @param agent
   * @param setting
   */
  private void initTaskSpecificOptions(int agent, int setting)
  {
	  globals[agent].setTaskSettingName( settingNames[setting] );
      TransferManager transManager = new TransferManager( globals[agent] );
      transManager.load( Global.TASK_OPTIONS );
      transManager.setSaveCode( Global.TASK_OPTIONS, Global.SAVE_TO_GLOBAL );
      transManager.save( Global.TASK_OPTIONS );
  }

  private void initAgents()
  {
     agents = new TaskAgent[runningAgentNum];
  }

  /** the start point of simulation of a task.*/
  abstract public void start();
}
