package clarion.system;
import clarion.common.*;
import clarion.tools.*;
import clarion.nacs.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is to implement what a task agent does in a task.
 * <li> summary of the important methods. <br>
 * <ol>
 * <li> abstract public void performAction( int netIdx, short[] action ): perform the selected action.
 * <li> abstract public void run(): simulate an agent performance by step.
 * </ol>
 * In order for a task agent to communicate with CLARION properly, all of the methods
 * in <tt>this</tt> class should be overrided by user codes.
 * </ul>
 * <li> Role in the CLARION model <br>
 * Acts as a task agent and it is part of a task simulator.
 * <li> Usage of this class. <br>
 * User needs extends <tt>this</tt> abstract class for concrete task agents to run CLARION.
 * This class is instantiated inside a <tt>Task</tt> object.
 * </ul>
 * @author Xi Zhang
 */

public abstract class TaskAgent{

   /** the global settings associated with <tt>this</tt> running agent. */
   protected Global  global;
   /** the embedded clarion model. */
   protected Clarion clarion;
   /** the task-specfic option settings in CLARION model*/
   protected TaskClarion taskClarion;
   protected Feature[] curSensoryInput;
   protected Feature[] curNacsInput;
   /** the retrieval results from NACS. */
   protected Feature[] nacsResults;
   /** the desired output. */
   protected Feature[] desiredOutput[];
   /** the reinforcements. */
   protected double reinforcements[];
   /** the active AMN network. */
   protected int activeAmnIdx;
   protected boolean learningTestOn;
  // protected WriteFile dataFile;

   public TaskAgent(){}

   /** class constructor with associated <tt>Global</tt> object. */
   public TaskAgent( Global g )
   {
      global = g;
      clarion = new Clarion( global, this );
      taskClarion = global.getTaskClarion();
   }

   /** reinitialization routine. */
   public void reinit()
   {
      clarion = new Clarion( global, this );
   }

   public boolean isInLearningTest()
   {
      return learningTestOn;
   }

  /** Returns the index on the active AMN network.
   *  @return the index on the active AMN network. */
   public int getActiveAmnIdx()
   {
      return activeAmnIdx;
   }

   /** Returns the RER rules in the foramt of string array generated in the specified network in ACS.
       @param netIdx the specified network in ACS.
       @return the RER rules in the foramt of string array generated in the specified network in ACS. */
   public String[] getRerRules( int netIdx )
   {
      return clarion.getRerRules( netIdx );
   }

   /** Returns the IRL rules generated in the specified network in ACS.
       @param netIdx the specified network in ACS.
       @return the IRL rules generated in the specified network in ACS in the foramt of string array. */
   public String[] getIrlRules( int netIdx )
   {
      return clarion.getIrlRules( netIdx );
   }

   /** Returns the FR Rules generated in the specified network in ACS.
       @param netIdx the specified network in ACS.
       @return the FR rules generated in the specified network in ACS in the foramt of string array. */
   public String[] getFixRules( int netIdx )
   {
      return clarion.getFixRules( netIdx );
   }

   /** Returns current sensory input. */
   public Feature[] getCurSensoryInput()
   {
      return curSensoryInput;
   }

   /** Returns current input to NACS. */
   public Feature[] getCurNacsInput()
   {
      //return (curNacsInput = curSensoryInput);
      return curNacsInput;
   }

   /** Returns current desired output. */
   public Feature[][] getCurDesiredOutput()
   {
      return desiredOutput;
   }

   public void setLearningTest( boolean flag )
   {
      learningTestOn = flag;
   }

   /** save the current into the specified data file. */
   public void saveOutput( String results )
   {
     // dataFile.saveStringData( results );
   }

   /** perform the chosen action.
    *  @param netIdx the ACS network index.
    *  @param action the selected actions by ACS networks. */
   public void performAction( int netIdx, short[][][] action )
   {
      if( global.isNacsOn() )
          global.getNACS().performAction( action );
   }

   /** Returns the output of current step.
    *  @param arr the array to be filled in with current output. */
   abstract public void getOutput( Object arr );
   /** Determines which AMN will be used given current action.
    *  @param netIdx the ACS network index.
    *  @param action the selected actions by ACS networks.
    *  @return the AMN network index. */
   abstract public int determineAmnNet( int netIdx, short[][][] action );
   /** send the retrieval results from NACS to ACS components.
    *  @param netIdx the ACS network index.
    *  @param results retrieval results from NACS */
   abstract public void setRetrievalResults( int netIdx, GKSChunk[] results );
   /** simulate a subject performance by step. */
   abstract public void run();
}
