package clarion.system;
import java.util.Arrays;
import clarion.common.*;
import clarion.nacs.*;
import clarion.acs.net.auxiliary.*;
import clarion.acs.gs.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class is the base class for user codes to extend. This class defines all
 * of the necessary methods relevant to task-specific part of CLARION options and
 * parameters. In order for a specific task to run CLARION properly, all of the
 * methods defined in <tt>this</tt> class should be overrided by user codes.
 * <li> summary of the important methods. <br>
 * All of the methods are defined according to the task-specific part of CLARION
 * options and parameters. See the appendix of the tutorial on Options and Parameters
 * to make sense of them.
 * </ul>
 * <li> Role in the CLARION model <br>
 * It is a part of task simulator.
 * <li> Usage of this class. <br>
 * Users need to override the methods if necessary. This class is instantiated
 * inside a <tt>TaskAgent</tt> object.
 * </ul>
 * @author Xi Zhang
 */

public class TaskClarion{
  /** constants for CLARION subsystems. */
  final public static int SUB_SYS_NUM = 4, ACS = 0, NACS = 1, RT = 2, MCS = 3;
  final public static int ACS_SAMPLE_NUM = 22, NACS_SAMPLE_NUM = 6;
  final public static int RT_SAMPLE_NUM = 9, MCS_SAMPLE_NUM = 1;
  final public static int GOAL = 0, GOAL_DIM = 1, BUR = 2, TDG = 3, ELIG = 4, REIN = 5;
  final public static int RER_WGT_C = 6, RER_WGT_R = 7, RER_UTL = 8, RER_POS = 9, RER_IG = 10;
  final public static int IRL_WGT_C = 11, IRL_WGT_R = 12, IRL_UTL = 13, IRL_POS = 14, IRL_IG = 15;
  final public static int IRL_PARAM = 16, FR_WGT_C = 17, FR_WGT_R = 18, FR_UTL = 19, FR_CON = 20;
  final public static int FR_ACT = 21;
  final public static int EX_CHUNK = 0, EX_RULE = 1, ASSOC_WGT_C = 2, ASSOC_WGT_R = 3, SIM = 4;
  final public static int DATA_TYPE = 5;
  final public static int OVER_RT = 0, BL_RT = 1, TL_RT = 2, PT_TL = 3, DT_TL = 4, AT_TL = 5;
  final public static int ASSOC_APP = 6, CHUNK_RETR = 7, AMN_PASS = 8;
  final public static int DRIVE = 0, DRV_REP = 1;
  /*
  final public static int REIN_FUNC = 0, GOAL_SETTING = 1;
  final public static int ACS_IN_DIM = 2, NACS_IN_DIM = 3, ACS_OUT_DIM = 4, NACS_OUT_DIM = 5;
  final public static int ACS_IN_VAL = 6, NACS_IN_VAL = 7, ACS_OUT_VAL = 8, NACS_OUT_VAL = 9;
  final public static int ACS_LEARN = 10, NACS_LEARN = 11, ACS_REASON = 12, NACS_REASON = 13;
  final public static int ACS_CL = 14, ACS_PARAM = 15, NACS_PARAM = 16;
  final public static int ACS_PERFORM = 17, NACS_PERFORM = 18, ACS_GAUGE = 19, NACS_GAUGE = 20;
  */
  final public static int REIN_FUNC = 0, GOAL_SETTING = 0;
  final public static int ACS_IN_DIM = 0, NACS_IN_DIM = 0, ACS_OUT_DIM = 0, NACS_OUT_DIM = 0;
  final public static int ACS_IN_VAL = 0, NACS_IN_VAL = 0, ACS_OUT_VAL = 0, NACS_OUT_VAL = 0;
  final public static int ACS_LEARN = 0, NACS_LEARN = 0, ACS_REASON = 0, NACS_REASON = 0;
  final public static int ACS_CL = 0, ACS_PARAM = 0, NACS_PARAM = 0;
  final public static int ACS_PERFORM = 0, NACS_PERFORM = 0, ACS_GAUGE = 0, NACS_GAUGE = 0;
  final static String className = "TaskClarion";
  final static String[][] methodNames = {
        {
           "getGoals",  "getGoalDims", "combineByBottomUpRectification",  "combineByTopDownGuidance",
           "getEligibility",  "getReinforcement", "getRerConditionWeights", "getRerRuleWeight",
           "getRerUtility", "getRerPositivity",  "getRerIG", "getIrlConditionWeights",
           "getIrlRuleWeight", "getIrlUtility", "getIrlPositivity",  "getIrlIG", "getIrlRule",
           "getFrConditionWeights", "getFrRuleWeight", "getFrUtility",  "getFrCondition",
           "getFrAction"
        },
        {
           "getExternalNacsChunks", "getExternalNacsRules",
           "getAssocConditionWeights", "getAssocDimensionWeights",
           "getSimilarity", "getAamDataPattern",
        },
        {
           "getOverallRT", "getBLOverallRT", "getTLOverallRT", "getPT_TL", "getDT_TL",
           "getAT_TL",  "getAssocAppTime", "getChunkRetrtime", "getAmnPassTime"
        },
        {  "getDriveSet" }
  };

  final static String[][] methodForms = {
        {
           "public short[][] getGoals()",
           "public short[] getGoalDims()",
           "public void combineByBottomUpRectification()",
           "public void combineByTopDownGuidance()",
           "public short[][][] getEligibility( int netIdx )",
           "public double getReinforcement( int netIdx, Feature[] input,         \n" +
           "       Chunk[] wmChunks, int goal, short[][][] action, Feature[] newInput,\n" +
           "       Chunk[] newWMChunks, int newGoal ) ",
           "public double[] getRerConditionWeights( int netIdx ) ",
           "public double getRerRuleWeight( int netIdx ) ",
           "public double getRerUtility( int netIdx, double PM, double NM, \n" +
           "       double cost ) ",
           "public boolean getRerPositivity( int learnOpt, double reward,  \n" +
           "       double qDiscount, double newMaxQVal, double qVal,       \n" +
           "       double threshold ) ",
           "public double getRerIG( int netIdx, double PMa, double NMa,    \n" +
           "       double PMb, double NMb )",
           "public double[] getIrlConditionWeights( int netIdx )",
           "public double getIrlRuleWeight( int netIdx )",
           "public double getIrlUtility( int netIdx, double PM, double NM, \n" +
           "       double cost )",
           "public boolean getIrlPositivity( int netIdx, double reward,    \n" +
           "       double qDiscount, double newMaxQVal, double qVal,       \n" +
           "       double threshold ) ",
           "public double getIrlIG( int netIdx, double PMa, double NMa )",
           "public IrlRuleForm getIrlRule(int groupIdx, int netIdx, int setIdx, int ruleIdx)",
           "public double[] getFrConditionWeights( int netIdx )",
           "public double getFrRuleWeight( int netIdx )",
           "public double getFrUtility( int netIdx, double PM, double NM,  \n" +
           "       double cost )",
           "public short[][] getFrCondition(int groupIdx, int netIdx, int ruleIdx)",
           "public short[][][] getFrAction(int groupIdx, int netIdx, int ruleIdx)"
        },
        {
           "public GKSChunk[] getExternalNacsChunks()",
           "public AssocRule[] getExternalNacsRules()",
           "public double[] getAssocConditionWeights()",
           "public double getAssocDimensionWeights()",
           "public double getSimilarity(GKSChunk chunk1, GKSChunk chunk2)",
           "public short[] getAamDataPattern(int netIdx, int ruleIdx)"
        },
        {
           "public int getOverallRT( int RT_BL, int RT_TL )",
           "public int getBLOverallRT( int PT_BL, int DT_BL, int AT_BL )",
           "public int getTLOverallRT( int PT_TL, int DT_TL, int AT_TL )",
           "public int getPT_TL( int PT_BL )",
           "public int getDT_TL( int DT_BL, int actionTime, double ruleBla, \n" +
           "           double chunkBla )",
           "public int getAT_TL( int action, int assocAppTime, int chunkRetrTime )",
           "public int getAssocAppTime( double assocRuleBla )",
           "public int getChunkRetrtime( double chunkBla )",
           "public int getAmnPassTime()"
        },
        {
           "public Drive getDriveSet()"
        }
  };

  final static String[][] descriptions = {
        {
           "to get all of goals used in goal structure in the form of its values   \n" +
           "in each parameter dimension.                                           \n" +
           "1st dimension - the goal NO.                                           \n" +
           "2nd dimension - eligible value in each parameter dimension.  ",
           "to get goal parameter dimension-value information. Currently, the goal \n" +
           "structure has several goal items representing goal and its parameters. \n" +
           "Those dimensions help describing what a specific goal wants to achieve.\n" +
           "The goal parameter dimension information CLARION wants to know is how  \n" +
           "many values in each dimension. How such dimensions works on goal action\n" +
           "decision is task-specific and so needs user's code to implement it     \n" +
           "whether in another option, user choose goal action by top level( Fixed \n" +
           "Rules) or both levels.             ",
           "the process of Bottom-Up correction",
           "the process of Top-Guidance correction",
           "to get eligible conditions of an external network. Eligible condition  \n" +
           "determines which network is eligible given current input. Each network \n" +
           "has an eligible condition for current input to match.                  \n" +
           "1. Eligible condition of a network has the format:                     \n" +
           "   if ( C1 or C2...or Cn ) then use net K. ( 'or' conditions )         \n" +
           "2. Ci has the format (ci1 and ci2... and cim). ( 'and' subconditions ) \n" +
           "   each condition cij is an array representing an input dimension.     \n" +
           "   In the array, only the elements those values are 1 is the eligible  \n" +
           "   elements in that dimension.                                         \n" +
           "3. The eligible conditions of an external network is represented by 3- \n" +
           "   dimension array.                                                    \n" +
           "   1st dimension - the 'or' eligible conditions of an external network \n" +
           "   that means, if any one of the condition is satisfied, the network   \n" +
           "   is eligible.                                                        \n" +
           "   2nd dimension - the 'and' subconditions composing a 'or' condition. \n" +
           "   In this dimension, the first element, 1-dimension array, stores the \n" +
           "   involved dimension indices on the input.                            \n" +
           "   3rd dimension - the eligible elements in one dimension composing    \n" +
           "   an 'and' subcondition.                                              \n" +
           "4. In user code, the eligible conditions should be given in the above  \n" +
           "   format. ",
           "to get reinforcement given current input, current working memory status\n" +
           "if WM is set, current goal if GS is set; current action; the new input,\n" +
           "the new working memory status if WM is set, new goal if GS is set      \n" +
           "resulting from the firing of current action. ",
           "to get dimension weights of RER condition which is used for          \n" +
           "calculating the rule support for the action of the RER rule",
           "to get weight of the chunk representing the condition of an RER rule. ",
           "to get RER Utility which measures the effectiveness, or usefulness of\n" +
           "a rule",
           "to get RER positivity which means whether or not the aciton chosen   \n" +
           "according to a rule is reasonably good.",
           "to get Information Gain of RER rule A over RER rule B.",
           "to get dimension weights of IRL condition which is used for          \n" +
           "calculating the rule support for the action of the IRL rule",
           "to get weight of the chunk representing the condition of an IRL rule. ",
           "to get IRL utility which measures the effectiveness, or usefulness of\n" +
           "a rule",
           "to get IRL positivity which means whether or not the aciton chosen   \n" +
           "according to a rule is reasonably good.",
           "to get Information Gain of IRL rule A over IRL rule B.",
           "to get an IRL rule. A IRL rule is composed of a set of parameters    \n" +
           "Most parameters are relevant to an input dimension and some are not. \n" +
           "Only knowing the parameters is not enough to make an action decision,\n" +
           "but also we need know how the parameters operates on action decision.\n" +
           "Since the action decision is task-specific, user still has to implement\n" +
           "the codes for how IRL rules work on action decision. That is a method \n" +
           "in class IrlRuleForm.java for user to implement. ",
           "to get dimension weights of Fixed Rule condition which is used for   \n" +
           "calculating the rule support for the action of the Fixed Rule.    ",
           "to get weight of the chunk representing the condition of a Fixed Rule. ",
           "to get Fixed Rule Utility which measures the effectiveness, or       \n" +
           "usefulness of a rule ",
           "to get a Fixed Rule condition. A Fixed Rule condition is represented \n" +
           "by a 2-dimension array.                                              \n" +
           "1st dimension - represents an input dimension.                       \n" +
           "2nd dimension - represents the eligible elements in that dimension.  \n" +
           "The value of each element is 0 or 1, 0: not eligible, 1: eligible.   \n" +
           "Actually, there are two kinds of Fixed Rules, one has fixed aciton   \n" +
           "and the other hasn't. In the first case, differently from IRL rules, \n" +
           "knowing Fixed Rule condition is enough to make an aciton decision by \n" +
           "simply doing a match of current input with the condition. In the     \n" +
           "second case, we still need user's code to implement how Fixed Rule   \n" +
           "works. Also this is a method in the class TaskAgent.java for user to \n" +
           "override.            ",

           " Get the actions of the specific Fixed Rule. The returned value is \n" +
           " in the format of 3-dimenisonal array.                             \n" +
           " The first dimension indicates action type :                       \n" +
           " index = 0 : normal action, index = 1: NACS control action.        \n" +
           " The second dimension indexes an action dimension.                 \n" +
           " The third stores the active values in that dimension.             \n"
        },
        {
           "to get the externally encoded NACS chunks. \n",
           "to get the externally encoded NACS rules.  \n",
           "to get chunk weights of associative rule condition which is used \n" +
           "for calculating the rule support for the action of the associative rule",
           "to get dimension weights in a chunk. ",
           "to get the similarity of two NACS chunks",
           "to get the type of data pattern for training AAM."
        },
        {
           "to get overall response time in each step, user should give a new   \n" +
           "formula to calculate it.",
           "to get overall Bottom-Level response time in each step, user should \n" +
           "give a new formula to calculate it.",
           "to get overall Top-Level response time in each step, user should    \n" +
           "give a new formula to calculate it.",
           "to get Top-Level Perceptual Time in each step, user should set a    \n" +
           "number to this variable or a new formula to calculate it.",
           "to get Top-Level Decision Time in each step, user should set a      \n" +
           "number to this variable or a new formula to calculate it.",
           "to get Top-Level Actuation Time in each step, user should set a     \n" +
           "number to this variable or a new formula to calculate it.",
           "to get Associatve rule Application Time of one iteration, user      \n" +
           "should set a number to this variable or a new formula to calculate  \n" +
           "it.",
           "to get Chunk Retrieval Time of one iteration, user should set a     \n" +
           "number to this variable or a new formula to calculate it.",
           "to get AMN one-pass time from mapping input to output, user should  \n" +
           "set a number to this variable or a new formula to calculate it.",
        },
        {
           "to user defined drive set."
        }
  };

  final static String[][] params = {
        {
           "null",
           "null",
           "null",
           "null",
           "int netIdx - the external network NO.              ",
           "int netIdx - the network NO.                                     \n" +
           "Feature[] input - current input,                                 \n" +
           " ( see the class Feature.java in subdirectory: clarion.common )  \n" +
           "WMChunk[] wmChunks - current working memory status.              \n" +
           " ( see the class WMChunk.java in subdirectory: clarion.acs.wm )  \n" +
           "int goal - current goal.                                         \n" +
           "int action - current action.                                     \n" +
           "Feature[] newInput - new input resulting from firing the action. \n" +
           "WMChunk[] newWMChunks - new working memory status.               \n" +
           "int newGoal - new goal.                             ",
           "int netIdx - the network NO.                        ",
           "int netIdx - the network NO.                        ",
           "int netIdx - the network NO.                                         \n" +
           "double PM - Positive Match of the rule, equals the number of times   \n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is positive.                                          \n" +
           "double NM - Negative Match of the rule, equals the number of times   \n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is negative.                                          \n" +
           "double cost - the execution cost of the rule.         ",
           "int netIdx - the network NO.                                         \n" +
           "double reward - currently received reward.                           \n" +
           "double qDiscount - the discount factor for Q-Learning that favors    \n" +
           "reinfocement received sooner relative to that received later.        \n" +
           "double newMaxQVal - maximum Q-value of the new state resuling from   \n" +
           "current action in current state, Q(next state, next possible action) \n" +
           "double qVal - current q-value, Q(current state, current action).     \n" +
           "double threshold - the threshold for rule positivity. ",
           "int netIdx - the network NO.                                         \n" +
           "double PMa - Positive Match of the rule a, equals the number of times\n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is positive.                                          \n" +
           "double NMa - Negative Match of the rule a, equals the number of times\n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is negative.                                          \n" +
           "double PMb - Positive Match of the rule b.                           \n" +
           "double NMb - Negative Match of the rule b.            ",
           "int netIdx - the network NO.                          ",
           "int netIdx - the network NO.                          ",
           "int netIdx - the network NO.                                         \n" +
           "double PM - Positive Match of the rule, equals the number of times   \n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is positive.                                          \n" +
           "double NM - Negative Match of the rule, equals the number of times   \n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is negative.                                          \n" +
           "double cost - the execution cost of the rule.          ",
           "int netIdx - the network NO.                                         \n" +
           "double reward - currently received reward.                           \n" +
           "double qDiscount - the discount factor for Q-Learning that favors    \n" +
           "reinfocement received sooner relative to that received later.        \n" +
           "double newMaxQVal - maximum Q-value of the new state resuling from   \n" +
           "current action in current state, Q(next state, next possible action) \n" +
           "double qVal - current q-value, Q(current state, current action).     \n" +
           "double threshold - the threshold for rule positivity.  ",
           "int netIdx - the network NO.                                         \n" +
           "double PMa - Positive Match of the rule a, equals the number of times\n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is positive.                                          \n" +
           "double NMa - Negative Match of the rule a, equals the number of times\n" +
           "that an input state matches rule condition, rule action is performed \n" +
           "and the result is negative.                                          \n" +
           "double threshold - compared threshold value.      ",
           "int groupIdx - learniing group the agent is in.                        \n" +
           "int netIdx - the network NO.                                         \n" +
           "int setIdx - the rule subset NO.                                     \n" +
           "int ruleIdx - the rule NO in that rule subset.    ",
           "int netIdx - the network NO.                      ",
           "int netIdx - the network NO.                      ",
           "int netIdx - the network NO.                                        \n" +
           "double PM - Positive Match of the rule, equals the number of times  \n" +
           "that an input state matches rule condition, rule action is performed\n" +
           "and the result is positive.                                         \n" +
           "double NM - Negative Match of the rule, equals the number of times  \n" +
           "that an input state matches rule condition, rule action is performed\n" +
           "and the result is negative.                                         \n" +
           "double cost - the execution cost of the rule.     ",
           "int groupIdx - learniing group the agent is in.                    \n" +
           "int netIdx - the network NO.                                     \n" +
           "int ruleIdx - the rule NO.                        ",
           "int groupIdx - learniing group the agent is in.                    \n" +
           "int netIdx - the network NO.                                     \n" +
           "int ruleIdx - the rule NO.                        ",
        },
        {
           "null",
           "null",
           "null",
           "null",
           "GKSChunk chunk1 - chunk for comparison.                         \n" +
           " ( see the class GKSChunk.java in subdirectory: clarion.nacs )  \n" +
           "GKSChunk chunk2 - chunk for comparison.              ",
           "null"
        },
        {
           "int RT_BL - Bottom-Level Response Time.   \n" +
           "int RT_TL - Top-Level Response Time.     ",
           "int PT_BL - Bottom-Level Perceptual Time. \n" +
           "int DT_BL - Bottom-Level Decision Time.   \n" +
           "int AT_BL - Bottom-Level Actuation Time. ",
           "int PT_TL - Top-Level Perceptual Time.    \n" +
           "int DT_TL - Top-Level Decision Time.      \n" +
           "int AT_TL - Top-Level Actuation Time.    ",
           "int PT_BL - Bottom-Level Perceptual Time.",
           "int DT_BL - Bottom-Level Decision Time.   \n" +
           "int actionTime - time determined by the operation dictated by \n" +
           "a particular rule.                                            \n" +
           "double ruleBla - rule base-level activation.                  \n" +
           "double chunkBla - action chunk base-level activation.        ",
           "int action - current action, deciding if NACS is involved or not.   \n" +
           "int assocAppTime - time of one iteration of associative applicaiton.\n" +
           "int chunkRetrTime - time of one iteration of chunk retrieval. ",
           "double assocRuleBla - associative rule base-level activation. ",
           "double chunkBla - nacs chunk base-level activation.           ",
           "null"
        },
        {
           "null"
        }
  };

  final static String[][] returns = {
        {
           "the goals represented by parameter dimension-value. ",
           "goal parameters' dimension-value information",
           "null",
           "null",
           "the eligible condtions of an external network.",
           "current reinforcement of the specific ACS network",
           "the dimension weights of RER condtion",
           "the weight of the RER rule",
           "the utility of RER",
           "if the action chosen according to a RER rule is reasonably good",
           "the information gain of RER rule A over RER rule B",
           "the dimension weights of IRL condtion",
           "the weight of the IRL rule",
           "the utility of IRL",
           "if the action chosen according to a IRL rule is reasonably good",
           "the information gain of IRL rule A over a random rule",
           "the IRL rule form",
           "the dimension weights of Fixed rule condtion",
           "the weight of the Fixed Rule",
           "the utility of Fixed rule",
           "the condition of Fixed Rule",
           "the action of Fixed Rule"
        },
        {
           "the externally encoded NACS chunks.",
           "the externally enocded NACS rules.",
           "the condition weights of Associative rule condtion",
           "the weights in a chunk",
           "the similarity between two NACS chunks using use defined measure",
           "return the dimension-value infomation of data pattern for training AAM"
        },
        {
           "the overall Response Time in each step",
           "the overall Bottom-Level Response Time in each step",
           "the overall Top-Level Response Time in each step",
           "the Top-Level Perceptual Time in each step",
           "the Top-Level Decision Time in each step",
           "the Top-Level Actuation Time in each step",
           "the time in GKS for one iteration of parallel applications of \n" +
           "associative rules",
           "the time of one iteration of chunk retrieval",
           "the time of one iteration of AMN mapping input to output"
        },
        {
           "Drive"
        }
  };

  final static String[][] seeSamples = {
        {
           "goals set is empty",
           "Such as in Tower of Hanoi task, each goal item has six parameter \n" +
           "dimensions: (DSIZE, FROM, TO, DSIZE1, FROM1, TO1). Each parameter\n" +
           "dimension has some possible values.  ",
           "null",
           "null",
            "public short[][][] getEligibility( int netIdx )                           \n" +
            "{                                                                         \n" +
            "  //**                                                                    \n" +
            "   * suppose the input has 4 dimensions and each dimension has 5 values.  \n" +
            "   * Eligibility of net 0 is :                                             \n" +
            "   * if ((c1 = v11 or v12) and (c2 > v23)) or ((c3 <= v34) and (c4 >= v42))) \n" +
            "   * then use net 0.                                                    \n" +
            "   * ( vij means the jth value of dimension i. )                        \n" +
            "   *//                                                                  \n" +
            "    short[][][] eligCondition = {                                       \n" +
            "           {                                                              \n" +
            "                {0,1},        /// relevant dims: 0, 1, in this condition. \n" +
            "                {1,1,0,0,0},  /// c1 = v11 or v12.                        \n" +
            "                {0,0,0,1,1}   /// c2 > v23.                               \n" +
            "           },                                                             \n" +
            "           {                                                              \n" +
            "                 {2,3},       /// relevant dims: 2, 3, in this condition. \n" +
            "                 {1,1,1,1,0}, /// c3 <= v34.                              \n" +
            "                 {0,1,1,1,1}  /// c4 >= v42.                              \n" +
            "           }                                                              \n" +
            "    };                                                                    \n" +
            "    return eligCondition;                                                 \n" +
            "}                                                                         ",

            "public double getReinforcement( int netIdx, Feature[] input, Chunk[] wmItems, \n" +
            "                                GoalChunk goalItem, short[][][] action,   \n" +
            "                                Feature[] newInput, Chunk[] newWmItems,   \n" +
            "                                GoalChunk newGoalItem )                   \n" +
            "{                                                                         \n" +
            "   int TARGET = 5;                                         \n" +
            "   int P = newInput[0].getValues()[0];                     \n" +
            "   int diff = Math.abs( P - TARGET );                      \n" +
            "   double fb = 0.0;                                        \n" +
            "   switch( diff ){                                         \n" +
            "       case 0:                                             \n" +
            "       case 1:                                             \n" +
            "            fb = 1.0;                                      \n" +
            "            break;                                         \n" +
            "       default:                                            \n" +
            "            fb = -0.2 * ( diff - 1 );                      \n" +
            "            break;                                         \n" +
            "   }                                                       \n" +
            "   return fb;                                              \n" +
            "}                                                          \n",

           "the default RER condition weights",
           "the default RER rule weight",
           "other RER utility measures in this option",
           "the default RER positivity criterion",
           "the default RER information gain measure",
           "the default IRL condition weights",
           "the default IRL rule weight",
           "other IRL utility measures in this option",
           "the default IRL positivity criterion",
           "the default IRL information gain measure",
           "/*\n In Process Control task, there are four sets of IRL rules.  \n" +
           "1. P = a * W + b           \n" +
           "2. P = a * W + c * P1 + b  \n" +
           "3. P = a * W1 + b          \n" +
           "4. P = a * W1 + c * P2 + b \n" +
           "a = {1,2},  b = {-1,-2,0,1,2},  c = {-1,-2,1,2}.            \n" +
           "P = the desired system output level(the goal),              \n" +
           "W = the current input to the system(to be determined),      \n" +
           "W1 = the previous input to the system,                      \n" +
           "P1 = the previous system output level(under W1),            \n" +
           "P2 = the system output level at the time step before P1.    \n" +
           "In this case, elements from the set a, b, c compose the IRL \n" +
           "parameters.   \n */                   \n\n" +
           "public class IrlRuleFormPC implements IrlFuleForm{       \n" +
           " int a, b, c, target, ruleSetIdx;                        \n" +
           " int curAction, nextAction;                              \n" +
           " public IrlRuleFormPC( int a, int b, int c, int target, int ruleSetIdx, int curAction ) \n" +
           " {                                                      \n" +
           "   this.a = a;                                          \n" +
           "   this.b = b;                                          \n" +
           "   this.c = c;                                          \n" +
           "   this.target = target;                                \n" +
           "   this.ruleSetIdx = ruleSetIdx;                        \n" +
           "   this.curAction = curAction;                          \n" +
           "   nextAction = ( int ) ( target - b ) / a;             \n" +
           " }               \n\n" +
           " public int determineAction( Feature[] input )          \n" +
           " {                                                      \n" +
           "    int p1 = input[0].getValues()[0];                   \n" +
           "    int chosenAction = Math.round( ( target - c * p1 - b ) / a ); \n" +
           "    if ( chosenAction < 0 )                             \n" +
           "         chosenAction = 0;                              \n" +
           "    else if ( chosenAction > 11 )                       \n" +
           "              chosenAction = 11;                        \n" +
           "    switch( type )                                      \n" +
           "    {                                                   \n" +
           "        case 0:  // action should be returned for this step. \n" +
           "        case 1:                                         \n" +
           "             curAction = chosenAction;                  \n" +
           "             return curAction;                          \n" +
           "        case 2:  // action should be returned for next step. \n" +
           "        case 3:                                         \n" +
           "             curAction = nextAction;                    \n" +
           "             nextAction = chosenAction;                 \n" +
           "             return curAction;                          \n" +
           "    }                                                   \n" +
           "    return curAction;                                   \n" +
           " }                                                      \n" +
           "}               ",
           "the default Fixed Rule condition weights",
           "the default Fixed Rule rule weight",
           "other Fixed Rule utility measures in this option",
           "public short[][] getFrCondition(int groupIdx, int netIdx, int ruleIdx)      \n" +
           "{                              \n" +
           "    //*                        \n" +
           "     * Suppose the input has 4 dimensions, one possible condtion of a Fixed Rule: \n" +
           "     * { {1,1,1,1}, {1,0,1}, {0,0,1}, {1,0} }                      \n" +
           "     *//                                                           \n" +
           "   short[][] conditions = { {1,1,1,1}, {1,0,1}, {0,0,1}, {1,0} };  \n" +
           "   return conditions;                                              \n" +
           "}                                ",

           "public short[][][] getFrAction(int groupIdx, int netIdx, int ruleIdx) \n" +
           "{                                                                  \n" +
           "   short[][][] action = {{ {0,1}, {0}, {2,3}, {0}}};               \n" +
           "   return action;                                                  \n" +
           "}                                                                  \n"

        },
        {
           "null",
           "null",
           "the default associative rule condition weights",
           "the default dimension weights in a chunk",
           "the default similarity measure in this option",
           "the default data pattern - experienced state"
        },
        {
           "the default overall response time formula",
           "the default overall bottome-level response time formula",
           "the default overall top-level response time formula",
           "the default top-level perceptual time formula",
           "the default top-level decision time formula",
           "public int getAT_TL( int action, int assocAppTime, int chunkRetrTime ) \n" +
           "{                                                 \n" +
           "  return 100 + 5 * assocAppTime + chunkRetrTime;  \n" +
           "}                                                   ",
           "the default associative rule application time formula",
           "the default chunk retrieval time formula",
           "the default AMN one-pass time formula"
        },
        {
           "see default drive set."
        }
  };

  final static String[][] explanations = {
        {
           "goals set is empty",
           "Such as in Tower of Hanoi task, each goal item has six parameter \n" +
           "dimensions: (DSIZE, FROM, TO, DSIZE1, FROM1, TO1). Each parameter\n" +
           "dimension has some possible values.  ",
           "null",
           "null",
            "public short[][][] getEligibility( int netIdx )                           \n" +
            "{                                                                         \n" +
            "  //**                                                                    \n" +
            "   * suppose the input has 4 dimensions and each dimension has 5 values.  \n" +
            "   * Eligibility of net 0 is :                                             \n" +
            "   * if ((c1 = v11 or v12) and (c2 > v23)) or ((c3 <= v34) and (c4 >= v42))) \n" +
            "   * then use net 0.                                                    \n" +
            "   * ( vij means the jth value of dimension i. )                        \n" +
            "   *//                                                                  \n" +
            "    short[][][] eligCondition = {                                       \n" +
            "           {                                                              \n" +
            "                {0,1},        /// relevant dims: 0, 1, in this condition. \n" +
            "                {1,1,0,0,0},  /// c1 = v11 or v12.                        \n" +
            "                {0,0,0,1,1}   /// c2 > v23.                               \n" +
            "           },                                                             \n" +
            "           {                                                              \n" +
            "                 {2,3},       /// relevant dims: 2, 3, in this condition. \n" +
            "                 {1,1,1,1,0}, /// c3 <= v34.                              \n" +
            "                 {0,1,1,1,1}  /// c4 >= v42.                              \n" +
            "           }                                                              \n" +
            "    };                                                                    \n" +
            "    return eligCondition;                                                 \n" +
            "}                                                                         ",

            "public double getReinforcement( int netIdx, Feature[] input, Chunk[] wmItems, \n" +
            "                                GoalChunk goalItem, short[][][] action,   \n" +
            "                                Feature[] newInput, Chunk[] newWmItems,   \n" +
            "                                GoalChunk newGoalItem )                   \n" +
            "{                                                                         \n" +
            "   int TARGET = 5;                                         \n" +
            "   int P = newInput[0].getValues()[0];                     \n" +
            "   int diff = Math.abs( P - TARGET );                      \n" +
            "   double fb = 0.0;                                        \n" +
            "   switch( diff ){                                         \n" +
            "       case 0:                                             \n" +
            "       case 1:                                             \n" +
            "            fb = 1.0;                                      \n" +
            "            break;                                         \n" +
            "       default:                                            \n" +
            "            fb = -0.2 * ( diff - 1 );                      \n" +
            "            break;                                         \n" +
            "   }                                                       \n" +
            "   return fb;                                              \n" +
            "}                                                          \n",

           "the default RER condition weights",
           "the default RER rule weight",
           "other RER utility measures in this option",
           "the default RER positivity criterion",
           "the default RER information gain measure",
           "the default IRL condition weights",
           "the default IRL rule weight",
           "other IRL utility measures in this option",
           "the default IRL positivity criterion",
           "the default IRL information gain measure",
           "/*\n In Process Control task, there are four sets of IRL rules.  \n" +
           "1. P = a * W + b           \n" +
           "2. P = a * W + c * P1 + b  \n" +
           "3. P = a * W1 + b          \n" +
           "4. P = a * W1 + c * P2 + b \n" +
           "a = {1,2},  b = {-1,-2,0,1,2},  c = {-1,-2,1,2}.            \n" +
           "P = the desired system output level(the goal),              \n" +
           "W = the current input to the system(to be determined),      \n" +
           "W1 = the previous input to the system,                      \n" +
           "P1 = the previous system output level(under W1),            \n" +
           "P2 = the system output level at the time step before P1.    \n" +
           "In this case, elements from the set a, b, c compose the IRL \n" +
           "parameters.   \n */                   \n\n" +
           "public class IrlRuleFormPC implements IrlFuleForm{       \n" +
           " int a, b, c, target, ruleSetIdx;                        \n" +
           " int curAction, nextAction;                              \n" +
           " public IrlRuleFormPC( int a, int b, int c, int target, int ruleSetIdx, int curAction ) \n" +
           " {                                                      \n" +
           "   this.a = a;                                          \n" +
           "   this.b = b;                                          \n" +
           "   this.c = c;                                          \n" +
           "   this.target = target;                                \n" +
           "   this.ruleSetIdx = ruleSetIdx;                        \n" +
           "   this.curAction = curAction;                          \n" +
           "   nextAction = ( int ) ( target - b ) / a;             \n" +
           " }               \n\n" +
           " public int determineAction( Feature[] input )          \n" +
           " {                                                      \n" +
           "    int p1 = input[0].getValues()[0];                   \n" +
           "    int chosenAction = Math.round( ( target - c * p1 - b ) / a ); \n" +
           "    if ( chosenAction < 0 )                             \n" +
           "         chosenAction = 0;                              \n" +
           "    else if ( chosenAction > 11 )                       \n" +
           "              chosenAction = 11;                        \n" +
           "    switch( type )                                      \n" +
           "    {                                                   \n" +
           "        case 0:  // action should be returned for this step. \n" +
           "        case 1:                                         \n" +
           "             curAction = chosenAction;                  \n" +
           "             return curAction;                          \n" +
           "        case 2:  // action should be returned for next step. \n" +
           "        case 3:                                         \n" +
           "             curAction = nextAction;                    \n" +
           "             nextAction = chosenAction;                 \n" +
           "             return curAction;                          \n" +
           "    }                                                   \n" +
           "    return curAction;                                   \n" +
           " }                                                      \n" +
           "}               ",
           "the default Fixed Rule condition weights",
           "the default Fixed Rule rule weight",
           "other Fixed Rule utility measures in this option",
           "public short[][] getFrCondition(int groupIdx, int netIdx, int ruleIdx)      \n" +
           "{                              \n" +
           "    //*                        \n" +
           "     * Suppose the input has 4 dimensions, one possible condtion of a Fixed Rule: \n" +
           "     * { {1,1,1,1}, {1,0,1}, {0,0,1}, {1,0} }                      \n" +
           "     *//                                                           \n" +
           "   short[][] conditions = { {1,1,1,1}, {1,0,1}, {0,0,1}, {1,0} };  \n" +
           "   return conditions;                                              \n" +
           "}                                ",

           "public short[][][] getFrAction(int groupIdx, int netIdx, int ruleIdx) \n" +
           "{                                                                  \n" +
           "   short[][][] action = {{ {0,1}, {0}, {2,3}, {0}}};               \n" +
           "   return action;                                                  \n" +
           "}                                                                  \n"
        },
        {
           "null",
           "null",
           "the default associative rule condition weights",
           "the default dimension weights in a chunk",
           "the default similarity measure in this option",
           "the default data pattern - experienced state"
        },
        {
           "the default overall response time formula",
           "the default overall bottome-level response time formula",
           "the default overall top-level response time formula",
           "the default top-level perceptual time formula",
           "the default top-level decision time formula",
           "public int getAT_TL( int action, int assocAppTime, int chunkRetrTime ) \n" +
           "{                                                 \n" +
           "  return 100 + 5 * assocAppTime + chunkRetrTime;  \n" +
           "}                                                   ",
           "the default associative rule application time formula",
           "the default chunk retrieval time formula",
           "the default AMN one-pass time formula"
        },
        {
           "see the deafault drive set."
        }
  };

  final static Sample[][] samples = new Sample[SUB_SYS_NUM][];{
        samples[0] = new Sample[ACS_SAMPLE_NUM];
        samples[1] = new Sample[NACS_SAMPLE_NUM];
        samples[2] = new Sample[RT_SAMPLE_NUM];
        samples[3] = new Sample[MCS_SAMPLE_NUM];
        for( int i = 0; i < SUB_SYS_NUM; i++ )
           for( int j = 0; j < samples[i].length; j++ ){
                samples[i][j] = new Sample( className, methodNames[i][j], methodForms[i][j],
                                descriptions[i][j], params[i][j], returns[i][j], seeSamples[i][j]);
                //System.out.println( samples[i][j] );
           }
  }
  /** the associated <tt>Global</tt> object. */
  protected Global global;

  public TaskClarion()
  {
  }

  /** Returns the specified sample code.
   *  @param subSysIdx the index on CLARION subsystem.
   *  @param sampleIdx the index on the sample within the specified subsystem.
   *  @return the specified sample code in the format of <tt>String</tt>.
   */
  public static String getSample( int subSysIdx, int sampleIdx )
  {
     return samples[subSysIdx][sampleIdx].toString();
  }

  public static String getExplanation( int subSysIdx, int sampleIdx )
  {
     return explanations[subSysIdx][sampleIdx].toString();
  }

  public void setGlobal( Global global )
  {
     this.global = global;
  }

  //============= the followings need be override by user's own codes. =================/
  /** Returns the goals. The returned value is in the format of 2-dimenstional array.
   *  The first dimension indexes a specific goal and the second dimension indexes a
   *  specific goal dimension. We assume that a goal is composed by an array of values
   *  each corresponding to a specific goal dimension.
   *  @return the goal set. */
  public short[][] getGoals()
  {
     return (short[][])null;
  }

  /** Returns the goal dimension-value info. The returned value is in the foramt of one
   *  dimensional array each slot storing the corresponding dimension value #.
   *  @return the goal dimension-value info */
  public short[] getGoalDimDVs()
  {
     return (short[])null;
  }

  /** bottom-up rectification process described in the Chapter on ACS in tutorial. */
  public void combineByBottomUpRectification()
  {
  }

  /** top-down guidance process described in the Chapter on ACS in tutorial. */
  public void combineByTopDownGuidance()
  {
  }

  /** Returns the eligible conditions of the specified network. The return value is in
   *  the format of 3-dimensional array. The first dimension indexes an eligible condition,
   *  the second dimension indexes an involved input dimension and the third dimnesion
   *  indexes a value in the involved dimension.
   *  Note: the first element of the second dimension stores all of the involved input
   *  dimensions.
   *  @param netIdx the ACS network NO.
   *  @return the eligible condition of the specified network.
   */
  public short[][][] getEligibility( int netIdx )
  {
     /**
      * suppose the input has 4 dimensions and each dimension has 5 values.
      * Eligibility of net 0 is :
      * if ((c1 = v11 or v12) and (c2 > v23)) or ((c3 <= v34) and (c4 >= v42)))
      * then use net 0.
      * ( vij means the jth value of dimension i. )
      */
    short[][][] eligCondition = {
                   {
                      {0,1},          // relevant dims: 0, 1, in this condition.
                      {1,1,0,0,0},    // c1 = v11 or v12.
                      {0,0,0,1,1}     // c2 > v23.
                   },
                   {
                      {2,3},          // relevant dims: 2, 3, in this condition.
                      {1,1,1,1,0},    // c3 <= v34.
                      {0,1,1,1,1}     // c4 >= v42.
                   }
                };
     return eligCondition;
  }

  /** Returns the reinforcement of the specified ACS network.
      @param  netIdx   the network NO.
      @param  input   current input,
      @param  wmChunks   current working memory status.
      @param  goal  current goal.
      @param  action  current action.
      @param  newInput   new input resulting from firing the action.
      @param  newWMChunks   new working memory status.
      @param  newGoal   new goal.
      @return the reinforcement of the specified ACS network.
   */
  public double getReinforcement( int netIdx, Feature[] input, Chunk[] wmItems, GoalChunk goalItem,
                short[][][] action, Feature[] newInput, Chunk[] newWmItems, GoalChunk newGoalItem )
  {
     int TARGET = 5;
     int P = newInput[0].getValues()[0];
     int diff = Math.abs( P - TARGET );
     double fb = 0.0;
     switch( diff ){
         case 0:
         case 1:
              fb = 1.0;
              break;
         default:
              fb = -0.2 * ( diff - 1 );
              break;
     }
     return fb;
  }

  /** Returns the RER dimensional weights used to calculate rule support. The returned
   *  value is in one dimensional array storing the weights of the input dimensions.
   *  Note: the sum of the weights should be 1.0.
   *  @param netIdx the ACS network NO.
   *  @return the dimensional RER weights.
   */
  public double[] getRerWeights( int netIdx )
  {
     int dimNum = global.getARInputDimNum( netIdx, Global.RER );
     double[] weights = new double[dimNum];
     for( int i = 0; i < dimNum; i++ )
          weights[i] = 1.0/dimNum;
     return weights;
  }

  /** Returns the utility of a specified set of RER rules.
   *  @param netIdx   the network NO.
      @param setIdx   the rule subset NO.
      @param PM Positive Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
      @param NM Negative Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
      @return the utility of a specified set of RER rules.
   */
  public double getRerUtility( int netIdx, int setIdx, double PM, double NM )
  {
     double cost = 1.0, v = 1.0, benefit = 2.0, c7, c8, utl = 0.0;
     int option = global.getRuleUtilityOption( netIdx, Global.RER, setIdx );
     if( option == Global.CONSTANT_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_V];
         benefit = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_BENE];
     }
     if( option == Global.VARIED_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_V];
         c7 = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_C7];
         c8 = global.getRuleUtilityParas(netIdx,Global.RER,setIdx)[Global.UTL_C8];
         benefit = (c7 + PM) / (c8 + PM + NM );
     }
     utl = benefit - v * cost;
     return utl;
  }

  /** Returns <tt>true</tt> if current step is positive or not according to RER positivity
   *  criterion, <tt>false</tt> otherwise.
      @param learnOption   learning method.
      @param reward   currently received reward.
      @param qDiscount   the discount factor for Q-Learning that favors
        reinfocement received sooner relative to that received later.
        used in q-learning method.
      @param newMaxQVal   maximum Q-value of the new state resuling from
        current action in current state, Q(next state, next possible action).
        used in q-learning method.
      @param qVal   current q-value, Q(current state, current action).
      @param threshold   the threshold for rule positivity.
      @return if current step is positive or not according to RER positivity.
   */
  public boolean getRerPositivity( int learnOption, double reward, double qDiscount,
                                   double newMaxQVal, double qVal, double threshold )
  {
     boolean positivity = false;
     if( learnOption == Global.QL ) {
         if( reward + qDiscount * newMaxQVal - qVal > threshold )
             positivity = true;
         else
             positivity = false;
     }
     else if( learnOption == Global.SQL ){
              if( reward > threshold )
                  positivity = true;
              else
                  positivity = false;
          }
          else{
              if( reward > 0 )
                  positivity = true;
              else
                  positivity = false;
          }
     return positivity;
  }

 /** Returns the Information Gain according to the RER Information Gain criterion.
  * @param netIdx   the network NO. decdies the values of c1 and c2.
  * @param PMa   Positive Match of the rule a, equals the number of times
    that an input state matches rule condition, rule action is performed
    and the result is positive.
    @param NMa   Negative Match of the rule a, equals the number of times
    that an input state matches rule condition, rule action is performed
    and the result is negative.
    @param PMb   Positive Match of the rule b.
    @param NMb   Negative Match of the rule b.
    @return the Information Gain.
  */
  public double getRerIG( int netIdx, double PMa, double NMa, double PMb, double NMb )
  {
      double c1 = 1.0, c2 = 2.0;
      c1 = global.getRerIGParaAbove( netIdx );
      c2 = global.getRerIGParaUnder( netIdx );
      double log2 = Math.log( 2 );
      double rateA = (PMa + c1) / ( PMa + NMa + c2 );
      double rateB = (PMb + c1) / ( PMb + NMb + c2 );
      double ig = Math.log( rateA / rateB ) / log2;
      return ig;
  }

  /** Returns the IRL dimensional weights used to calculate rule support. The returned
   *  value is in one dimensional array storing the weights of the input dimensions.
   *  Note: the sum of the weights should be 1.0.
   *  @param netIdx the ACS network NO.
   *  @return the dimensional IRL weights.
   */
  public double[] getIrlWeights( int netIdx )
  {
     int dimNum = global.getARInputDimNum( netIdx, Global.IRL );
     double[] weights = new double[dimNum];
     for( int i = 0; i < dimNum; i++ )
          weights[i] = 1.0/dimNum;
     return weights;
  }

  /** Returns the utility of a specified set of IRL rules.
   *  @param netIdx   the network NO.
      @param setIdx   the rule subset NO.
      @param PM   Positive Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
      @param NM   Negative Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
      @return the utility of a specified set of IRL rules.
   */
  public double getIrlUtility( int netIdx, int setIdx, double PM, double NM )
  {
     double cost = 1.0, v = 1.0, benefit = 2.0, c7, c8, utl = 0.0;
     int option = global.getRuleUtilityOption( netIdx, Global.IRL, setIdx );
     if( option == Global.CONSTANT_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_V];
         benefit = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_BENE];
     }
     if( option == Global.VARIED_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_V];
         c7 = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_C7];
         c8 = global.getRuleUtilityParas(netIdx,Global.IRL,setIdx)[Global.UTL_C8];
         benefit = (c7 + PM) / (c8 + PM + NM );
     }
     utl = benefit - v * cost;
     return utl;
  }

  /** Returns <tt>true</tt> if current step is positive or not according to IRL positivity
   *  criterion, <tt>false</tt> otherwise.
      @param learnOption   learning method.
      @param reward   currently received reward.
      @param qDiscount   the discount factor for Q-Learning that favors
        reinfocement received sooner relative to that received later.
        used in q-learning method.
      @param newMaxQVal   maximum Q-value of the new state resuling from
        current action in current state, Q(next state, next possible action).
        used in q-learning method.
      @param qVal   current q-value, Q(current state, current action).
      @param threshold   the threshold for rule positivity.
      @return if current step is positive or not according to IRL positivity.
   */
  public boolean getIrlPositivity( int learnOption, double reward, double qDiscount,
                                   double newMaxQVal, double qVal, double threshold )
 {
     boolean positivity = false;
     if( learnOption == Global.QL ) {
         if( reward + qDiscount * newMaxQVal - qVal > threshold )
             positivity = true;
         else
             positivity = false;
     }
     else if( learnOption == Global.SQL ){
              if( reward > threshold )
                  positivity = true;
              else
                  positivity = false;
          }
          else {
              if( reward > 0 )
                  positivity = true;
              else
                  positivity = false;
          }
     return positivity;
  }

  /** Returns the Information Gain according to the IRL Information Gain criterion.
  * @param netIdx   the network NO. decdies the values of c5 and c6.
    @param PM   Positive Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
    @param NM   Negative Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
    @param threshold   compared threshold value.
    @return the Information Gain.
  */
  public double getIrlIG( int netIdx, double PM, double NM, double threshold )
  {
      double c5 = 1.0, c6 = 2.0;
      c5 = global.getIrlIGParaAbove( netIdx );
      c6 = global.getIrlIGParaUnder( netIdx );
      double log2 = Math.log( 2 );
      double rate = (PM + c5) / ( PM + NM + c6 );
      double ig = Math.log( rate / threshold ) / log2;
      return ig;
  }

  /** Returns an IRL rule form corresponding to the rule index.
   *  @param groupIdx  agent group index.
   *  @param netIdx  ACS network index.
   *  @param setIdx  rule subset index.
   *  @param ruleIdx rule index.
   *  @return an IRL rule form corresponding to the rule index.
   */
  public IrlRuleForm getIrlRule( int groupIdx, int netIdx, int setIdx, int ruleIdx )
  {
   /**
    * In Process Control task, there are four sets of IRL rules.
       1. P = a * W + b
       2. P = a * W + c * P1 + b
       3. P = a * W1 + b
       4. P = a * W1 + c * P2 + b
       a = {1,2},  b = {-1,-2,0,1,2},  c = {-1,-2,1,2}.
       P = the desired system output level(the goal),
       W = the current input to the system(to be determined),
       W1 = the previous input to the system,
       P1 = the previous system output level(under W1),
       P2 = the system output level at the time step before P1.
       In this case, elements from the set a, b, c compose the IRL
       parameters.

     smaple code is the following:

     public class IrlRuleFormPC implements IrlRuleForm{
          int a, b, c, target, ruleSetIdx;
          int curAction, nextAction;

          public IrlRuleFormPC( int a, int b, int c, int target, int ruleSetIdx, int curAction )
          {
                this.a = a;
                this.b = b;
                this.c = c;
                this.target = target;
                this.ruleSetIdx = ruleSetIdx;
                this.curAction = curAction;
                nextAction = ( int ) ( target - b ) / a;
          }

          public double determineAction( short[] input, short[] suggestedAction )
          {
             int p1 = 0;
             for( int i = 0; i < input.length; i++ )
                  if( input[i] == 1 ){
                      p1 = i;
                      break;
                  }
             int chosenAction = Math.round( ( target - c * p1 - b ) / a );
             if ( chosenAction < 0 )
                  chosenAction = 0;
             else if ( chosenAction > 11 )
                       chosenAction = 11;
             switch( ruleSetIdx )
             {
                 case 0:  // action should be returned for this step.
                 case 1:
                      curAction = chosenAction;
                      suggestedAction[0] = (short)curAction;
                      break;
                 case 2:  // action should be returned for next step.
                 case 3:
                      curAction = nextAction;
                      nextAction = chosenAction;
                      suggestedAction[0] = (short)curAction;
                      break;
             }
             return 1.0;
          }

          public void generalize()
          {
          }

          public void specialize()
          {
          }
     }
     */
     return (IrlRuleForm)null;
  }

  /** Returns the FR dimensional weights used to calculate rule support. The returned
   *  value is in one dimensional array storing the weights of the input dimensions.
   *  Note: the sum of the weights should be 1.0.
   *  @param netIdx the ACS network NO.
   *  @return the dimensional FR weights.
   */
  public double[] getFrWeights( int netIdx )
  {
     int dimNum = global.getARInputDimNum( netIdx, Global.FR );
     double[] weights = new double[dimNum];
     for( int i = 0; i < dimNum; i++ )
          weights[i] = 1.0/dimNum;
     return weights;
  }

  /** Returns the utility of a specified set of FR rules.
   *  @param netIdx   the network NO.
      @param setIdx   the rule subset NO.
      @param PM   Positive Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is positive.
      @param NM   Negative Match of the rule, equals the number of times
        that an input state matches rule condition, rule action is performed
        and the result is negative.
      @return the utility of a specified set of FR rules.
   */
  public double getFrUtility( int netIdx, int setIdx, double PM, double NM )
  {
     double cost = 1.0, v = 1.0, benefit = 2.0, c7, c8, utl = 0.0;
     int option = global.getRuleUtilityOption( netIdx, Global.FR, setIdx );
     if( option == Global.CONSTANT_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_V];
         benefit = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_BENE];
     }
     if( option == Global.VARIED_OPTION ) {
         cost = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_COST];
         v = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_V];
         c7 = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_C7];
         c8 = global.getRuleUtilityParas(netIdx,Global.FR,setIdx)[Global.UTL_C8];
         benefit = (c7 + PM) / (c8 + PM + NM );
     }
     utl = benefit - v * cost;
     return utl;
  }

  /** Returns <tt>true</tt> if current step is positive or not according to FR positivity
   *  criterion, <tt>false</tt> otherwise.
      @param learnOption   learning method.
      @param reward   currently received reward.
      @param qDiscount   the discount factor for Q-Learning that favors
        reinfocement received sooner relative to that received later.
        used in q-learning method.
      @param newMaxQVal   maximum Q-value of the new state resuling from
        current action in current state, Q(next state, next possible action).
        used in q-learning method.
      @param qVal   current q-value, Q(current state, current action).
      @param threshold   the threshold for rule positivity.
      @return if current step is positive or not according to FR positivity.
   */
  public boolean getFrPositivity( int learnOption, double reward, double qDiscount,
                                   double newMaxQVal, double qVal, double threshold )
  {
     boolean positivity = false;
     if( learnOption == Global.QL ) {
         if( reward + qDiscount * newMaxQVal - qVal > threshold )
             positivity = true;
         else
             positivity = false;
     }
     else if( learnOption == Global.SQL ){
              if( reward > threshold )
                  positivity = true;
              else
                  positivity = false;
          }
          else{
              if( reward > 0 )
                  positivity = true;
              else
                  positivity = false;
          }
     return positivity;
  }

  /** Returns a FR condition. The returned value is in the format of 2-dimenisonal
   *  array. The first dimension indexes an dimenison and the second indexes one
   *  deimensional value. The value in each slot of the returned array:
   *  1, if the dimensional value is in the condition,
   *  0, otherwise.
   *  @param groupIdx agent group index.
   *  @param netIdx ACS network index.
   *  @param ruleIdx rule index.
   *  @return a FR condition. */
  public short[][] getFrCondition(int groupIdx, int netIdx, int ruleIdx)
  {
     short[][] conditions = { {1,1,1,1}, {1,0,1,0}, {0,0,0,1}, {1,0,1,0} };
     return conditions;
  }

  /** Returns FR actions. The returned value is in the format of 3-dimenisonal array.
   *  The first dimension indicates action type :
   *  index = 0 : normal action, index = 1: NACS control action.
   *  The second dimension indexes an action dimension.
   *  The third stores the active values in that dimension.
   *  @param groupIdx learniing group the agent is in.
   *  @param netIdx the net index.
   *  @param ruleIdx the rule index.
   *  @return the action of the specific fixed rule. */
  public short[][][] getFrAction(int groupIdx, int netIdx, int ruleIdx)
  {
     short[][][] action = {{ {0,1}, {0}, {2,3}, {0}}};
     return action;
  }

  /** Decides action by a Fixed Rule. It is used for the action (of a Fixed Rule)
   *  which is variable and depends on current input (EX + GS + WM).
   *  Fills in the suggestedAction under current input and returns the rule
   *  support for current input.
   *  @param groupIdx learniing group the agent is in.
   *  @param netIdx the net index.
   *  @param setIdx the rule set index.
   *  @param sensoryInput current sensory input.
   *  @param gsItem the top goal item.
   *  @param wmItems current WM content.
   *  @param sugggestedAction the array to store the suggested action.
   *  @return the reinforcement. */
  public double decideFrAction( int groupIdx, int netIdx, int setIdx, Feature[] sensoryInput,
                GoalChunk gsItem, Chunk[] wmItems, short[][][] suggestedAction )
  {
     /**
      *  sample code of PC task.
      *  target: the target production.
      *  begin:  the start pooint of P1 dimension in the input array.
      *  end:    the end pooint of P1 dimension in the input array.
      *  idx:    current P1 value index on the P1 dimension.
      */
     int target = 5, begin = 0, end = 11, idx = 0;
     short[] state = sensoryInput[0].getContent();
     for( int i = begin; i <= end; i++ ){
          if( state[i] == 1 ){
              idx = i;
              break;
          }
     }
     short[] action = new short[1];
     action[0] = (short)(( idx + target ) / 2);
     suggestedAction[Global.ACTION_NORMAL][0] = action;
     return 1.0;
  }

  /** Returns the externally given NACS chunks.
   *  @return the externally given NACS chunks. */
  public GKSChunk[] getExternalNacsChunks()
  {
     return (GKSChunk[])null;
  }

  /** Returns the externally given associative rules.
   *  @return the externally given associative rules. */
  public AssocRule[] getExternalNacsRules()
  {
     return (AssocRule[]) null;
  }

  /** Returns the weights for computing associative rule support from all of the
   *  condition chunks. the sum of the weights should be 1.0.
   *  @return the weights in the format of an array each indexes a condition. */
  public double[] getAssocWeights( int condNum )
  {
     double[] weights = new double[condNum];
     for( int i = 0; i < condNum; i++ )
          weights[i] = 1.0/condNum;
     return weights;
  }

  /** Returns the weights for computing associative rule support from one condition
   *  chunk. the sum of the weights should be 1.0.
   *  @return the weights in the format of an array each indexes a dimension. */
  public double[] getAssocDimWeights( int dimNum )
  {
      /**
       * the sum of the weights should be 1.0.
       */
     //int dimNum = global.getGksInputDimNum();
     double[] weights = new double[dimNum];
     for( int i = 0; i < dimNum; i++ )
          weights[i] = 1.0/dimNum;
     return weights;
  }

  /** Returns the similarity between two chunks. The chunk similarity measure is
   *  similar to: # of equal dimensions between chunk1 and chunk2 / # of dimensions in chunk2.
   *  @param chunk1 the source chunk.
   *  @param chunk2 the destination chunk.
   *  @return the similarity. */
  public double getSimilarity(GKSChunk chunk1, GKSChunk chunk2)
  {
       double similarity = 0.0;
       if( chunk1 == null || chunk2 == null )
           return similarity;
       // calculate similarity.
       Feature[] featList1 = chunk1.getFeatures();
       Feature[] featList2 = chunk2.getFeatures();
       int matchCount = 0;
       for( int i = 0; i < featList1.length; i++ ){
            // search the index of the ith feature in list1 on list2.
            int idx = chunk2.searchFeature( featList1[i].getLabel() );
            if( idx == -1 || featList1[i].getType() != featList2[idx].getType() )
                continue;
            short[] arr1 = featList1[i].getValues();
            short[] arr2 = featList2[idx].getValues();
            Arrays.sort( arr1 );
            Arrays.sort( arr2 );
            if( Arrays.equals( arr1, arr2 ) )
                matchCount++;
       }
       // similarity-based reasoning, use the default similarity measure.
       double val = Math.pow( featList2.length, 1.001 );
       similarity = (double)matchCount / val;
       return similarity;
  }

  /** Returns the types of data pettern for training AAM.
   *  @param netIdx the AAM net index.
   *  @param chunkIdx the chunk the rule is associated with.
   *  @param ruleIdx the rule index on the rule group associated with the chunk.
   *  @return the types of data pettern for training AAM. */
  public short[] getAamDataPattern(int netIdx, int chunkIdx, int ruleIdx)
  {
     short[] types = {0};
     return types;
  }

  /** new routine. Performs encoding external knowledge in NACS. */
  public void performEncodeExternalKnowledge( short[][] action )
  {
  }

  /** new routine. Performs the assimilation in NACS. */
  public void performNacsAssimilation( short[][] action )
  {
  }

  /** new routine. Gets retreival results from NACS. */
  public Object[] getNacsResults()
  {
     return (Object[])null;
  }

  /** Returns the overall response time computed by the formula use defines.
   *  @param RT_BL response time from Bottom Level.
   *  @param RT_TL response time from Top Level.
   *  @return overall response time. */
  public int getOverallRT( int RT_BL, int RT_TL )
  {
     return Math.max( RT_BL, RT_TL );
  }

  /** Returns the overall Bottom Level response time computed by the formula use
   *  defines.
   *  @param  PT_BL Bottom-Level Perceptual Time.
      @param  DT_BL Bottom-Level Decision Time.
      @param  AT_BL Bottom-Level Actuation Time.
      @return the overall Bottom Level response time. */
  public int getBLOverallRT( int PT_BL, int DT_BL, int AT_BL )
  {
     return PT_BL + DT_BL + AT_BL;
  }

  /** Returns the overall Top Level response time computed by the formula use
   *  defines.
   *  @param  PT_BL Top-Level Perceptual Time.
      @param  DT_BL Top-Level Decision Time.
      @param  AT_BL Top-Level Actuation Time.
      @return the overall Top Level response time. */
  public int getTLOverallRT( int PT_TL, int DT_TL, int AT_TL )
  {
     return PT_TL + DT_TL + AT_TL;
  }

  /** Returns Top-Level Perceptual Time.
   *  @param  PT_BL Bottom-Level Perceptual Time.
   *  @return Top-Level Perceptual Time. */
  public int getPT_TL( int PT_BL )
  {
     return PT_BL + 150;
  }

  /** Returns the Top-Level Decision Time.
    * @param  DT_BL Bottom-Level Decision Time.
    * @param  actionTime time determined by the operation dictated by
      a particular rule.
      @param  ruleBla rule base-level activation.
      @param  chunkBla action chunk base-level activation.
      @return the Top-Level Decision Time. */
  public int getDT_TL( int DT_BL, int actionTime, double ruleBla, double chunkBla )
  {
     double t0 = global.getDT_TL_t0();
     double t1 = global.getDT_TL_t1();
     int TLDT = (int)( actionTime + t0 / ruleBla + t1 / chunkBla );
     return TLDT;
  }

  /** Returns the Top-Level Actuation Time.
   * @param flag decides if NACS is involved or not.
     @param assocAppTime time of one iteration of associative applicaiton.
     @param chunkRetrTime time of one iteration of chunk retrieval.
     @returns the Top-Level Actuation Time. */
  public int getAT_TL( boolean flag, int assocAppTime, int chunkRetrTime )
  {
     if( flag )
         return 200 + assocAppTime + chunkRetrTime;
     else
         return 500;
  }

  /** Returns associative rule application time.
   *  @param  assocRuleBla associative rule base-level activation.
   *  @return associative rule application time.
   */
  public int getAssocAppTime( double assocRuleBla )
  {
      double t4 = global.getAssocAppTime_t4();
      double t5 = global.getAssocAppTime_t5();
      int assocRuleAppTime = (int)(t4 + t5 / assocRuleBla );
      return assocRuleAppTime;
  }

  /** Returns the NACS chunk retrieval time.
   *  @param  chunkBla - nacs chunk base-level activation.
   *  @return the NACS chunk retrieval time.  */
  public int getChunkRetrTime( double chunkBla )
  {
      double t2 = global.getChunkRetTime_t2();
      double t3 = global.getChunkRetTime_t3();
      int chunkRetTime = (int)(t2 + t3 / chunkBla);
      return chunkRetTime;
  }

  /** Returns AMN one-pass time.
   *  @return AMN one-pass time. */
  public int getAmnPassTime()
  {
     return global.getAmnOnePassTime();
  }

  class Sample{
     String cls;
     String method;
     String form;
     String description;
     String params;
     String returns;
     String seeSample;

     Sample( String cls, String method, String form, String description, String params,
             String returns, String seeSample )
     {
        this.cls = cls;
        this.method = method;
        this.form = form;
        this.description = description;
        this.params = params;
        this.returns = returns;
        this.seeSample = seeSample;
     }

     public String toString()
     {
        String str = "";
        str += "USER CLASS : \n" + cls + "\n\n";
        str += "USER METHOD : \n" + method + "\n\n";
        str += "METHOD CALLING FORM : \n" + form + "\n\n";
        str += "METHOD DESCRIPTION : \n" + description + "\n\n";
        str += "PARAMETERS : \n" + params + "\n\n";
        str += "RETURNS : \n" + returns + "\n\n";
        str += "SAMPLES : \n" + seeSample + "\n\n";
        return str;
     }
  }
}
