package tasksimulators.newsimu.AGL;
import java.util.Vector;
import java.util.Stack;
import java.util.Arrays;
import clarion.system.*;
import clarion.common.*;
import clarion.nacs.*;
import clarion.tools.*;
import clarion.acs.gs.*;

/**
 * <ul>
 * <li> General overview.
 * <ul>
 * <li> summary of this class <br>
 * This class extends the <tt>TaskClarion</tt> class and implements all
 * of the necessary routines specific to tje Artificial Intelligence Learning task.
 * <li> summary of the important methods. <br>
 * <ol>
 *  <li> public double decideFrAction( int groupIdx, int netIdx, int setIdx, Feature[] sensoryInput,
 * GoalChunk gsItem, Chunk[] wmItems, short[][][] suggestedAction ) :
 * decides action by a Fixed Rule.
 * <li> public void performEncodeExternalKnowledge( short[][] action ) :
   Perform the process of encoding external knowledge.
 * <li> public void performNacsAssimilation( short[][] action ) :
   Performs the process of assimilation in NACS.
 *  <li> public void combineByBottomUpRectification(): Performs the Bottom-Up
 * Rectification combination.
 *  <li> public void combineByTopDownGuidance(): Performs the Top-down Guidance
 * combination.
 * </ol>
 * </ul>
 * <li> Role in the CLARION model <br>
 * It is a part of the AGL task simulator.
 * <li> Usage of this class. <br>
 * This class is instantiated inside a <tt>TaskAgent</tt> object.
 * </ul>
 * @author Xi Zhang
 */

public class TaskClarionAGL extends TaskClarion{

  /** the variables to represent the learning group. */
  final int CONTROL_TYPE = 0, IMPLICIT_TYPE = 1;
  final int EXPLICIT_TYPE = 2, SYNERGY_TYPE = 3;
  /** used for constructing of associative rules in AGL task. */
  final int COND_NUM = 1;
  /** Numbers of iteration of training. */
  final int ITER_NUMS[] = { 1, 1, 1, 1 };
  /** the probability of associative rule application. */
  final double RULE_APP_PROB = 1.0;
  //final double RULE_APP_PROB = 0.9;
  //final double encodeProb = 0.16;
  final double encodeProb = 0.23;
  short[][] chunkMap;
  /** the active AMN network index. */
  int activeAmnIdx;
  /** current learning group. */
  int learningType;
  /** variables for implementing the NACS-related process in AGL task. */
  int searchNum;
  int totSearchNum;
  Feature[] gksInput;
  Feature[] amnInput;
  Feature[] completion;
  Feature[] gksOutput;
  Feature[] amnOutput;
  Feature[] trainSample;
  Feature[] testSample;
  Feature[][] beginFeatures;
  /** the General Knowledge Store. */
  GKS gks;
  /** the Associative Memory Networks. */
  AMNet amns[];
  /** the variables for I/O dimension info. */
  final int MAX_STR_LEN = 11, MIN_STR_LEN = 5, SAMPLE_NUM = 177;
  final int INPUT_DIM_NUM = MAX_STR_LEN + 2, OUTPUT_DIM_NUM = MAX_STR_LEN + 1;
  final int INPUT_NUM = 83, OUTPUT_NUM = 89, ACTION_NUM = 7;
  final int LETTER_DIM_LEN = 7, LEARN_TYPE_NUM = 4, STAGE_NUM = 2, POSITION_NUM = MAX_STR_LEN + 1;
  final int L_CON = 0, L_IMP = 1, L_EMP = 2, L_SYN = 3;
  final int S_TRAIN = 0, S_TEST = 1;
  /** the conditions of the Fixed Rules. */
  short[] frConditions[][][];
  /** the actions of the Fixed Rules. */
  short[] frActions[][][][];
  {  // initialize FR rules.
     frConditions = new short[1][ACTION_NUM][INPUT_DIM_NUM][];
     frActions = new short[1][ACTION_NUM][][][];
     short[][][] actionConditions = {
       {{L_CON},{S_TRAIN}}, {{L_IMP},{S_TRAIN}}, {{L_EMP},{S_TRAIN}}, {{L_SYN},{S_TRAIN}},
       {{L_CON,L_IMP},{S_TEST}}, {{L_EMP},{S_TEST}}, {{L_SYN},{S_TEST}}
     };
     for( int i = 0; i < ACTION_NUM; i++ ){
       for( int j = 0; j < INPUT_DIM_NUM-2; j++ ){
            frConditions[0][i][j] = new short[ LETTER_DIM_LEN ];
            for( int k = 0; k < frConditions[0][i][j].length; k++ )
                 frConditions[0][i][j][k] = 1;
       }
       frConditions[0][i][INPUT_DIM_NUM-2] = new short[LEARN_TYPE_NUM];
       for( int j = 0; j < actionConditions[i][0].length; j++ )
            frConditions[0][i][INPUT_DIM_NUM-2][actionConditions[i][0][j]] = 1;
       frConditions[0][i][INPUT_DIM_NUM-1] = new short[STAGE_NUM];
       for( int j = 0; j < actionConditions[i][1].length; j++ )
            frConditions[0][i][INPUT_DIM_NUM-1][actionConditions[i][1][j]] = 1;
     }
  }

  /* the default constructor. The system call call this constructor to instantiate
     an instance of this class when AGL task is being loaded. */
  public TaskClarionAGL() {}

  /** Gets the NACS retrieval results.
   *  @return the results. */
  public Object[] getNacsResults()
  {
     return completion;
  }

  /** Reset the NACS result buffer. */
  public void resetNacsResults()
  {
     completion = null;
  }

  /** Set the GKS. */
  public void setGks()
  {
     gks = global.getNACS().getGks();
  }

  /** Set the AMN networks. */
  public void setAmns()
  {
     amns = global.getNACS().getAmns();
  }

  /** Set the learning type. */
  public void setLearningType( int type )
  {
     learningType = type;
  }

  /** Reset the search number in the searching process. */
  public void resetTotSearchNum()
  {
     totSearchNum = 0;
  }

  /** Sets the sample for training NACS.
   * @param curNacsInput the sample for training. */
  public void setTrainSample( Feature[] curNacsInput )
  {
     trainSample = curNacsInput;
  }

  /** Sets the sample for testing NACS.
   * @param curNacsInput the sample for testing. */
  public void setTestSample( Feature[] curNacsInput )
  {
     testSample = curNacsInput;
  }

  /** Sets the features as begin tag of reasoning.
   * @param features the features to set as begin tags.*/
  public void setBeginFeatures( Feature[][] features )
  {
     beginFeatures = new Feature[ features.length ][];
     for( int i = 0; i < features.length; i++ ){
       beginFeatures[i] = new Feature[ features[i].length ];
       for( int j = 0; j < features[i].length; j++ )
            beginFeatures[i][j] = (Feature)features[i][j].clone();
     }
  }

  /** Returns the reinforcement of the specified ACS network.
      @param  netIdx   the network NO.
      @param  input   current input,
      @param  wmChunks   current working memory status.
      @param  goal  current goal.
      @param  action  current action.
      @param  newInput   new input resulting from firing the action.
      @param  newWMChunks   new working memory status.
      @param  newGoal   new goal.
      @return the reinforcement of the specified ACS network.  */
  public double getReinforcement( int netIdx, Feature[] input, Chunk[] wmItems, GoalChunk goalItem,
                short[][][] action, Feature[] newInput, Chunk[] newWmItems, GoalChunk newGoalItem )
  {
     return 0.9;
  }

  /** Returns a FR condition. The returned value is in the format of 2-dimenisonal
   *  array. The first dimension indexes an dimenison and the second indexes one
   *  deimensional value. The value in each slot of the returned array:
   *  1, if the dimensional value is in the condition,
   *  0, otherwise.
   *  @param groupIdx agent group index.
   *  @param netIdx ACS network index.
   *  @param ruleIdx rule index.
   *  @return a FR condition. */
  public short[][] getFrCondition(int groupIdx, int netIdx, int ruleIdx)
  {
     return CommonMethods.copy( frConditions[netIdx][ruleIdx] );
  }

  /** Returns FR actions. The returned value is in the format of 2-dimenisonal
   *  array. The first dimension indexes an action dimension and the second dimension
   *  stores the active values in that dimension.
   *  Suppose the output has 4 dimensions and each dimension has four values.
   *  one sample of action : short[][] action = { {0,1}, {0}, {2,3}, {0}};
   *  @param ther FR actions.  */
  public short[][][] getFrAction(int groupIdx, int netIdx, int ruleIdx)
  {
    return CommonMethods.copy( frActions[netIdx][ruleIdx] );
  }

  /** Decides action by a Fixed Rule. It is used for the action (of a Fixed Rule)
   *  which is variable and depends on current input (EX + GS + WM).
   *  Fills in the suggestedAction under current input and returns the rule
   *  support for current input.
   *  @param groupIdx learniing group the agent is in.
   *  @param netIdx the net index.
   *  @param setIdx the rule set index.
   *  @param sensoryInput current sensory input.
   *  @param gsItem the top goal item.
   *  @param wmItems current WM content.
   *  @param sugggestedAction the array to store the suggested action.
   *  @return the reinforcement. */
  public double decideFrAction( int groupIdx, int netIdx, int setIdx, Feature[] sensoryInput,
                                GoalChunk gsItem, Chunk[] wmItems, short[][][] suggestedAction )
  {
     final int INPUT_DIM_NUM = 13;
     final short L_CON = 0, L_IMP = 1, L_EMP = 2, L_SYN = 3;
     final short S_TRAIN = 0, S_TEST = 1;

     short[][] formalAction = suggestedAction[Global.ACTION_NORMAL];
     for( int i = 0; i < formalAction.length; i++ ){
          formalAction[i] = CommonMethods.copy( (sensoryInput[i].getValueNum()==0)?null:sensoryInput[i].getValues() );
     }
     short[][] controlAction = suggestedAction[Global.ACTION_CONTROL];
     for( int i = 0; i < controlAction.length; i++ ){
          if( i == 6 || i == 8 ){
              controlAction[i] = new short[MAX_STR_LEN];
              for( int j = 0; j < MAX_STR_LEN; j++ )
                   controlAction[i][j] = (short)j;
          }
          else if( i == 7 || i == 9 )
                   controlAction[i] = null;
               else
                   controlAction[i] = new short[1];
     }
     double ruleSupport = 0.0;
     if( sensoryInput[INPUT_DIM_NUM-1].getValues()[0] == S_TRAIN ){
        //showFeatures( "train sample = ", trainSample );

        if(setIdx < 4)
           ruleSupport += 0.5;

        // fill in the action dimensions of same value.
        controlAction[ControlActionFormat.ACT_REP][0] = ControlActionFormat.ACT_REP_ACT_NACS;
        controlAction[ControlActionFormat.ASSIM_EXP][0] = ControlActionFormat.ASSIM_EXP_USER_ASSIM;
        controlAction[ControlActionFormat.INFER_RETR][0] = ControlActionFormat.INFER_RETR_DO_NOTHING;
        controlAction[ControlActionFormat.SET_STRENTH][0] = ControlActionFormat.SET_STRENTH_DO_NOTHING;
        controlAction[ControlActionFormat.PARA_GKS_RSN_MTHD][0] = ControlActionFormat.GKS_RSN_MTHD_FORWARD;
        controlAction[ControlActionFormat.PARA_AMNS][0] = 0;

        // fill in the action dimensions of different values.
        switch (sensoryInput[INPUT_DIM_NUM-2].getValues()[0] ){
           case L_CON:
                ruleSupport += (setIdx == 0) ? 0.5 : 0.0;
                controlAction[ControlActionFormat.CHOOSE_LEVEL][0] = ControlActionFormat.CHOOSE_LEVEL_AMN;
                controlAction[ControlActionFormat.ENCODE_EXT][0] = ControlActionFormat.ENCODE_EXT_DO_NOTHING;
                break;
           case L_IMP:
                ruleSupport += (setIdx == 1) ? 0.5 : 0.0;
                controlAction[ControlActionFormat.CHOOSE_LEVEL][0] = ControlActionFormat.CHOOSE_LEVEL_AMN;
                controlAction[ControlActionFormat.ENCODE_EXT][0] = ControlActionFormat.ENCODE_EXT_DO_NOTHING;
                break;
           case L_EMP:
                ruleSupport += (setIdx == 2) ? 0.5 : 0.0;
                controlAction[ControlActionFormat.CHOOSE_LEVEL][0] = ControlActionFormat.CHOOSE_LEVEL_BOTH;
                controlAction[ControlActionFormat.ENCODE_EXT][0] = ControlActionFormat.ENCODE_EXT_RULES_USER;
                break;
           case L_SYN:
                ruleSupport += (setIdx == 3) ? 0.5 : 0.0;
                controlAction[ControlActionFormat.CHOOSE_LEVEL][0] = ControlActionFormat.CHOOSE_LEVEL_BOTH;
                controlAction[ControlActionFormat.ENCODE_EXT][0] = ControlActionFormat.ENCODE_EXT_RULES_USER;
                break;
         }
     }
     if( sensoryInput[INPUT_DIM_NUM-1].getValues()[0] == S_TEST ){
         //showFeatures( "test sample = ", testSample );

         if(setIdx >= 4)
           ruleSupport += 0.5;


        // fill in the action dimensions of same value.
        controlAction[ControlActionFormat.ACT_REP][0] = ControlActionFormat.ACT_REP_ACT_NACS;
        controlAction[ControlActionFormat.ENCODE_EXT][0] = ControlActionFormat.ENCODE_EXT_DO_NOTHING;
        controlAction[ControlActionFormat.ASSIM_EXP][0] = ControlActionFormat.ASSIM_EXP_DO_NOTHING;
        controlAction[ControlActionFormat.INFER_RETR][0] = ControlActionFormat.INFER_RETR_ONE;
        controlAction[ControlActionFormat.SET_STRENTH][0] = ControlActionFormat.SET_STRENTH_DO_NOTHING;
        controlAction[ControlActionFormat.PARA_GKS_RSN_MTHD][0] = ControlActionFormat.GKS_RSN_MTHD_FORWARD;
        controlAction[ControlActionFormat.PARA_AMNS][0] = 0;
        controlAction[ControlActionFormat.PARA_CHK_TYPES][0] = ControlActionFormat.CHK_TYPES_STATE;

        // fill in the action dimensions of different values.
        switch (sensoryInput[INPUT_DIM_NUM-2].getValues()[0] ){
           case L_CON:
                ruleSupport += (setIdx == 4) ? 0.5 : 0.0;
                controlAction[ControlActionFormat.CHOOSE_LEVEL][0] = ControlActionFormat.CHOOSE_LEVEL_AMN;
                controlAction[ControlActionFormat.PARA_COMB_MODE][0] = ControlActionFormat.COMB_DO_NOTHING;
                break;
           case L_IMP:
                ruleSupport += (setIdx == 4) ? 0.5 : 0.0;
                controlAction[ControlActionFormat.CHOOSE_LEVEL][0] = ControlActionFormat.CHOOSE_LEVEL_AMN;
                controlAction[ControlActionFormat.PARA_COMB_MODE][0] = ControlActionFormat.COMB_DO_NOTHING;
                break;
           case L_EMP:
                ruleSupport += (setIdx == 5) ? 0.5 : 0.0;
                controlAction[ControlActionFormat.CHOOSE_LEVEL][0] = ControlActionFormat.CHOOSE_LEVEL_BOTH;
                controlAction[ControlActionFormat.PARA_COMB_MODE][0] = ControlActionFormat.COMB_USER_TDG;
                break;
           case L_SYN:
                ruleSupport += (setIdx == 6) ? 0.5 : 0.0;
                controlAction[ControlActionFormat.CHOOSE_LEVEL][0] = ControlActionFormat.CHOOSE_LEVEL_BOTH;
                controlAction[ControlActionFormat.PARA_COMB_MODE][0] = ControlActionFormat.COMB_USER_BUR;
                break;
         }
     }

     return ruleSupport;

     /*
      public void training( int learningType )
      {
          //showFeatures( "train sample = ", trainSample );
          trainAmn( learningType );
          if( learningType == EXPLICIT_TYPE || learningType == SYNERGY_TYPE )
              trainGks();
      }

      public void testing()
      {
        //showFeatures( "test sample = ", testSample );

        searchNum = 0;
        gks.resetRT();
        amns[ activeAmnIdx ].resetRT();
        if( learningType == CONTROL_TYPE || learningType == IMPLICIT_TYPE ){
            amnInput = testSample;
            testAmn();
            completion = amnOutput;
        }
        if( learningType == EXPLICIT_TYPE )
            combineByTopDownGuidance();
        if( learningType == SYNERGY_TYPE )
            combineByBottomUpRectification();

        //System.out.println( "Search Num = " + searchNum );
        //totSearchNum += searchNum;
        //System.out.println( "Total search Num = " + totSearchNum );
      }
      */
  }

  /** Perform the process of encoding external knowledge.
   *  @param action current action. */
  public void performEncodeExternalKnowledge( short[][] action )
  {
     //System.out.println( "in training GKS" );
     short[] conditions = new short[ 2 ];
     int beginChunkIdx = gks.getChunkIdx( "begin letter" );
     if( beginChunkIdx == global.EMPTY ){
         //System.out.println( "create begin chunk in TaskClarionAGL.java" );
         setupBeginLetterChunk();
     }
     beginChunkIdx = gks.getChunkIdx( "begin letter" );
     conditions[ 0 ] = (short)beginChunkIdx;
     conditions[ 1 ] = (short)beginChunkIdx;
     int length = trainSample.length;
     for( int i = 0; i < length; i++ ){
          Feature[] features = new Feature[ 1 ];
          features[ 0 ] = trainSample[ i ];
          short conclusion = (short)gks.getChunkIdx( features );
          if( conclusion == global.EMPTY )
              conclusion = (short)gks.insertChunk( 0, 0, 0, "letterChunk", features );
          if( Math.random() < encodeProb )
              gks.insertRule( conditions, conclusion );
          conditions[ 0 ] = conditions[ 1 ];
          conditions[ 1 ] = conclusion;
     }
  }



  /** new routine. Performs encoding external knowledge in NACS. */
  /*
  public void performEncodeExternalKnowledge( short[][] action )
  {
     //System.out.println( "in training GKS" );
     short[] conditions = new short[ 2 ];
     int beginLetterChunkIdx = gks.getChunkIdx( "begin letter" );
     if( beginLetterChunkIdx == global.EMPTY ){
         System.out.println( "create begin letter chunk in TaskClarionAGL.java" );
         setupBeginLetterChunk();
         beginLetterChunkIdx = gks.getChunkIdx( "begin letter" );
     }
     int beginPositionChunkIdx = gks.getChunkIdx( "begin position" );
     if( beginPositionChunkIdx == global.EMPTY ){
         System.out.println( "create begin position chunk in TaskClarionAGL.java" );
         setupBeginPositionChunk();
         beginPositionChunkIdx = gks.getChunkIdx( "begin position" );
     }
     conditions[ 0 ] = (short)beginLetterChunkIdx;
     conditions[ 1 ] = (short)beginPositionChunkIdx;
     int length = trainSample.length;

     //System.out.println("sample len = " + length);
     //System.out.println("feature len = " + global.getFeatureLength(Global.FT_OUTPUT, OUTPUT_DIM_NUM-1) );
     //showFeatures( "train sample = ", trainSample );


     for( int i = 0; i < length; i++ ){
          Feature[] features = new Feature[ 1 ];
          features[ 0 ] = trainSample[ i ];
          short conclusion = (short)gks.getChunkIdx( features );
          if( conclusion == global.EMPTY ){
              System.out.println( this.getClass().getName() + ", chunk create error " );
              conclusion = (short)gks.insertChunk( 0, 0, 0, "letterChunk", features );
          }
          if( Math.random() <= encodeProb )
              gks.insertRule( conditions, conclusion );
          // update the conditions.
          conditions[ 0 ] = conclusion;
          short[] values = new short[ 1 ];
          values[ 0 ] = (short)(i + 1);
          features[ 0 ] = new Feature(global.getFeatureNames(Global.FT_OUTPUT, OUTPUT_DIM_NUM-1),
                              (short)Global.FT_OUTPUT,(short)(OUTPUT_DIM_NUM-1),
                              (short)global.getFeatureLength(Global.FT_OUTPUT, OUTPUT_DIM_NUM-1),values,global);
          conditions[ 1 ] = (short)gks.getChunkIdx( features );
          if( conditions[1] == global.EMPTY )
              conditions[1] = (short)gks.insertChunk( 0, 0, 0, "positionChunk", features );
     }
  }
  */

 /** Performs the process of assimilation in NACS.
   * @param aciton current action. */
  public void performNacsAssimilation( short[][] action )
  {
     //System.out.println( "in training AMN" );
     int inputLength = trainSample.length;
     int iterNum = ITER_NUMS[ learningType ];
     // generate samples.
     Vector sands = new Vector( inputLength );
     for( int i = 0; i < inputLength; i++ )
          sands.add( new Integer( i ) );
     int[] sampleNums = new int[ inputLength - 1 ];
     int sampleNum = calcSampleNums( inputLength, sampleNums );
     Vector[] samples = new Vector[ sampleNum ];
     int offset = 0;
     for( int i = 0; i < inputLength-1; i++ ){
        for( int j = 0; j < sampleNums[i]; j++ ){
             int size = i + 2;
             samples[ offset + j ] = new Vector( size );
             Vector s = (Vector)(sands.clone());
             for( int k = 0; k < size; k++ )
                  samples[ offset + j ].addElement(s.remove((int)(Math.random()*s.size())));
        }
        offset += sampleNums[i];
     }
     // train AMN.

     //System.out.println( "activeAmnIdx = " + activeAmnIdx );

     // construct desired output.
     /*
     int length = trainSample.length;
     Feature[] desiredOutput = new Feature[ length ];
     for( int i = 0; i < length; i++ ){
           short type = Global.FT_OUTPUT;
           short order = (short)trainSample[i].getOrder();
           String label = global.getFeatureNames( type, order );
           int fLength = global.getFeatureLength( type, order );
           short[] actVals = CommonMethods.copy( trainSample[i].getValues() );
           desiredOutput[i] = new Feature( label, type, order, (short)fLength, actVals, global );
     }
     amns[activeAmnIdx].setDesiredOutput( desiredOutput );
     */
     // set desired output to currently active AMN.
     amns[activeAmnIdx].setDesiredOutput( trainSample );
     for( int i = 0; i < iterNum; i++ ){
          sands = new Vector( sampleNum );
          for( int j = 0; j < sampleNum; j++ )
               sands.add( new Integer( j ) );
          for( int j = 0; j < sampleNum; j++ ){
               int sampleIdx = ((Integer)(sands.remove((int)(Math.random()*sands.size())))).intValue();
               int size = samples[sampleIdx].size();
               Feature[] inputs = new Feature[ inputLength ];
               for( int k = 0; k < size; k++ ){
                    int dimIdx = ((Integer)(samples[sampleIdx].elementAt( k ))).intValue();
                    inputs[dimIdx] = (Feature)(trainSample[dimIdx].clone());
               }
               amns[activeAmnIdx].setInput( inputs );
               amns[activeAmnIdx].training();
          }
     }
  }

 /** Performs the Bottom-Up Rectification combination. */
  public void combineByBottomUpRectification()
  {
     completion = null;
     if( testSample == null || learningType != SYNERGY_TYPE )
         return;

     // if( learningType == SYNERGY_TYPE )
     //     combineByBottomUpRectification();

     amnInput = testSample;
     testAmn();
     gksInput = amnOutput;
     validateByGks();
     completion = gksOutput;
  }

  /** Performs the Top-down Guidance combination. */
  public void combineByTopDownGuidance()
  {
     completion = null;
     if( testSample == null || learningType != EXPLICIT_TYPE )
         return;

     //if( learningType == EXPLICIT_TYPE )
     //    combineByTopDownGuidance();


     gksInput = testSample;
     boolean flag = dfsSearch();
     if( flag )
         completion = gksOutput;
     else{
         amnInput = gksOutput;
         testAmn();
         completion = amnOutput;
     }

      //System.out.println( "TaskClarionAGL.java, completion len = " + completion.length +
      //                    ", test sample len = " + testSample.length +
      //                    ", Top search succes flag = " + flag );

  }

  /** Search in GKS using DFS search. */
  private boolean dfsSearch()
  {
    boolean completeFlag = false;
     int length = gksInput.length;
     short[] conditions = new short[ 2 ];
     conditions[ 0 ] = (short)gks.getChunkIdx( "begin letter" );
     conditions[ 1 ] = (short)gks.getChunkIdx( "begin letter" );
     short[] maxPartial = new short[ length + 2 ];
     short[] dynPartial = new short[ length + 2 ];
     for( int i = 0; i < length + 2; i++ ){
          maxPartial[ i ] = (short)global.EMPTY;
          dynPartial[ i ] = (short)global.EMPTY;
     }
     maxPartial[ 0 ] = conditions[ 0 ];
     maxPartial[ 1 ] = conditions[ 1 ];
     dynPartial[ 0 ] = conditions[ 0 ];
     dynPartial[ 1 ] = conditions[ 1 ];
     short[] filledLocs = new short[ 2 ];
     int count = 0;
     for( int i = 0; i < length; i++ )
          if( gksInput[i].getValueNum() == 1 )
              filledLocs[count++] = (short)i;
     // construct stack for Depth First Search.
     Stack stack = new Stack();
     Vector beginElem = new Vector( 2 );
     beginElem.addElement( new Integer(conditions[0]) );
     beginElem.addElement( new Integer(conditions[1]) );
     stack.push( beginElem );
     Vector conc = gks.searchConclusions( conditions );
     int idx = 0;
     while( idx < conc.size() ){
          if( Math.random() > RULE_APP_PROB )
              conc.removeElementAt( idx );
          else
              idx++;
     }
     stack.push( conc );
     searchNum++;
     int curPos = 1;
     int exitCode = 0;
     while( exitCode == 0 ){
        if( curPos == 0 ){
            exitCode = -1;
            continue;
        }
        Vector top = (Vector)stack.peek();
        int curChunkIdx;
        Feature feature;
        if( top.isEmpty() ){ // this is dead path, needs replace the last letter if exists.
            stack.pop();
            curPos--;
            ((Vector)stack.peek()).removeElementAt( 0 );
            continue;
        }
        if( curPos == filledLocs[1] + 1 ){ // reach the second filled letter( the end )
            curChunkIdx = ((Integer)top.elementAt(0)).intValue();
            feature = gks.getChunk(curChunkIdx).getFeature(0);
            //if( feature.equals( gksInput[ filledLocs[1] ] ) && Math.random() < 0.8 ){
            if( feature.equals( gksInput[ filledLocs[1] ] ) ){
                dynPartial[ curPos + 1 ] = (short)curChunkIdx;
                for( int i = 0; i < curPos + 2; i++ )
                     maxPartial[i] = dynPartial[i];
                exitCode = 1;
                continue;
            }
            else
                top.removeElementAt(0);
        }
        else{
            if( curPos == filledLocs[0] + 1 ){ // reach the first filled letter.
                idx = 0;
                while( idx < top.size() ){
                   curChunkIdx = ((Integer)top.elementAt(idx)).intValue();
                   feature = gks.getChunk(curChunkIdx).getFeature(0);
                   // if mismatched, remove it.
                   if( !feature.equals( gksInput[ filledLocs[0] ] ) )
                        top.removeElementAt( idx );
                   else{ // if matched, record maximum partial string currently available.
                        dynPartial[ curPos + 1 ] = (short)curChunkIdx;
                        for( int i = 0; i < curPos + 2; i++ )
                             maxPartial[i] = dynPartial[i];
                        idx++;
                   }
                }
                if( top.isEmpty() ) // no match exists.
                    continue;
            }
            curChunkIdx = ((Integer)top.elementAt(0)).intValue();
            dynPartial[ curPos + 1 ] = (short)curChunkIdx;
            conditions[ 0 ] = dynPartial[ curPos ];
            conditions[ 1 ] = dynPartial[ curPos + 1 ];
            conc = gks.searchConclusions( conditions );
            idx = 0;
            while( idx < conc.size() ){
                   if( Math.random() > RULE_APP_PROB )
                       conc.removeElementAt( idx );
                   else
                       idx++;
            }
            stack.push( conc );
            searchNum++;
            curPos++;
        }
     }
     if( exitCode == -1 ){ // fails.
         for( int i = 0; i < 2; i++ )
            if( maxPartial[filledLocs[i] + 2] == global.EMPTY ){
                Feature[] features = new Feature[ 1 ];
                features[0] = gksInput[filledLocs[i]];
                maxPartial[filledLocs[i] + 2] = (short)gks.getChunkIdx( features );
            }
     }
     gksOutput = new Feature[ length ];
     for( int i = 0; i < length; i++ ){
          if( maxPartial[i + 2] != global.EMPTY && Math.random() < 0.98 ){
              gksOutput[i] = gks.getChunk( maxPartial[i + 2] ).getFeature( 0 );
          }
          else
              gksOutput[i] = (Feature)null;
     }

     //showFeatures( "gks output = ", gksOutput );

     if( exitCode == 1 )
         completeFlag = true;
     return completeFlag;
 }

 /** Perform the process of validation by GKS. */
 private boolean validateByGks()
  {
     boolean completeFlag = false;
     int length = testSample.length;
     short[] conditions = new short[ 2 ];
     conditions[ 0 ] = (short)gks.getChunkIdx( "begin letter" );
     conditions[ 1 ] = (short)gks.getChunkIdx( "begin letter" );
     short[] dynPartial = new short[ length + 2 ];
     Vector[] candidateConclusions = new Vector[ length + 2 ];
     for( int i = 0; i < length + 2; i++ )
          dynPartial[ i ] = (short)global.EMPTY;
     dynPartial[ 0 ] = conditions[ 0 ];
     dynPartial[ 1 ] = conditions[ 1 ];
     int curPos = 0;  // indicates current position needs validation.
     double failNum = 0.0;
     while( curPos < length ){
          Vector conclusions = gks.searchConclusions( conditions );
          searchNum++;
          Feature[] features = new Feature[ 1 ];
          features[ 0 ] = gksInput[ curPos ];
          int curChunkIdx = gks.getChunkIdx( features );
          boolean inFlag = false;
          if( !conclusions.isEmpty() ){
              for( int i = 0; i < conclusions.size(); i++ )
                   if( ((Integer)conclusions.elementAt(i)).intValue() == curChunkIdx ){
                       inFlag = true;
                       break;
                   }
          }
          if( inFlag ){
              dynPartial[ curPos + 2 ] = (short)curChunkIdx;
              conditions[ 0 ] = conditions[ 1 ];
              conditions[ 1 ] = (short)dynPartial[curPos + 2];
              curPos++;
              continue;
          }
          // no match. determine failure or not.
          failNum = length - curPos;
          //if( Math.random() > failNum / length ) // if determine not failure.
          //    break;
          // candidate string rectified by GKS.
           if( conclusions == null || conclusions.isEmpty() ){
               conditions[ 0 ] = conditions[ 1 ];
               conditions[ 1 ] = (short)curChunkIdx;
               curPos++;
               continue;
           }
           if( conclusions.size() == 1 || curPos == length -1 ){
               int idx = (int)(Math.random() * conclusions.size());
               dynPartial[curPos + 2] = (short)((Integer)conclusions.elementAt(idx)).intValue();
               conditions[ 0 ] = conditions[ 1 ];
               conditions[ 1 ] = (short)dynPartial[curPos + 2];
               curPos++;
               continue;
           }
           // find the most suitable letter to fill in the failed validaiton position.
           Vector[] newConclusions = new Vector[3];
           for( int j = 0; j < 3; j++ )
                newConclusions[j] = new Vector();
           newConclusions[0] = conclusions;
           int offset = 0;
           while( offset < 2 ){
                 if(curPos + offset >= length - 1)
                    break;
                 features[ 0 ] = gksInput[curPos + offset + 1];
                 int concChunkIdx = gks.getChunkIdx( features );
                 short[][] ruleConditions = gks.getConditions( concChunkIdx );
                 if(ruleConditions == null){
                    offset++;
                    continue;
                 }
                 short[] conds = new short[2];
                 for (int j = 0; j < ruleConditions.length; j++) {
                       Arrays.sort(ruleConditions[j]);
                       boolean flag = false;
                       for (int k = 0; k < newConclusions[offset].size(); k++) {
                           short chunkIdx = (short) ( (Integer) newConclusions[offset].elementAt(k)).intValue();
                           conds[0] = (offset == 0) ? dynPartial[curPos + 1] : chunkIdx;
                           features[ 0 ] = gksInput[curPos + offset];
                           conds[1] = (offset == 0) ? chunkIdx : (short)gks.getChunkIdx( features );
                           Arrays.sort(conds);
                           if (conds[0] == ruleConditions[j][0] && conds[1] == ruleConditions[j][1])
                               newConclusions[offset+1].addElement( new Integer(chunkIdx) );
                       }
                 }
                 if( newConclusions[offset+1].isEmpty() )
                     break;
                 offset++;
           }
           Vector candidate = (!newConclusions[2].isEmpty()) ? newConclusions[2] :
                              (!newConclusions[1].isEmpty()) ? newConclusions[1] : newConclusions[0];
           int idx = (int)(Math.random() * candidate.size());
           dynPartial[curPos + 2] = (short)((Integer)candidate.elementAt(idx)).intValue();
           conditions[ 0 ] = conditions[ 1 ];
           conditions[ 1 ] = (short)dynPartial[curPos + 2];
           curPos++;
     } // end of while( curPos < length )

     //showFeatures( "gks output = ", gksOutput );

     gksOutput = new Feature[ length ];
     for( int i = 0; i < length; i++ )
       if (dynPartial[i + 2] != Global.EMPTY)
          gksOutput[i] = gks.getChunk(dynPartial[i + 2]).getFeature(0);
       else
          gksOutput[i] = null;

     completeFlag = true;
     return completeFlag;
  }


  /*
  private boolean dfsSearch()
  {
     boolean completeFlag = false;
     int length = gksInput.length;
     short[] conditions = new short[ 2 ];
     conditions[ 0 ] = (short)gks.getChunkIdx( "begin letter" );
     conditions[ 1 ] = (short)gks.getChunkIdx( "begin position" );
     short[] maxPartial = new short[ length + COND_NUM ];
     short[] dynPartial = new short[ length + COND_NUM ];
     for( int i = 0; i < length + COND_NUM; i++ ){
          maxPartial[ i ] = (short)global.EMPTY;
          dynPartial[ i ] = (short)global.EMPTY;
     }
     maxPartial[ 0 ] = conditions[ 0 ];
     dynPartial[ 0 ] = conditions[ 0 ];
     // get the filled-in positions.
     short[] filledLocs = new short[ 2 ];
     int count = 0;
     for( int i = 0; i < length; i++ )
          if( gksInput[i].getValueNum() == 1 )
              filledLocs[count++] = (short)i;
     // construct stack for Depth First Search.
     Stack stack = new Stack();
     Vector beginElem = new Vector( COND_NUM );
     beginElem.addElement( new Integer( conditions[0] ) );
     stack.push( beginElem );
     Vector conc = gks.searchConclusions( conditions );
     int idx = 0;
     while( idx < conc.size() ){
          if( Math.random() > RULE_APP_PROB )
              conc.removeElementAt( idx );
          else
              idx++;
     }
     stack.push( conc );
     searchNum++;
     int curPos = 1;
     // init current position variable
     short[] posValues = new short[ 1 ];
     posValues[0] = (short)curPos;
     int exitCode = 0;
     while( exitCode == 0 ){
        if( curPos == 0 ){
            exitCode = -1;
            continue;
        }
        Vector top = (Vector)stack.peek();
        int curChunkIdx;
        Feature feature;
        if( top.isEmpty() ){ // this is dead path, needs replace the last letter if exists.
            stack.pop();
            curPos--;
            posValues[0] = (short)curPos;
            ((Vector)stack.peek()).removeElementAt( 0 );
            continue;
        }
        if( curPos == filledLocs[1] + 1 ){ // reach the second filled letter( the end )
            curChunkIdx = ((Integer)top.elementAt(0)).intValue();
            feature = gks.getChunk(curChunkIdx).getFeature(0);
            if( feature.equals( gksInput[ filledLocs[1] ] ) ){
                dynPartial[ curPos ] = (short)curChunkIdx;
                for( int i = 0; i < curPos + 1; i++ )
                     maxPartial[i] = dynPartial[i];
                exitCode = 1;
                continue;
            }
            else
                top.removeElementAt(0);
        }
        else{
            if( curPos == filledLocs[0] + 1 ){ // reach the first filled letter.
                idx = 0;
                while( idx < top.size() ){
                   curChunkIdx = ((Integer)top.elementAt(idx)).intValue();
                   feature = gks.getChunk(curChunkIdx).getFeature(0);
                   // if mismatched, remove it.
                   if( !feature.equals( gksInput[ filledLocs[0] ] ) )
                        top.removeElementAt( idx );
                   else{ // if matched, record maximum partial string currently available.
                        dynPartial[ curPos ] = (short)curChunkIdx;
                        for( int i = 0; i < curPos + 1; i++ )
                             maxPartial[i] = dynPartial[i];
                        idx++;
                   }
                }
                if( top.isEmpty() ) // no match exists.
                    continue;
            }
            // fill in the conditions.
            curChunkIdx = ((Integer)top.elementAt(0)).intValue();
            dynPartial[ curPos ] = (short)curChunkIdx;
            conditions[ 0 ] = dynPartial[ curPos ];
            Feature[] features = new Feature[ 1 ];
            features[0] = new Feature(global.getFeatureNames(Global.FT_OUTPUT, OUTPUT_DIM_NUM-1),
                              (short)Global.FT_OUTPUT,(short)(OUTPUT_DIM_NUM-1),
                              (short)global.getFeatureLength(Global.FT_OUTPUT, OUTPUT_DIM_NUM-1),posValues,global);
            conditions[ 1 ] = (short)gks.getChunkIdx( features );
            if( conditions[1] == global.EMPTY ){
                System.out.println( "conditions error in TaskClarionAGL.java" );
                exitCode = -1;
                continue;
            }
            conc = gks.searchConclusions( conditions );
            idx = 0;
            while( idx < conc.size() ){
                   if( Math.random() > RULE_APP_PROB )
                       conc.removeElementAt( idx );
                   else
                       idx++;
            }
            stack.push( conc );
            searchNum++;
            curPos++;
            posValues[0] = (short)curPos;


            String str = "condition[0] = " + gks.getChunk(conditions[0]).getFeature(0);
            str += ", condition[1] = " + gks.getChunk(conditions[1]).getFeature(0);
            str +=", conclusion = ";
            for( int i = 0; i < conc.size(); i++ )
                 str += gks.getChunk(((Integer)conc.elementAt(i)).intValue()).getFeature(0) + ", ";
            System.out.println( str );


        }
     }


     // fill in the filled-in positions of the testing sample.
     if( exitCode == -1 ){ // fails.
         for( int i = 0; i < 2; i++ )
            if( maxPartial[filledLocs[i] + COND_NUM] == global.EMPTY ){
                Feature[] features = new Feature[ 1 ];
                features[0] = gksInput[filledLocs[i]];
                maxPartial[filledLocs[i] + COND_NUM] = (short)gks.getChunkIdx( features );
            }
     }
     // get output from GKS.
     gksOutput = new Feature[ length ];
     for( int i = 0; i < length; i++ ){
          if( maxPartial[i + COND_NUM] != global.EMPTY ){
              gksOutput[i] = gks.getChunk( maxPartial[i + COND_NUM] ).getFeature( 0 );
          }
          else
              gksOutput[i] = (Feature)null;
     }

     //showFeatures( "gks output = ", gksOutput );

     if( exitCode == 1 )
         completeFlag = true;
     return completeFlag;
  }
  */


  /*
  private boolean validateByGks()
  {
     boolean completeFlag = false;
     int length = testSample.length;
     short[] conditions = new short[ 2 ];
     conditions[ 0 ] = (short)gks.getChunkIdx( "begin letter" );
     conditions[ 1 ] = (short)gks.getChunkIdx( "begin position" );
     short[] dynPartial = new short[ length + COND_NUM ];
     for( int i = 0; i < length + COND_NUM; i++ )
          dynPartial[ i ] = (short)global.EMPTY;
     dynPartial[ 0 ] = conditions[ 0 ];
     // init current position
     short[] posValues = new short[ 1 ];
     posValues[ 0 ] = 1;
     int idx = 0;    // indicates current position needs validation.
     while( idx < length ){
          Vector conclusions = gks.searchConclusions( conditions );
          searchNum++;
          Feature[] features = new Feature[ 1 ];
          features[ 0 ] = gksInput[ idx ];
          int curChunkIdx = gks.getChunkIdx( features );
          boolean inFlag = false;
          if( !conclusions.isEmpty() ){
              for( int i = 0; i < conclusions.size(); i++ )
                   if( ((Integer)conclusions.elementAt(i)).intValue() == curChunkIdx ){
                       inFlag = true;
                       break;
                   }
          }
          if( inFlag ){
              dynPartial[ idx + COND_NUM ] = (short)curChunkIdx;
              idx++;
              if( idx < length ){
                  posValues[0] = (short)(idx + 1);
                  // fill in conditions.
                  conditions[ 0 ] = (short)curChunkIdx;
                  features = new Feature[ 1 ];
                  features[0] = new Feature(global.getFeatureNames(Global.FT_OUTPUT, OUTPUT_DIM_NUM-1),
                              (short)Global.FT_OUTPUT,(short)(OUTPUT_DIM_NUM-1),
                              (short)global.getFeatureLength(Global.FT_OUTPUT, OUTPUT_DIM_NUM-1),posValues,global);
                  conditions[ 1 ] = (short)gks.getChunkIdx( features );
                  if( conditions[1] == global.EMPTY ){
                      System.out.println("conditions error in TaskClarionAGL.java" );
                      //System.exit( 1 );        // new revision 11/01/04
                  }
               }
               continue;
          }
          // no match.
          // conclusions is empty
          idx += COND_NUM + 1;
          posValues[0] = (short)(idx + 1);
          if( idx < length ){
              if( gksInput[idx-1].getValueNum() > 0 ){
                  for( int i = 0; i < COND_NUM; i++ ){
                       features[ 0 ] = gksInput[ idx - COND_NUM + i ];
                       conditions[ i ] = (short)gks.getChunkIdx( features );
                       if( conditions[ i ] == global.EMPTY ){
                           System.out.println( "chunk null error in condition in TaskClarionAGL.java" );
                           //System.out.println( "feature = " + features[0] );   // new revision 11/01/04
                           //System.exit( 1 );
                       }
                  }
              }
              else{
                  System.out.println( "feature null error in TaskClarionAGL.java" );
                  break;
              }
          }
     } // end of while( idx < length )
     // fill in GKS output.
     int failNum = 0;
     gksOutput = new Feature[ length ];
     for( int i = 0; i < length; i++ ){
          if( dynPartial[i + COND_NUM] != global.EMPTY ){
              gksOutput[i] = gks.getChunk( dynPartial[i + COND_NUM] ).getFeature( 0 );
          }
          else{
              gksOutput[i] = (Feature)null;
              failNum++;
          }
     }

     //showFeatures( "gks output = ", gksOutput );

     //if( Math.random() >= (double)failNum/length )
     completeFlag = true;
     return completeFlag;
  }
  */

  private void showGKS()
  {
    System.out.println( gks );
  }

  private void testAmn()
  {
     amns[activeAmnIdx].setInput( amnInput );
     amns[activeAmnIdx].retrieval( false );
     amnOutput = amns[activeAmnIdx].getResults();
  }

  private int calcSampleNums( int length, int[] sampleNums )
  {
     int sampleNum = 0;
     for( int i = 2; i <= length; i++ ){
          long num = 1;
          for( int j = length; j > length-i; j-- )
               num *= j;
          for( int j = 1; j < i+1; j++ )
               num /= j;
          sampleNums[ i-2 ] = 1 + (int)num/2;
          sampleNum += 1 + (int)num/2;
     }
     return sampleNum;
  }

  private void setupBeginLetterChunk()
  {
     gks.insertChunk( 0, 0, 0, "begin letter", beginFeatures[0] );
  }

  private void setupBeginPositionChunk()
  {
     gks.insertChunk( 0, 0, 0, "begin position", beginFeatures[1] );
  }

  private void showFeatures( String head, Feature[] features )
  {
     String str = head;
     for( int i = 0; i < features.length; i++ )
          str += features[i];
     System.out.println( str );
  }
}
